/*----------------------------------
 * i2dghw.c
 *
 * Contains the routines to 
 * support the low-level diagnostics
 * used by ipdiag.c
 *---------------------------------
 */

/*---------------------------------------------------------------
 * Revision History:
 * 6/25/91	MAG	1.02	moved message in interrupt test:
 *						at verbosity > 1 WAS printing "0 spurrious..."
 * 7/19/91	MAG	1.10.2	Interrupt routine fixed.  IT_FAIL only set 
 * 12/12/91	MAG	2.00.4	Adding microchannel support.
 * 12/13/91 MAG 2.01.0	Add pos_checkout()
 *  3/24/92 MAG 2.01.8	FIFO mailbox test: board changes the data
 *						so we know we're not just looking at our own
 *						mailbox.
 *  3/24/92 MAG 2.01.9	POS test: if it fails, we restore POS contents
 *						re-enable interrupts, and so on before returning.
 *						(Previous would return without cleanup).
 *						
 *  4/02/92 MAG 2.02.0	EISA support added
 *						In interrupt handling routine, set IRQ mask bit
 *						to prevent extra level-trigger ints, then
 *						EOI first (so we don't lose edge-triggered ints), then 
 * 						service the interrupt.				
 *						Also, routine improved to perform more interrupts
 *						faster (EISA/MCA) and to check for wrong interrupts
 *						more thoroughly (ISA)
 *						findMCAslot() and findEISAslot changed so 
 *						addresses of the address and interrupt are passed.
 *						If either value is 0, it is treated as a wild card
 *						and it matches whatever card is there.
 *	6/02/92	MAG	2.02.2	dma test now clears two low bits before testing
 *						the values: low two mailbox bits are indeterminate.
 *  7/23/92	MAG	2.02.6	Added register_exercise
 *  3/15/93	MAG	2.02.7	Added "missed" to interrupt test.  Previous versions
 *						could pass, by accident, a motherboard which didn't
 *						handle properly level triggered EISA interrupts.
 *  7/08/93 MAG 2.03.1	Changes to support asic version of EISA-8/CEX.
 *						As of this rev, the regular interrupt test tests
 *						EISA interrupts in level-trigger mode, regardless
 *						of how configured.  The high-level test tests 
 *						them in edge-triggered mode.
 *  2/17/93 MAG 2.04.2	Times for memory test increased from 3, 16 to 6,32
 *  6/22/95 MAG 2.05.0	PCI support added: recognition, interrupt test.
 *  1/23/97	MAG	3.00.0	ST654 support added.
 *---------------------------------------------------------------
 */

typedef unsigned short ushort;
typedef unsigned char uchar;

#include <dos.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <time.h>
#include <conio.h>

#include "i2dghw.h"
#include "i2lib.h"

//function prototypes
void interrupt i_handler_chief (void);
void interrupt i_handler_3 (void);
void interrupt i_handler_4 (void);
void interrupt i_handler_5 (void);
void interrupt i_handler_7 (void);
void interrupt i_handler_10 (void);
void interrupt i_handler_11 (void);
void interrupt i_handler_12 (void);
void interrupt i_handler_15 (void);
static int wait_mailbox_in (i2eBordStrPtr, int);
static int byte_data_in4i (i2eBordStrPtr, int);
static int word_data_read4i (i2eBordStrPtr, int, ushort *);
static int byte_data_out (i2eBordStrPtr, int, uchar);
static int word_data_out (i2eBordStrPtr, int, ushort);
static int word16_data_out (i2eBordStrPtr, int, ushort);
static int byte_data_in (i2eBordStrPtr, int);
static int byte_data_out8e (i2eBordStrPtr, int, uchar);
static int word_data_out8e (i2eBordStrPtr, int, ushort);
static int word_data_read (i2eBordStrPtr, int, unsigned short *);
static int word16_data_read (i2eBordStrPtr, int, unsigned short *);
static int byte_data_read (i2eBordStrPtr, int, unsigned char *);
static int memory_diag (pfunk, int);
static int do_an_interrupt (i2eBordStrPtr, int, int);
static int checkInterrupts(void);
static int fifo_4701_test (int, int, uchar *, uchar *);
static int fifo_local_stress8 (pfunk);
static int fifo_local_stress16 (pfunk);
static void sendLcdString(i2eBordStrPtr, char *);
static int waitForAnyKey(i2eBordStrPtr);
static void waitForSpecificKey(i2eBordStrPtr, int);
static void interactive_lcd_one(i2eBordStrPtr, int);
static int selMCAirq(int);
static int selEISAirq(int);
static int convertMask(int);
static int levelsForEISA(int);

/*-----------------------------------
 * Statics for the interrupt handlers
 * to use.
 *-----------------------------------
 */
#define LS_SIZE 64 
static unsigned short localStack[LS_SIZE + 2];
static i2eBordStrPtr pTestBoard;


/*-----------------------------------
 * Handy macros for winding up a
 * routine: stores error condition
 * and returns.
 *-----------------------------------
 */

#define MEM_ERROR(f,tipe,e1, e2)	\
	if(1) {							\
	f->f_error = (tipe ? e1 : e2);	\
	return -1;						\
	}

#define POST_ERROR(f,e1)	\
	if(1) {					\
	f->f_error = e1;		\
	return -1;				\
	}
#define SUCCEED(f)			\
	if(1) {					\
	f->f_error = OKAY;		\
	return 0;				\
	}


/* Table for mapping MC-8 or MC-EX pos[1] (setup address 103)
   values into the selected i/o address.
 */

int	mcAddresses[] = {
	0,	0,	0,	0, 
	0x2020,	 0x2028,	 0x2030,	 0x2038,	
	0,	0,	0,	0, 
	0x2060,	 0x2068,	 0x2070,	 0x2078,	
	0x8880,	 0x8888,	 0x8890,	 0x8898,	
	0xa8a0,	 0xa8a8,	 0xa8b0,	 0xa8b8,	
	0x88c0,	 0x88c8,	 0x88d0,	 0x88d8,	
	0xa8e0,	 0xa8e8,	 0xa8f0,	 0xa8f8,	
	0,		 0x0108,	 0x0110,	 0x0118,	
	0x2120,	 0x2128,	 0x2130,	 0x2138,	
	0x0140,	 0x0148,	 0x0150,	 0x0158,	
	0x2160,	 0x2168,	 0x2170,	 0x2178,	
	0x8980,	 0x8988,	 0x8990,	 0x8998,	
	0xa9a0,	 0xa9a8,	 0xa9b0,	 0xa9b8,	
	0x89c0,	 0x89c8,	 0x89d0,	 0x89d8,	
	0xa9e0,	 0xa9e8,	 0xa9f0,	 0xa9f8,	
	0,		 0x0208,	 0x0210,	 0x0218,	
	0x2220,	 0x2228,	 0x2230,	 0x2238,	
	0x0240,	 0x0248,	 0x0250,	 0x0258,	
	0x2260,	 0x2268,	 0x2270,	 0x2278,	
	0x8a80,	 0x8a88,	 0x8a90,	 0x8a98,	
	0xaaa0,	 0xaaa8,	 0xaab0,	 0xaab8,	
	0x8ac0,	 0x8ac8,	 0x8ad0,	 0x8ad8,	
	0xaae0,	 0xaae8,	 0xaaf0,	 0xaaf8,	
	0,		 0x0308,	 0x0310,	 0x0318,	
	0x2320,	 0x2328,	 0x2330,	 0x2338,	
	0x0340,	 0x0348,	 0x0350,	 0x0358,	
	0x2360,	 0x2368,	 0x2370,	 0x2378,	
	0x8b80,	 0x8b88,	 0x8b90,	 0x8b98,	
	0xaba0,	 0xaba8,	 0xabb0,	 0xabb8,	
	0x8bc0,	 0x8bc8,	 0x8bd0,	 0x8bd8,	
	0xabe0,	 0xabe8,	 0xabf0,	 0xabf8,	
	0x5400,	 0x5408,	 0x5410,	 0x5418,	
	0x7420,	 0x7428,	 0x7430,	 0x7438,	
	0x5440,	 0x5448,	 0x5450,	 0x5458,	
	0x7460,	 0x7468,	 0x7470,	 0x7478,	
	0xdc80,	 0xdc88,	 0xdc90,	 0xdc98,	
	0xfca0,	 0xfca8,	 0xfcb0,	 0xfcb8,	
	0xdcc0,	 0xdcc8,	 0xdcd0,	 0xdcd8,	
	0xfce0,	 0xfce8,	 0xfcf0,	 0xfcf8,	
	0x5500,	 0x5508,	 0x5510,	 0x5518,	
	0x7520,	 0x7528,	 0x7530,	 0x7538,	
	0x5540,	 0x5548,	 0x5550,	 0x5558,	
	0x7560,	 0x7568,	 0x7570,	 0x7578,	
	0xdd80,	 0xdd88,	 0xdd90,	 0xdd98,	
	0xfda0,	 0xfda8,	 0xfdb0,	 0xfdb8,	
	0xddc0,	 0xddc8,	 0xddd0,	 0xddd8,	
	0xfde0,	 0xfde8,	 0xfdf0,	 0xfdf8,	
	0x5600,	 0x5608,	 0x5610,	 0x5618,	
	0x7620,	 0x7628,	 0x7630,	 0x7638,	
	0x5640,	 0x5648,	 0x5650,	 0x5658,	
	0x7660,	 0x7668,	 0x7670,	 0x7678,	
	0xde80,	 0xde88,	 0xde90,	 0xde98,	
	0xfea0,	 0xfea8,	 0xfeb0,	 0xfeb8,	
	0xdec0,	 0xdec8,	 0xded0,	 0xded8,	
	0xfee0,	 0xfee8,	 0xfef0,	 0xfef8,	
	0x5700,	 0x5708,	 0x5710,	 0x5718,	
	0x7720,	 0x7728,	 0x7730,	 0x7738,	
	0x5740,	 0x5748,	 0x5750,	 0x5758,	
	0x7760,	 0x7768,	 0x7770,	 0x7778,	
	0xdf80,	 0xdf88,	 0xdf90,	 0xdf98,	
	0xffa0,	 0xffa8,	 0xffb0,	 0xffb8,	
	0xdfc0,	 0xdfc8,	 0xdfd0,	 0xdfd8,	
	0xffe0,	 0xffe8,	 0xfff0,	 0xfff8,	
	};

/* Table to map high three bits of pos[0] (setup address 102) into
 * the selected interrupt.
 * 0x2 bit enables interrupt.
 */

int mcInterrupts[] = {
	3, 4, 5, 7, 10, 11, 12, 15 };



/*--------------------------------------------
 * selMCAirq(value)
 * Returns the currently selected pos[0] value,
 * and replaces it with the new one.
 * if value < 0, only saves: preserves original
 * contents.
 *--------------------------------------------
 */

static int
selMCAirq(int value)
{
int i;
int setup_save;

	if (!slotMorE) return 0;
	SAVE_AND_DISABLE_INTS
	setup_save = INB(0x96);

	OUTB(0x96, slotMorE);
	i = INB(0x102);
	if (value >= 0)
	{
		OUTB(0x102, value);
	}

	OUTB(0x96, setup_save);
	RESTORE_INTS
	return i;
}

/*--------------------------------------------
 * selEISAirq(value)
 * Returns the currently selected  zC86 value,
 * and replaces it with the new one.
 * if value < 0, only saves: preserves original
 * contents.
 *--------------------------------------------
 */

static int
selEISAirq(int value)
{
int i;

	if (!slotMorE) return 0;
	i = INB(slotMorE + 0xc86);

	if (value >= 0)
	{
		OUTB(slotMorE + 0xc86, value);
	}
	return i;
}



#define NO_CHANGE_LEVELS -1

/* For diagnostic purposes, we must guarantee that any interrupt
 * line we are testing on EISA is configured (temporarily at least)
 * for level-triggered operation.
 * (Except, of couse, that some cards can now do edges)
 */

static int
levelsForEISA(int value)
{
int i = NO_CHANGE_LEVELS;

	if (value != NO_CHANGE_LEVELS)
	{
		i = (INB(0x4d1) << 8) | INB(0x4d0);
		OUTB(0x4d1, value >> 8);
		OUTB(0x4d0, value);
	}
	return i;
}

/* The following converts IRQ_? masks used internally to
 * the bit mask for the Edge-Level configuration register
 * in EISA machines.
 */

static int
convertMask(register int bits)
{
register int mbits;
	mbits  = (bits & IRQ_3) ? ( 1 << 3) : 0;
	mbits |= (bits & IRQ_4) ? ( 1 << 4) : 0;
	mbits |= (bits & IRQ_5) ? ( 1 << 5) : 0;
	mbits |= (bits & IRQ_7) ? ( 1 << 7) : 0;
	mbits |= (bits & IRQ_10) ? ( 1 << 10) : 0;
	mbits |= (bits & IRQ_11) ? ( 1 << 11) : 0;
	mbits |= (bits & IRQ_12) ? ( 1 << 12) : 0;
	mbits |= (bits & IRQ_15) ? ( 1 << 15) : 0;
	return mbits;
}


/*--------------------------------------------
 * validMCA()
 * Returns true or false according to whether
 * we seem to be running on a Micro Channel
 * machine.  
 * This is determined by io (byte) address being
 * valid.
 *--------------------------------------------
 */

int
validMCA()
{
int i;
int goodness = 1;	/* good until proved bad */
int setup_save;

	SAVE_AND_DISABLE_INTS
	setup_save = INB(0x96);

	for (i = 0; i < 15; i++)
	{
		OUTB(0x96, i);
		if (i != (INB(0x96) & 0xf) )
		{
			goodness = 0;
			break;	 /* from the loop on values */
		}
	}
	OUTB(0x96, setup_save);
	RESTORE_INTS
	return goodness;
}



/*--------------------------------------------
 * validEISA()
 * Returns true or false according to whether
 * we seem to be running on an EISA 
 * machine.  
 * This is determined by io (byte) address being
 * valid.
 *--------------------------------------------
 */

int
validEISA()
{
int i,j, value;

	/* The idea is that on an EISA machine 0xxx, 1xxx, 2xxx
	 * will address different slots, but on an ISA they will be
	 * masked off to 10 bits.
	 * So if changing the top nibble changes the value read,
	 * this might be an EISA machine.  (It also might be a
	 * Micro Channel machine as well.)
	 * An EISA machine will always pass this test, however, because
	 * register zC80-3 select the product I.D registers.  z=0 for
	 * the motherboard.  You cannot install a system board in a
	 * system board.  Therefore the product I.D. will always differ.
	 */
	for (i = 0x0c80; i <= 0x0c83; i++)
	{
		value = INB(i);
		for (j = i + 0x4000; j > i; j -= 0x1000)
		{
			if (value != INB(j)) return 1;
		}
	}
	return 0;	
}


/*--------------------------------------------
 * findMCAslot()
 * Returns true or false according to whether
 * we have found a board for the selected address
 * in any Microchannel slot.
 * Stores the slot in slotMorE global
 * 
 * If argument irq == 0, then we don't care what
 * irq we were configured for.
 *--------------------------------------------
 */

int
findMCAslot(int * paddress, int * pirq)
{
register int i;
int goodness = 0;	/* bad until proved good */
int setup_save;
unsigned int id[8];
unsigned int value;
int setup_address[8];
int setup_irq[8];
int ismine[8];

	SAVE_AND_DISABLE_INTS
	setup_save = INB(0x96);

	for (i = 0; i < 7; i++)
	{
		OUTB(0x96, i | 8);	/* Select the slot */
		id[i] = (INB(0x101) << 8) | (INB(0x100) & 0xff);
		if (id[i] != 0x61f2	&&	/* MC-8 */
			id[i] != 0x61f3) 	/* MC-CEX */
		{
			ismine[i] = 0;
			continue;
		}
		ismine[i] = 1;
		/* It's some sort of microchannel card, but 
		 * at what address is it configured?
		 */

		value = (INB(0x103) & 0xff);
		setup_address[i] = mcAddresses[value];


		value = (INB(0x102) & 0xff);

		/* Wild card support: first board found is matched */
		if (*paddress == 0) 
		{
			*paddress = setup_address[i];
			*pirq = (value & 2) ?
				mcInterrupts[(value >> 5) & 7] : 0;
		}

		if (*paddress == setup_address[i])
		{
			goodness = 1;
			slotMorE = i | 8;
		}

		value = (INB(0x102) & 0xff);
		setup_irq[i] = (value & 2) ?
			mcInterrupts[(value >> 5) & 7] : 0;

		if (setup_irq[i] == *pirq || !*pirq)
		{
			goodness |= 2;
		}
	}

	OUTB(0x96, setup_save);
	RESTORE_INTS

	if ((mainVerbosity & V_STARTUP) || goodness != 3)
	{
		for (i = 0; i < 7; i++)
		{
			if (ismine[i])
			{
				printf("Computone MC board in slot %d, I.D. 0x%x, Address 0x%x",
					i + 1, id[i], setup_address[i]);
				if (setup_irq[i]) 
				{
					printf(", Interrupt %d\n", setup_irq[i]);
				}
				else
				{
					printf(", (polled)\n");
				}
			}
		}
	}
	if (goodness != 3)
	{
		printf("Micro Channel POS: No board configured for ");
		if (*paddress)
		{
			printf("address 0x%x", *paddress);
		}
		else
		{
			printf("any address");
		}
		if (*pirq)
		{
			printf(", interrupt %d\n", *pirq);
		}
		else
		{
			printf(".\n");
		}
	}
	return (goodness == 3);
}


/*--------------------------------------------
 * findEISAslot()
 * Returns true or false according to whether
 * we have found a board for the selected address
 * in any EISA slot()
 * Stores the slot in slotMorE global
 * 
 * If argument irq == 0, then we don't care what
 * irq we were configured for.
 *--------------------------------------------
 */

int
findEISAslot(int *paddress, int *pirq)
{
register int i;
int goodness = 0;	/* bad until proved good */
unsigned int idm[16];
unsigned int idp[16];
unsigned int base;
unsigned int value;
int setup_address[16];
int setup_irq[16];
int ismine[16];

	for (i = 1; i < 16; i++)
	{
		base = i << 12;
		idm[i] = (INB(base + 0xc80) << 8) | (INB(base + 0xc81) & 0xff);
		idp[i] = (INB(base + 0xc82) << 8) | (INB(base + 0xc83) & 0xff);
		ismine[i] = 0;
		if (idm[i] == 0x0e8e)
		{
			if(idp[i] == 0x0281 || idp[i] == 0x0218)
			{
				ismine[i] = 1;
			}
			else if(idp[i] == 0x0282 || idp[i] == 0x0283)
			{
				ismine[i] = 3;	/* Can do edge-trigger */
			}
			else 
			{
				continue;
			}
		}
		else 
		{
			continue;
		}


		/* It's some sort of eisa card, but 
		 * at what address is it configured?
		 */

		setup_address[i] = base + 0xc88;
		value = INB(base + 0xc86);

		if (*paddress == 0)
		{
			*paddress = setup_address[i];
			if (*pirq == 0)
			{
				*pirq = (value & 8) ?
					mcInterrupts[value & 7] : 0;
			}
		}
		if (*paddress == setup_address[i])
		{
			goodness = 1;
			slotMorE = base;
		}

		setup_irq[i] = (value & 8) ?
			mcInterrupts[value & 7] : 0;

		if ( (ismine[i] & 2) && !(value & 0x10))
		{
			ismine[i] = 1;	/* Could be edging, but not */
		}

		if (setup_irq[i] == *pirq || !*pirq)
		{
			goodness |= 2;
		}
	}


	if ((mainVerbosity & V_STARTUP) || goodness != 3)
	{
		for (i = 1; i < 16; i++)
		{
			if (ismine[i])
			{
				printf("Computone EISA board in slot %d, I.D. 0x%x%x, Address 0x%x",
					i, idm[i], idp[i], setup_address[i]);
				if (setup_irq[i]) 
				{
					printf(", Interrupt %d %s\n", setup_irq[i],
						(ismine[i] & 2) ? "(edge)" : "(level)");
				}
				else
				{
					printf(", (polled)\n");
				}
			}
		}
	}
	if (goodness != 3)
	{
		printf("EISA Configuration: No board at ");
		if (*paddress)
		{
			printf("address 0x%x", *paddress);
		}
		else
		{
			printf("any address");
		}
		if (*pirq)
		{
			printf(", interrupt %d\n", *pirq);
		}
		else
		{
			printf(".\n");
		}
	}
	return (goodness == 3);
}


/*--------------------------------------
 * pos_checkout(f)
 * 
 * Checks out pos[2] and pos[3] registers of
 * a Micro Channel card.
 *
 * Input parameters:
 * 		f_passes:	number of passes (0 --> 65536)
 * Output:
 * 		f_error:	OKAY:			no error
 *
 *		f_error:	BAD_POS_2		pos register 2 failed
 *		f_error:	BAD_POS_3		pos register 3 failed
 *		f_wstring[0]	data written
 *		f_rstring[0]	data read back
 *		f_error:	BAD_MC			not seemingly on a microchannel
 *-------------------------------------------------------------
 */
			
int
pos_checkout (pfunk f)
{
int setup_save;
int savePOS;
int i, j;
int passes;

	/* Note that this test has the interrupts off almost the
	 * whole time: implication is that ctrl-C cannot break us
	 * out.
	 */

	if (!hasUsedB || (busType != MCA_OPT) || !slotMorE)
	{
		POST_ERROR(f, BAD_MC);
	}
	
	SAVE_AND_DISABLE_INTS
	setup_save = INB(0x96);
	OUTB(0x96, slotMorE);	/* select the slot */

	passes = f->f_passes;
	f->f_passfail = 0;	//spin through the passes
	do {
		/* All the bits of 103 should be pretty much a
		 * free choice.  It selects address.
		 */
		savePOS = INB(0x103);	/* save pos103 contents */
		for (i = 0; i < 256; i++)
		{
			OUTB(0x103, i);
			j = INB(0x103);
			if (i != j)
			{
				f->f_wstring[0] = i;
				f->f_rstring[0] = j;
				OUTB(0x103, savePOS);
				OUTB(0x96, setup_save);
				RESTORE_INTS
				POST_ERROR(f, BAD_POS_3);
			}
		}
		OUTB(0x103, savePOS);
	
		/* Don't change all the bits here, some are reserved.
		 * The ones we do change select the interrupt.
		 */
		savePOS = INB(0x102);	/* save pos102 contents */
		for (i = 3; i < 256; i += 0x20)
		{
			OUTB(0x102, i);
			j = INB(0x102);
			if (i != j)
			{
				f->f_wstring[0] = i;
				f->f_rstring[0] = j;
				OUTB(0x102, savePOS);
				OUTB(0x96, setup_save);
				RESTORE_INTS
				POST_ERROR(f, BAD_POS_2);
			}
		}
		OUTB(0x102, savePOS);
		f->f_passfail++;
	} while (--passes);
	
	OUTB(0x96, setup_save);
	RESTORE_INTS
	SUCCEED(f);
}


/*-------------------------------------------------------------
 * memory_walk(f)
 *
 * Does a walking bit test on memory -- utilizes downloaded code
 * Returns 0 if ok, -1 if failure of any sort
 *
 * memory_rand(f)
 * 
 * Does a random test on memory -- utilizes downloaded code
 * Returns 0 if ok, -1 if failure of any sort
 *
 * 
 * Input parameters:
 * 		f_pB:		Pointer to board structure
 * 		f_passes:	number of passes (0 --> 65536)
 * Output:
 * 		f_error:	OKAY:			no error
 * 		f_results[0]:	memory found, in 1K blocks
 * MW_BAD_MEM_SIZE:	invalid memory size
 * MR_BAD_MEM_SIZE:	invalid memory size
 * 		f_results[0]:	memory found, in 1K blocks
 * MW_MEM_FAIL:	test failed
 * MR_MEM_FAIL:	test failed
 * 		f_passfail:		pass number failed
 * 		f_results[0]:	memory found, in 1K blocks
 * 		f_results[1]:	segment of failure
 * 		f_results[2]:	offset of failure
 * 		f_wstring[0]:	value written (low 8 bits)
 * 		f_wstring[1]:	value written (high 8 bits)
 * 		f_rstring[0]:	value read back (low 8 bits)
 * 		f_rstring[1]:	value read back (high 8 bits)
 * MW_TIMEOUT1:	timed out starting the test up
 * MR_TIMEOUT1:	timed out starting the test up
 * 		no other return information
 * MW_TIMEOUT2:	timed out waiting for a pass to complete
 * MR_TIMEOUT2:	timed out waiting for a pass to complete
 * 		f_passfail:		pass number failed
 * 
 *-------------------------------------------------------------
 */

int 
memory_walk (pfunk f)
{
	return memory_diag (f, 0);	//the walking bit test
}

int 
memory_rand (f)
funk *f;
{
	return memory_diag (f, 1);	//the random memory test
}



static int memory_diag (pfunk f, int isRandom)
{
int i;
ushort passes;
i2eBordStrPtr pB = f->f_pB;

	iiDelay(pB, 10);		/* ? Delay after downloading the firmware ? */
							/* ? is it really needed ? */

	/* Tell it how many passes to do */
	if (word_data_out (pB, 1, f->f_passes))	/* Should be immediately ready */
	{
		MEM_ERROR(f, isRandom, MR_TIMEOUT1, MW_TIMEOUT1);
	}

	i = byte_data_in(pB, 2);
	if (i < 0)
	{
		MEM_ERROR(f, isRandom, MR_TIMEOUT1, MW_TIMEOUT1);
	}

	i <<= 5;			/* Convert to K */

	f->f_results[0] = i;

	/* Check memory report for a reasonable value, and that
	 * it matches the value given in power-up message
	 */

	if ( (i & 127) || (i < 128) || (i > 2048) ||
		(i != (pB->i2ePom.e.porMemory << 5)) )
	{
		MEM_ERROR(f, isRandom, MR_BAD_MEM_SIZE, MW_BAD_MEM_SIZE);
	}

	passes = f->f_passes;
	f->f_passfail = 0;	//spin through the passes
	do {
		i = byte_data_in (pB, (isRandom ? 6 : 32) );

		if (i < 0)
		{
			MEM_ERROR(f, isRandom, MR_TIMEOUT2, MW_TIMEOUT2);
		}
		if (i)	//failure
		{
			if	(word_data_read (pB, 2, &(f->f_results[1])) ||
				 word_data_read (pB, 2, &(f->f_results[2])) ||
				 word_data_read (pB, 2, (ushort *)&(f->f_wstring[0])) ||
				 word_data_read (pB, 2, (ushort *)&(f->f_wstring[2]))
			)
			{
				MEM_ERROR(f, isRandom, MR_TIMEOUT2, MW_TIMEOUT2);
			}
			MEM_ERROR(f, isRandom, MR_MEM_FAIL, MW_MEM_FAIL);
		}
		f->f_passfail++;
	} while (--passes);
	SUCCEED(f);	
}



/*------------------------------------------------------
 * fifo_sanity(f)
 
 * Does a basic sanity check on the bi-fifo -- involves no downloading
 * Returns 0 if ok, -1 if failure of any sort
 * 
 * ONLY APPLIES TO NON-EXPANDABLE BOARDS
 * Input parameters:
 *  f_pB:		pointer to board structure
 * Output:
 * 	f_error:	OKAY:			no error
 * 				f_results[0]:	type of board:
 * 								NON_EXPANDABLE
 *
 * 			FS_AF_AE:		initial AF/AE register error
 * 				f_results[0]:	type of board:
 * 								NON_EXPANDABLE
 * 				f_results[1]:	# of ports
 * 				f_wstring[0]:	expected value
 * 				f_rstring[0]:	value read
 * 			FS_MASK:		initial mask register error
 * 				f_results[0]:	type of board:
 * 								NON_EXPANDABLE
 * 				f_results[1]:	# of ports
 * 				f_wstring[0]:	expected value
 * 				f_rstring[0]:	value read
 * 			FS_AF_AE_D:		AF/AE register diagnostic failure
 * 				f_results[0]:	type of board:
 * 								NON_EXPANDABLE
 * 				f_results[1]:	# of ports
 * 				f_wstring[0]:	value written
 * 				f_rstring[0]:	value read
 * 			FS_MASK_D:		mask register diagnostic failure
 * 				f_results[0]:	type of board:
 * 								NON_EXPANDABLE
 * 				f_results[1]:	# of ports
 * 				f_wstring[0]:	value written
 * 				f_rstring[0]:	value read
 * 			FS_BYTE_D:		byte detect register diagnostic failure
 * 				f_results[0]:	type of board:
 * 								NON_EXPANDABLE
 * 				f_results[1]:	# of ports
 * 				f_wstring[0]:	value written
 * 				f_rstring[0]:	value read
 * 			FS_MAILBOX_D:	outbound mailbox register diagnostic failure
 * 				f_results[0]:	type of board:
 * 								NON_EXPANDABLE
 * 				f_results[1]:	# of ports
 * 				f_wstring[0]:	value written
 * 				f_rstring[0]:	value read
 *----------------------------------------------------
 */

int fifo_sanity (pfunk f)
{
i2eBordStrPtr pB = f->f_pB;
int i;
char kar;
int port;
uchar *dw;
uchar *dr;

	port = pB->i2ePointer;
	
	/* AE/AF register should be 0 after reset and reading init. message */
	OUTB(port, SEL_AEAF);
	if ((kar = INB(port)) != 0)
	{
		f->f_wstring[0] = 0;
		f->f_rstring[0] = kar;
		POST_ERROR(f, FS_AF_AE);
	}

	/* Initial Am4701 Mask after reset has ...*/ 
	OUTB(port, SEL_MASK);
	if ((kar = INB(port)) != (ST_OUT_FULL | ST_IN_EMPTY))
	{
		f->f_wstring[0] = (ST_OUT_FULL | ST_IN_EMPTY);
		f->f_rstring[0] = kar;
		POST_ERROR(f, FS_MASK);
	}
	dw = f->f_wstring;
	dr = f->f_rstring;

	/* Test AE/AF register */
	i = fifo_4701_test(port, SEL_AEAF, dw, dr);
	if (i)
	{
		POST_ERROR(f, FS_AF_AE_D);
	}

	/* test the mask register */
	i = fifo_4701_test(port, SEL_MASK, dw, dr);
	if (i)
	{
		POST_ERROR(f, FS_MASK_D);
	}

	/* test the byte-detect register */
	i = fifo_4701_test(port, SEL_BYTE_DET, dw, dr);
	if (i)
	{
		POST_ERROR(f, FS_BYTE_D);
	}

	/* test outgoing mailbox register */
	i = fifo_4701_test(port, SEL_OUTMAIL, dw, dr);
	if (i)
	{
		POST_ERROR(f, FS_MAILBOX_D);
	}
	SUCCEED(f);
}



static int fifo_4701_test (port, select, pwr, prd)
int port;	/* address of the pointer register */
int select;	/* Which register */
uchar *pwr;	/* Store what we wrote */
uchar *prd;	/* Store what we read */
{
int i;
int j;

uchar kar;
uchar kar2;
uchar validBits;

	kar = 19;
	/* Mask register has only 7 bits */
	j = (select == SEL_MASK ? 128 : 256);
	validBits = j - 1;

	for (i=0; i<j; i++)
	{
		OUTB(port, select);	/* Write the value */
		OUTB(port, kar);

		OUTB(port, select);	/* Read the value and compare */
		kar2 = INB(port);
		if (kar != kar2)
		{
			*pwr = kar;
			*prd = kar2;
			return -1;
		}
		kar = (kar + 19) & validBits;
	}
	return 0;
}

/*-----------------------------------------------------------------
 * int fifo_local_stress (pfunk f)
 *
 * Invokes either fifo_local_stress8 or fifo_local_stress16
 * depending on the fifo size used.
 * See those routines (below) for return values, etc.
 *-----------------------------------------------------------------
 */

int
fifo_local_stress(pfunk f)
{
	return (f->f_pB->i2eDataWidth16 ? 
		fifo_local_stress16(f) : fifo_local_stress8(f) );
}


/*-----------------------------------------------------------------
 * int fifo_local_stress8 (pfunk f)
 *
 * Does a "local" loopback check on the bi-fifo -- utilizes downloaded code
 * (local loopback is having the board loop back data sent by the host)
 * Returns 0 if ok, -1 if failure of any sort
 *
 * This flavour assumes 8-bit interface!
 *
 * Input parameters:
 *	f_pB:		pointer to board structure.
 *	f_passes:	number of passes (0 --> 65536)
 *
 * Output:
 *	f_error:	OKAY:			no error
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *				FLL_BAD_SIZE:	bad fifo size
 *					f_results[0]:	size of fifo found
 *				FLL_TIMEOUT2:	timed-out looping data back
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *				FLL_DATA_ERROR:	data error in loopback
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *					f_wstring[0]:	value written
 *					f_rstring[0]:	value read
 *					f_results[1]:	0
 *					f_results[2]:	0
 *					f_results[3]:	0
 *				FLL_GENERAL:	some goofy kind of error
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *					f_results[1]:	error code
 *-----------------------------------------------------------------
 */

static int
fifo_local_stress8 (pfunk f)
{
int i;
uchar kar, ksave;
int k2;
ushort passes;
i2eBordStrPtr pB = f->f_pB;

	f->f_results[0] = pB->i2eFifoSize;
	f->f_results[1] = f->f_results[2] = f->f_results[3] = 0;
	passes = f->f_passes;
	f->f_passfail = 0;
	kar =  (time(NULL) & 0xff);	/* pick up a random value here */
	do {
		/* First part of a pass, try one by each! */
		for (i=0; i<256; i++)
		{
			if (byte_data_out (pB, 2, kar))
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			k2 = byte_data_in (pB, 2);
			if (k2 < 0)	//timeout
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			if ((k2 & 0xff) != kar)
			{
				f->f_wstring[0] = kar;
				f->f_rstring[0] = k2;
				POST_ERROR(f, FLL_DATA_ERROR);
			}
			kar += passes;
		}
		kar++;
		ksave = kar;

		/* Next, write some, read some */

		for (i=0; i < 256/8; i++)
		{
			if (byte_data_out8e (pB, 2, kar))
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			kar += passes;
		}
		iiDelay(pB, 2);
		iiDelay(pB, 1);	/* Allow time for the firmware to absolutely
						 * finish transferring the data.
						 * The loadware should be able to handle
						 * this well within a millisecond: we'll
						 * actually insert two separate 1-Ms delays
						 * here for safety, and this does not allow
						 * for the further advantage that the
						 * firmware started copying while we were
						 * still writing data.
						 */
	
		kar = ksave;
		for (i = 0; i < 256 / 8; i++)
		{
			k2 = byte_data_in4i (pB, kar);
			if (k2 < 0)	//timeout
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			if (k2  != kar)
			{
				f->f_wstring[0] = kar;
				f->f_rstring[0] = k2;
				POST_ERROR(f, FLL_DATA_ERROR);
			}

			kar = ~kar;
			k2 = byte_data_in4i (pB, kar);
			if (k2 < 0)	//timeout
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			if (k2 != kar)
			{
				f->f_wstring[0] = kar;
				f->f_rstring[0] = k2;
				POST_ERROR(f, FLL_DATA_ERROR);
			}
			kar = ~kar;

			kar += passes;
		}

		f->f_passfail++;
		kbhit();		/* This ensures ctrlC detection */

	} while (--passes);
	SUCCEED(f);
}

/*-----------------------------------------------------------------
 * int fifo_local_stress16 (pfunk f)
 *
 * Does a "local" loopback check on the bi-fifo -- utilizes downloaded code
 * (local loopback is having the board loop back data sent by the host)
 * Returns 0 if ok, -1 if failure of any sort
 *
 * This flavour assumes 16-bit interface!
 *
 * Input parameters:
 *	f_pB:		pointer to board structure.
 *	f_passes:	number of passes (0 --> 65536)
 *
 * Output:
 *	f_error:	OKAY:			no error
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *				FLL_BAD_SIZE:	bad fifo size
 *					f_results[0]:	size of fifo found
 *				FLL_TIMEOUT2:	timed-out looping data back
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *				FLL_DATA_ERROR16:		data error in loopback
 *				FLL_DATA_ERROR16_EZ:	data error in loopback (single access)
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *					f_wstring[0]:	value written (lsb)
 *					f_wstring[1]:	value written (msb)
 *					f_rstring[0]:	value read	(lsb)
 *					f_rstring[1]:	value read	(msb)
 *					f_results[1]:	0
 *					f_results[2]:	0
 *				FLL_GENERAL:	some goofy kind of error
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *					f_results[1]:	error code
 *-----------------------------------------------------------------
 */

static int
fifo_local_stress16 (pfunk f)
{
int i;
ushort passes;
i2eBordStrPtr pB = f->f_pB;
ushort ivalue, ovalue, osave;

	f->f_results[0] = pB->i2eFifoSize;
	f->f_results[1] = f->f_results[2] = 0;
	passes = f->f_passes;
	f->f_passfail = 0;
	ovalue =  (time(NULL) & 0xffff);	/* pick up a random value here */
	do {
		/* First part of a pass, try one by each! */
		for (i=0; i<256; i++)
		{
			if (word16_data_out (pB, 2, ovalue))
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			if (word16_data_read (pB, 2, &ivalue) < 0)	/* Timeout? */
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			if (ivalue != ovalue)
			{
				f->f_wstring[0] = ovalue & 0xff;
				f->f_wstring[1] = ovalue >> 8;
				f->f_rstring[0] = ivalue & 0xff;
				f->f_rstring[1] = ivalue >> 8;
				POST_ERROR(f, FLL_DATA_ERROR16_EZ);
			}
			ovalue += 0x2317 + passes;
		}
		ovalue++;
		osave = ovalue;

		/* Next, write some, read some */

		for (i=0; i < 256/8; i++)
		{
			if (word_data_out8e (pB, 2, ovalue))
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			ovalue += passes;
		}
	
		iiDelay(pB, 2);
		iiDelay(pB, 1);	/* Allow time: see notes above */

		ovalue = osave;
		for (i = 0; i < 256 / 8; i++)
		{
			if (word_data_read4i (pB, ovalue, &ivalue) < 0)
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			if (ivalue != ovalue)
			{
				f->f_wstring[0] = ovalue & 0xff;
				f->f_wstring[1] = ovalue >> 8;
				f->f_rstring[0] = ivalue & 0xff;
				f->f_rstring[1] = ivalue >> 8;
				POST_ERROR(f, FLL_DATA_ERROR16);
			}
			ovalue = ~ovalue;
			if (word_data_read4i (pB, ovalue, &ivalue) < 0)
			{
				POST_ERROR(f, FLL_TIMEOUT2);
			}
			if (ivalue != ovalue)
			{
				f->f_wstring[0] = ovalue & 0xff;
				f->f_wstring[1] = ovalue >> 8;
				f->f_rstring[0] = ivalue & 0xff;
				f->f_rstring[1] = ivalue >> 8;
				POST_ERROR(f, FLL_DATA_ERROR16);
			}
			ovalue = ~ovalue;

			ovalue += passes;
		}

		f->f_passfail++;
		kbhit();		/* This ensures ctrlC detection */

	} while (--passes);
	SUCCEED(f);
}

/*-----------------------------------------------
 * int fifo_isolate_loop(f)
 *
 * Fifo-loopback test in which both the board and
 * the host verify the integrity of the data.
 *
 * Uses native mode transfers (byte or word depending on 
 * product and configuration)
 *
 * Input parameters:
 *	f_pB:		pointer to board structure
 *	f_passes:	number of passes (0 --> 65536)
 *
 * Output:
 *	f_error:	OKAY:			no error
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *
 *				FLL_DATA_ERROR:	data error in loopback
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *					f_wstring[0]:	value written
 *					f_rstring[0]:	value read
 *					f_results[1]:	1 = low-byte, 2 = high-byte, 3 = both
 *									(if using 16-bit transfers)
 *					f_results[2]:	1 = error writing to board
 *									2 = error reading from board.
 *					f_results[3]:	1 (means report a while word)
 *				FLL_GENERAL:	some goofy kind of error
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *					f_results[1]:	error code
 *------------------------------------------------------------------------
 */

int
fifo_isolate_loop(pfunk f)
{
int i;
unsigned short ovalue, osave;
unsigned short ivalue;
int mail;
ushort passes;
i2eBordStrPtr pB = f->f_pB;

	f->f_results[0] = pB->i2eFifoSize;
	f->f_results[1] = f->f_results[2] = 0;
	f->f_results[3] = 1;
	passes = f->f_passes;
	f->f_passfail = 0;
	ovalue = 0xface;		/* This must match what the loadware expects */

	do {
		/* Write a block, saving where we started */
		osave = ovalue;
		for (i = 0; i < 256; i++)
		{
			/* 2.05.0 per AKM: 1 may be too soon */
			if (word_data_out (pB, 2, ovalue))
			{
				f->f_results[1] = 0x8110;
				POST_ERROR(f, FLL_GENERAL);
			}
			ovalue += 0x2211;		/* This must also match firmware */
		}

		/* Read it back, comparing values */
		ovalue = osave;
		for (i = 0; i < 256; i++)
		{
			/* 2.05.0 per AKM: 1 may be too soon */
			if (word_data_read(pB, 2, &ivalue))
			{
				f->f_results[1] = 0x8120;
				POST_ERROR(f, FLL_GENERAL);
			}
			/* Check for an error */
			if (ivalue != ovalue)
			{
				/* For whatever type of error, we need the values */
				f->f_wstring[0] = ovalue & 0xff;
				f->f_wstring[1] = ovalue >> 8;
				f->f_rstring[0] = ivalue & 0xff;
				f->f_rstring[1] = ivalue >> 8;

				/* If there is something in the mailbox soon, then
				 * this is an error writing to the board.
				 * 10h - low byte error, 20h - high byte error,
				 * 30h - both
				 */

				if ((mail = wait_mailbox_in (pB, 2)) >= 0)
				{
					f->f_results[1] = mail >> 4;	/* high, low, whatever */
					f->f_results[2] = 1;	/* error writing to board */
				}
				else
				{
					f->f_results[2] = 2;	/* Error reading from board */
					if (f->f_wstring[0] == f->f_rstring[0])
					{
						f->f_results[1] = 2;	/* error in high byte */
					}
					else if (f->f_wstring[1] == f->f_rstring[1])
					{
						f->f_results[1] = 1;	/* error in low byte */
					}
					else 
					{
						f->f_results[1] = 3;	/* error in both bytes */
					}
				}
				POST_ERROR(f, FLL_DATA_ERROR);
			}
			ovalue += 0x2211;		/* This must also match firmware */
		}
		f->f_passfail++;
		kbhit();		/* This ensures ctrlC detection */
	}
	while (--passes);
	SUCCEED(f);
}

/*-----------------------------------------------
 * int fifo_dma_loopback(f)
 *
 * Uses the FIFO in native mode, and can only be used
 * for the IntelliPort-IIEX, since the other does not
 * have DMA.
 *
 * The loadware waits for a mailbox hit,
 * Then sets up DMA to receive 16k words or bytes
 * Then sends 16k words or bytes to the host.
 * (The host loops back the data as he gets it.  Because of DMA
 * priorities host can assume that if there is data available loop back,
 * there must be room in the fifo to loop it back to).
 * When 16k words or bytes have been looped back, we wait for a mailbox
 * message from the board, which indicates the status of the pass.
 *
 *
 * Input parameters:
 *	f_pB:		pointer to board structure
 *	f_passes:	number of passes (0 --> 65536)
 *
 * Output:
 *	f_error:	OKAY:			no error
 *
 *				FDMA_XTIMO:		board timed out trying to send data to host
 *				FDMA_RTIMO:		board timed out trying to read data from host
 *				FDMA_MISM:		board reports mis-matched characters
 *				FDMA_HDATA:		host doesn't get data as soon as expected
 *				FDMA_HMAIL:		host doesn't get end-of-pass mailbox message
 *								or couldn't send start-of-pass message
 *				
 *------------------------------------------------------------------------
 */

int
fifo_dma_loopback(pfunk f)
{
ushort passes;
register i2eBordStrPtr pB = f->f_pB;
register int count;
int data_address;

	passes = f->f_passes;
	f->f_passfail = 0;

	data_address = pB->i2eData;

	iiDelay(pB, 20);		/* Wait plenty of time for loadware to be
							 * started
							 */
	do {
		/* Write to board mailbox to start things off */
		if (!iiTrySendMail(pB, 0x30))	/* value doesn't matter */
		{
			POST_ERROR(f, FDMA_HMAIL);
		}

		/* Wait up to 100 MS for the first data byte to appear:
		 * once the mail has been picked up, the board must set
		 * up the DMA transfers...
		 */

		for (count = 0; count < 10; count++)
		{
			if (HAS_INPUT(pB)) break;
			iiDelay(pB, 10);
		}
		/* Once the DMA has been turned on, we expect that the
		 * the board ought to stay ahead, since the host accesses
		 * to the board are limited by bus speed even on very fast
		 * hosts.
		 */

		/* Write back any data received, up to 16k things */
		/* The status read should guarantee that the board will
		 * stay ahead of the host!.
		 */

		/* An additional delay first, to ensure the board will
		 * run ahead.  This also guarantees that the board shall
		 * have filled his output fifo (not overfilled) before
		 * we shall have read the data.  If the fifo overfills,
		 * causing data to be lost, it would mean the fifo's full
		 * flag ain't stopping the DMA.  This delay gives such a
		 * failure a chance to happen.
		 */

		iiDelay(pB, 100);

		if(pB->i2eDataWidth16)
		{
			for (count = 0; count < 16 * 1024; count++)
			{
				if (HAS_INPUT(pB))
				{
					OUTW(data_address, INW(data_address));
				}
				else
				{
					if (mainVerbosity & V_FIFO)
					{
						printf("\n16-bit mode, waiting for input at count %d\n",
							count);
					}
					POST_ERROR(f, FDMA_HDATA);
				}
			}
		}
		else
		{
			for (count = 0; count < 16 * 1024; count++)
			{
				if (HAS_INPUT(pB))
				{
					OUTB(data_address, INB(data_address)); 
				}
				else
				{
					if (mainVerbosity & V_FIFO)
					{
						printf("\n8-bit mode, waiting for input at count %d\n",
							count);
					}
					POST_ERROR(f, FDMA_HDATA);
				}
			}
		}

		/* Clear two lowest bits of mailbox before testing */
		switch(wait_mailbox_in(pB, 2) & (~3))
		{
		case 0:
			break;
		case 0x10:
			POST_ERROR(f, FDMA_XTIMO);
		case 0x20:
			POST_ERROR(f, FDMA_RTIMO);
		case 0x30:
			POST_ERROR(f, FDMA_MISM);

			/* The default covers unexpected values as well as
			 * time-out getting the mail.
			 */
		default:
			if (mainVerbosity & V_FIFO)
			{
				printf("\nBoard is late with mail: status = %x\n",
					INB(pB->i2eStatus) );
			}
			POST_ERROR(f, FDMA_HMAIL);
		}
		f->f_passfail++;
		kbhit();		/* This ensures ctrlC detection */
	}
	while (--passes);
	SUCCEED(f);
}


/*-----------------------------------------------------------------
 * int fifo_verify_size(f)
 *
 * Determines the number of bytes in the fifo by writing words (or
 * pairs of bytes) to it.
 * 
 * Compares it with the number reported in Power-on test.
 *
 * Input parameters:
 *	f_pB:		pointer to board structure.
 *	f_passes:	number of passes (0 --> 65536)
 *
 * Output:
 *	f_results[0]:	size reported by power-on message (anyway)
 *
 *	f_error:	OKAY:			no error
 *				FV_NODRAIN:
 *					Cannot start the test: output fifo won't drain
 *				FV_BAD_SIZE:	bad fifo size
 *					f_results[1]:	size measured
 *---------------------------------------------------------
 */
int fifo_verify_size(pfunk f)
{
i2eBordStrPtr pB = f->f_pB;
unsigned int capacity;
unsigned short us;

	f->f_results[0] = pB->i2eFifoSize;
	if (!iiWaitForTxEmpty(pB, 10))
	{
		POST_ERROR(f, FV_NODRAIN);
	}	

	/* For complete robustness, count the number of
	 * words we could write, one by each.
	 */

	/* For this test, we cannot tolerate a fifo larger
	 * than 16k words, or 32k bytes
	 */ 
	
	capacity = us = 0;
	while(HAS_OUTROOM(pB))
	{
		iiWriteWord(pB, us++);		/* This will write 1 word, or 2 bytes */
		if (++capacity >= 32767) break;
	}
	capacity <<= 1;		/* Change from word count to byte count */
	f->f_results[1] = capacity;
	if (f->f_results[0] != capacity)
	{
		POST_ERROR(f, FV_BAD_SIZE);
	}
	SUCCEED(f);
}

/*--------------------------------------------------------------------
 * int fifo_block_loopback (f)
 *
 * Does a "local" loopback check on the bi-fifo -- utilizes downloaded code
 * (local loopback is having the board loop back data sent by the host)
 * Uses a block-oriented protocol:  copies 256 bytes, hits mailbox,
 * waits for mailbox hit, reads 256 bytes, compares...
 * Returns 0 if ok, -1 if failure of any sort
 *
 * Uses native mode transfers (byte or word depending on 
 * product and configuration)
 *
 * Input parameters:
 *	f_pB:		pointer to board structure
 *	f_passes:	number of passes (0 --> 65536)
 *
 * Output:
 *	f_error:	OKAY:			no error
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *
 *				FLL_DATA_ERROR:	data error in loopback
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *					f_wstring[0]:	value written
 *					f_rstring[0]:	value read
 *					f_results[1]:	1 = low-byte, 2 = high-byte, 3 = both
 *									(if using 16-bit transfers)
 *					f_results[2]:	0 
 *					f_results[3]:	0
 *				FLL_GENERAL:	some goofy kind of error
 *					f_passfail:		pass number failed
 *					f_results[0]:	size of fifo, in bytes (each direction)
 *					f_results[1]:	error code
 *------------------------------------------------------------------------
 */
extern int forceStringLength;
extern int argInt;

int fifo_block_loopback (pfunk f)
{
ushort passes;
uchar inbuffer1[256];
uchar outbuffer1[256];
uchar inbuffer2[256];
uchar outbuffer2[256];
uchar *puW, *puR, *puWlimit;
uchar startvalue;
uchar adder = 0;
i2eBordStrPtr pB = f->f_pB;
int packsize1 = 124;
int packsize2 = 120;
unsigned char myMask = 0;
int length;

	f->f_results[0] = pB->i2eFifoSize;
	f->f_results[1] = f->f_results[2] = f->f_results[3] = 0;
	passes = f->f_passes;
	f->f_passfail = 0;

	startvalue =  (time(NULL) & 0xff);	/* pick up a random value here */

	/* Initialize value in mask register */
	/* Make sure MSB is not set: has special meaning for expandables */
	iiWriteMask(pB, (myMask & 0x7f));

	do {
		/* Change the value in mask register in case we are sensitive to it */
		myMask += 0x71;
		iiWriteMask(pB, (myMask & 0x7f));

		if (forceStringLength)
		{
			for (puW = outbuffer1, puR = outbuffer2; puW < &outbuffer1[256]; )
			{
				*puW++ = argInt & 0xff;
				*puW++ = argInt >> 8;
				*puR++ = argInt & 0xff;
				*puR++ = argInt >> 8;
			}
		}
		else
		{
			startvalue += adder;
			adder += 0x33;

			/* Generate "random" data */
			for (puW = outbuffer1, puR = outbuffer2; puW < &outbuffer1[256]; )
			{
				/* The first two guarantee great "bit-delta", but
			 	* the results are, overall balanced: as many bits
			 	* set as clear.
			 	*/
	
				*puW++ = startvalue;
				*puW++ = ~startvalue;
	
				/* These results will be more "unbalanced" */
	
				*puR++ = 0x55 ^ adder;
				*puR++ = 0xAA ^ startvalue;
	
				startvalue += adder;
			}
		}
 
		/* String-move the data from first buffer.*/

		iiWriteWord(pB, packsize1);
		iiWriteBuf(pB, outbuffer1, packsize1);
			
		/* Hit mailbox to notify board data is there */

		iiDelay(pB,5);	/* Delay, to allow data in fifo to corrupt itself */

		if(!iiTrySendMail(pB, 1))
		{
			POST_ERROR(f, FM_TIMEOUT_OUT);
		}

		/* Wait for incoming mailbox: data has been read */
		/* Must take less than two seconds for sure! */

		if (wait_mailbox_in (pB, 2) < 0)
		{
			POST_ERROR(f, FM_TIMEOUT_IN);
		}

		/* String-move the data from second buffer.*/

		iiWriteWord(pB, packsize2);
		iiWriteBuf(pB, outbuffer2, packsize2);

		/* Hit mailbox to notify board data is there */

		iiDelay(pB, 5);	/* delay to allow data in fifo to corrupt itself */

		if(!iiTrySendMail(pB, 1))
		{
			POST_ERROR(f, FM_TIMEOUT_OUT);
		}

		/* Wait for incoming mailbox: first block data has been written */
		/* Must take less than two seconds for sure! */

		if (wait_mailbox_in (pB, 2) < 0)
		{
			POST_ERROR(f, FM_TIMEOUT_IN);
		}

		/* Read in the data from block 1 */

		length = iiReadWord(pB) & 0xff;
		iiReadBuf(pB, inbuffer1, length);

		if (length != packsize1 && (mainVerbosity & V_FIFO) )
		{
			printf("\n     Length Mismatch (block 1): wrote %d, read %d\n",
				packsize1, length);
		}

		for (puW = outbuffer1, puR = inbuffer1, 
			puWlimit = &outbuffer1[packsize1];
				puW < puWlimit; puW++, puR++)
		{
			if (*puW != *puR)
			{
				if (mainVerbosity & V_FIFO)
				{
				int ee;
				int mm;
					mm = puW - outbuffer1;
					mm = (mm < 4) ? 0 : mm-4;
					ee = packsize1 - mm;
					ee = (ee > 12) ? 12 : ee;
					printf("     Starting %d into buffer 1, values written|read\n     ", mm);
					for (; ee; ee--, mm++)
					{
						printf("%x|%x ", outbuffer1[mm], inbuffer1[mm]);
					}
					printf("\n");
				}
					
				f->f_wstring[0] = *puW;
				f->f_rstring[0] = *puR;
				if (pB->i2eDataWidth16)
				{
					if ((puW - outbuffer1) & 1)
					{
						f->f_results[1] = 2;	/* High-byte */
					}
					else if (puW[1] != puR[1])
					{
						f->f_wstring[1] = puW[1];
						f->f_rstring[1] = puR[1];
						f->f_results[1] = 3;	/* Both bytes */
					}
					else
					{
						f->f_results[1] = 1;	/* Low byte */
					}
				}
				POST_ERROR(f, FLL_DATA_ERROR);
			}
		}

		/* Wait for incoming mailbox: second block data has been written */
		/* Must take less than two seconds for sure! */

		if (wait_mailbox_in (pB, 2) < 0)
		{
			POST_ERROR(f, FM_TIMEOUT_IN);
		}

		/* Read in the data from block 2 */

		length = iiReadWord(pB) & 0xff;
		iiReadBuf(pB, inbuffer2, length);

		if (length != packsize2 && (mainVerbosity & V_FIFO) )
		{
			printf("\n     Length Mismatch (block 2): wrote %d, read %d\n",
				packsize1, length);
		}

		for (puW = outbuffer2, puR = inbuffer2, 
			puWlimit = &outbuffer2[packsize2];
				puW < puWlimit; puW++, puR++)
		{
			if (*puW != *puR)
			{
				if (mainVerbosity & V_FIFO)
				{
				int ee;
				int mm;
					mm = puW - outbuffer2;
					mm = (mm < 4) ? 0 : mm-4;
					ee = packsize2 - mm;
					ee = (ee > 12) ? 12 : ee;
					printf("     Starting %d into buffer 2, values written|read\n     ", mm);
					for (; ee; ee--, mm++)
					{
						printf("%x|%x ", outbuffer2[mm], inbuffer2[mm]);
					}
					printf("\n");
				}
				f->f_wstring[0] = *puW;
				f->f_rstring[0] = *puR;
				if (pB->i2eDataWidth16)
				{
					if ((puW - outbuffer2) & 1)
					{
						f->f_results[1] = 2;	/* High-byte */
					}
					else if (puW[1] != puR[1])
					{
						f->f_wstring[1] = puW[1];
						f->f_rstring[1] = puR[1];
						f->f_results[1] = 3;	/* Both bytes */
					}
					else
					{
						f->f_results[1] = 1;	/* Low byte */
					}
				}
				POST_ERROR(f, FLL_DATA_ERROR);
			}
		}

		packsize1 -= 6;
		if (packsize1 <= 0)
		{
			packsize1 = 252;
		}

		packsize2 += 2;
		if (packsize2 >= 254)
		{
			packsize2 = 6;
		}

		f->f_passfail++;
		kbhit();		/* This ensures ctrlC detection */

	} while (--passes);
	return 0;
}


/*-----------------------------------------------------------------------
 * fifo_mailbox (f)
 *
 * Does testing of fifo mailboxes -- utilizes downloaded code
 * Returns 0 if ok, -1 if failure of any sort
 * 
 * Input parameters:
 * 	f_pB:		pointer to board structure
 * 	f_passes:	number of passes (0 --> 65536)
 * 
 * Output:
 * 	f_error:	OKAY:			no error
 * 				FM_TIMEOUT_IN:	timed-out getting data in inbound
 * 					f_passfail:		pass number failed
 * 				FM_TIMEOUT_OUT:	timed-out getting data to outbound
 * 					f_passfail:		pass number failed
 * 				FM_DATA_IN:		data in inbound incorrect
 * 					f_passfail:		pass number failed
 * 					f_wstring[0]:	data expected
 * 					f_rstring[0]:	data read
 * 				FM_DATA_OUT:	data in outbound incorrect
 * 					f_passfail:		pass number failed
 * 					f_results[0]:	data expected
 * 					f_results[0]:	data read
 * 
 *-----------------------------------------------------------------------
 */

int fifo_mailbox (pfunk f)
{
int i;
uchar kar;
int k2;
ushort passes;
i2eBordStrPtr pB = f->f_pB;
unsigned char validMask;

	/* We write to mailbox, the board reads it, complements it,
	 * and writes it back.  We read, complement again, and check.
	 */

	/* The -IIEX has only a 6-bit Mailbox!, where the
	 * -II has an 8-bit one
	 */

	validMask = ((pB->i2eFifoStyle == FIFO_IIEX) ? 0xfc : 0xff) ;
	passes = f->f_passes;
	f->f_passfail = 0;
	do {
		kar = 1;		/* Don't use 0 value: ambiguous */
		for (i=1; i<256; i++)
		{
			if (!iiTrySendMail(pB, kar))
			{
				POST_ERROR(f, FM_TIMEOUT_OUT);
			}
			if ((k2 = wait_mailbox_in (pB, 2)) < 0)
			{
				POST_ERROR(f, FM_TIMEOUT_IN);
			}
			
			if ( ((~k2) ^ kar) & validMask)
			{
				f->f_wstring[0] = kar & validMask;
				f->f_rstring[0] = k2 & validMask;
				POST_ERROR(f, FM_DATA_IN);
			}
			kar += 19;
			if (!kar) kar += 19;	/* Prevent 0 value from being used */
		}
		f->f_passfail++;
		kbhit();		/* This ensures ctrlC detection */
	} while (--passes);
	return 0;
}



/*-----------------------------------------------------------------------
 * register_test (f)
 * Tests registers on the board -- utilizes downloaded code
 * Returns 0 if ok, -1 if failure of any sort
 * 
 * Input parameters:
 *	f_pB:		pointer to board structure
 * 	f_passes:	number of passes (0 --> 65536)
 * 
 * Output:
 * 	f_error:	OKAY:			no error
 * 				RT_FAIL:		a failure occurred
 * 					f_passfail:		pass number failed
 * 					f_results[0]:	failed device:
 * 									DEVICE_1400
 * 									DEVICE_8259
 * 									DEVICE_80186
 *									DEVICE_LCD
 *									DEVICE_654
 * 					f_results[1]:	if DEVICE_80186, value written (MSB)
 * 								if DEVICE_8259, which one (0-3)
 * 								if DEVICE_1400 or
 * 								   DEVICE_654, which one (0-15)
 *								if DEVICE_LCD, which one (0-3)
 * 					f_results[2]:	if DEVICE_1400:
 * 									GLOBAL_REG if global register failed
 * 									CHANNEL_REG if channel register failed
 *									if DEVICE_80186, value read (MSB)
 * 					f_results[3]:	register address that failed
 * 					f_wstring[0]:	value written
 * 					f_rstring[0]:	value read
 * 				RT_GENERAL:		misc. failure
 * 					f_passfail:		pass number failed
 * 					f->results[0]:	failure code
 * 
 *-----------------------------------------------------------------------
 */
int register_test (pfunk f)
{
uchar val;
ushort passes;
i2eBordStrPtr pB = f->f_pB;

	/* Tell it how many passes to do */
	if (word_data_out (pB, 1, f->f_passes))	/* Should be immediately ready */
	{
		f->f_results[0] = 0x8010;
		POST_ERROR(f, RT_GENERAL);
	}
	/* Clear any result MSB's */
	f->f_results[0] = 
		f->f_results[1] = 
		f->f_results[2] = 
		f->f_results[3] = 0;

	passes = f->f_passes;
	f->f_passfail = 0;
	do {
		val = byte_data_in (pB, 10);
								/* Each pass should be less than 10 seconds */
								/* Even for IntelliPort-IIEX with 4 boxes */
		if (val)	//error
		{
			if (	(byte_data_read(pB, 2 , (uchar *)&(f->f_results[0])) )	||
					(byte_data_read(pB, 2 , (uchar *)&(f->f_results[1])) )	||
					(byte_data_read(pB, 2 , (uchar *)&(f->f_results[2])) )	||
					(byte_data_read(pB, 2 , (uchar *)&(f->f_results[3])) )	||
					(byte_data_read(pB, 2 , (uchar *)&(f->f_wstring[0])) )	||
					(byte_data_read(pB, 2 , (uchar *)&(f->f_rstring[0])) ) )
			{
				f->f_results[0] = 0x8000;
				POST_ERROR(f, RT_GENERAL);
			}
			POST_ERROR(f, RT_FAIL);
		}
		f->f_passfail++;
		kbhit();		/* This ensures ctrlC detection */
	} while (--passes);
	SUCCEED(f);
}



/*-----------------------------------------------------------------------
 * register_exercise (f)
 * Exercises registers on the board's external bus -- uses downloaded code
 * Runs for the specified period, then returns (always successful)
 * 
 * Input parameters:
 *	f_pB:		pointer to board structure
 * 	f_passes:	number of passes (0 --> 65536) (= #seconds)
 * 
 * Output:
 * 	f_error:	OKAY:			no error
 * 
 *-----------------------------------------------------------------------
 */
int register_exercise (pfunk f)
{
time_t later = time(NULL) + f->f_passes;
	
	while (time(NULL) < later)
	{
		kbhit();
	}

	SUCCEED(f);
}



/*-----------------------------------------------------------------------
 * interrupt_test (f)
 * Tests interrupt vectoring from the board -- utilizes downloaded code
 * Returns 0 if ok, -1 if failure of any sort
 * 
 * Input parameters:
 * 	f_pB:			Pointer to board structure.
 * 	f_passes:		number of passes (0 --> 65536)
 * 	f_interrupts:	bitmap of interrupts to test:
 * 					("or" of the manifests IRQ_3 thru IRQ_15)
 * 					bit 0: 1 --> test IRQ 3
 * 					bit 1: 1 --> test IRQ 4
 * 					bit 2: 1 --> test IRQ 5
 * 					bit 3: 1 --> test IRQ 7
 * 					bit 4: 1 --> test IRQ 10
 * 					bit 5: 1 --> test IRQ 11
 * 					bit 6: 1 --> test IRQ 12
 * 					bit 7: 1 --> test IRQ 15
 * 
 * Output:
 * 	f_error:	OKAY:			no error
 * 				IT_FAIL:		a failure occurred
 * 					f_passfail:		pass number failed
 *					f_results[0-7]:	The number of interrupts issued on
 *									IRQ 3, 4, 5, 7, 10, 11, 12, 15
 *					f_results[8-15]: The number of interrupts caught on
 *									IRQ 3, 4, 5, 7, 10, 11, 12, 15
 *									
 * 				IT_GENERAL:		misc. failure
 * 					f_passfail:		pass number failed
 * 					f->results[0]:	failure code
 * 
 *-----------------------------------------------------------------------
 */

static int interruptsIssued[16];
static int interruptsCaught[16];
static int xlatIrq[8] = { 3, 4, 5, 7, 10, 11, 12, 15 };

static int bitsList[] = {IRQ_3, IRQ_4, IRQ_5, IRQ_7,
			IRQ_10, IRQ_11, IRQ_12, IRQ_15};

static int posList[] = {0x03, 0x23, 0x43, 0x63, 0x83, 0xa3, 0xc3, 0xe3 };
static int eisaList[] = {8, 9, 10, 11, 12, 13, 14, 15};
static int bitsel[4] = {0x55, 0xaa, 0x0f, 0xf0};

static int (*selectIrq)(int);
static int *selList;

/* The interrupt test for the PCI bus is
 * quite different from the others:
 * The card has only a single interrupt
 * request line, which is driven by any
 * of the separate interrupt requests
 * out of the CEX asic.  The responsibility
 * of "seeing" the correct interrupt number
 * is the "hosts", so we need only test the
 * interrupt line that the PCI board is
 * configured for.  However, we should ensure
 * that whatever interrupt we tell the board
 * to issue, it will still cause _an interrupt.
 */

extern int chiefInterrupt;
int pci_interrupt_test (pfunk f)
{
int i, j, k;
ushort passes;
unsigned int status;
i2eBordStrPtr pB = f->f_pB;
int bits = f->f_interrupts;
int bcount = 0;
unsigned int missed = 0;

	passes = f->f_passes;

	if(bits == 0) bits = 0xff;	/* Ask for nothing in particular, and
								 * it means check everything!
								 */
	if(bits & 1)bcount++;
	if(bits & 2)bcount++;
	if(bits & 4)bcount++;
	if(bits & 8)bcount++;
	if(bits & 0x10)bcount++;
	if(bits & 0x20)bcount++;
	if(bits & 0x40)bcount++;
	if(bits & 0x80)bcount++;

	pTestBoard = pB;

	/* just link in the interrupt handler for the interrupt
	 * we are configured to use.
	 */

	if (	!i2LinkEnableIrq(0, &pB, chiefInterrupt,  
				(void (far *)())i_handler_chief) )
	{
		i2RestoreIrqs();
		f->f_results[0] = 0x8222;
		POST_ERROR(f, IT_GENERAL);
	}	

	iiEnableMailIrq(pB);	/* Enable mailbox interrupts */

	iiDelay(pB, 3);			/* a Teeny delay to let interrrupts in... */

	/* Before we try to GENERATE any interrupts, we 
	 * allow ourselves to service any "leftover" interrupts
	 * which might have accrued from previous tests, previous
	 * boards, etc...
	 * But there shouldn't be more than just one!
	 */
	
	for (i = 0; i < 16; i++)
	{
		if (interruptsCaught[i] > 1)
		{
			i2RestoreIrqs();
			f->f_results[0] = 0x8226;
			POST_ERROR(f, IT_GENERAL);
		}
		interruptsCaught[i] = 0;
	}

	/* Now, try to generate some interrupts: one for each specified pass. */

	/* Note ctrlexit from this might imply not restoring the selected
	 * interrupt properly.
	 */

	do {
		for (k = 0; k < 4; k++)
		{
			if (	((bits & 1)    && byte_data_out(pB, 1, 3)) ||
					((bits & 2)    && byte_data_out(pB, 1, 4)) ||
					((bits & 4)    && byte_data_out(pB, 1, 5)) ||
					((bits & 8)    && byte_data_out(pB, 1, 7)) ||
					((bits & 0x10) && byte_data_out(pB, 1, 10)) ||
					((bits & 0x20) && byte_data_out(pB, 1, 11)) ||
					((bits & 0x40) && byte_data_out(pB, 1, 12)) ||
					((bits & 0x80) && byte_data_out(pB, 1, 15)) )
			{
				i2RestoreIrqs();
				f->f_results[0] = 0x8000;
				POST_ERROR(f, IT_GENERAL);
			}	
			interruptsIssued[chiefInterrupt] += bcount;	

			/* Now all the characters are in the fifo to tell the
	 		* board to issue this and that type of interrupt.
	 		*/
			/* Now ensure the board has read all the data */
	
			i = 8;	/* Enough to flush out any bad interrupts and let the
				 	* good ones come through.
				 	*/
#define DEBUG_CLUDGE_3
#ifdef DEBUG_CLUDGE_3
			i = 300;	/* 300 is plenty of time */
			i = 100;
#endif

			while ((status = checkInterrupts()) != 0)
			{
				if (!(i--)) break;		/* Waited plenty long */
#ifndef DEBUG_CLUDGE_3
				iiDelay(pB, 1);		/* Delay to allow plenty of time 
								 	* for board to place interrupt
								 	*/
#endif
				/* Shouldn't really check the mail until
				 * we have waited a long time for it to
				 * give us an interrupt (10 uS anyway), since
				 * it could take that long for us to see
				 * the interrupt.
				 */
				if (i == 1 && iiGetMail(pB) != NO_MAIL_HERE)
				{
				OUTB(pB->i2eData+4, 0xaa);	/* DEBUGGING */
					missed++;
					iiDelay(pB, 1);
				}					/* If there is any mail here, it must
								 	* be that the interrupt we wanted is
								 	* inoperative.  So we read the mailbox
								 	* to allow the board to continue with
								 	* other interrupts, then insert an
								 	* additional delay to allow the board
								 	* plenty of time for the next int.
								 	*/
			}
			if (status)
			{
				OUTB(pB->i2eData+4, 0xab);	/* DEBUGGING */
				break;	/* Error ends the test */
			}
	
			f->f_passfail++;
		}
		if (status) break;

	} while (--passes);

	f->f_passfail--;	/* So that the pass count will be reported right */

	/* Use library routines to unlink our handlers */

	i2RestoreIrqs();

	for (i = 0; i < 8; i++)
	{
		j = xlatIrq[i];
		f->f_results[i]   = interruptsIssued[j];
		f->f_results[i+8] = interruptsCaught[j];
	}
	if (missed)
	{
		printf("\nWarning: missed %d interrupts; recovered in software\n", missed);
		POST_ERROR(f, IT_FAIL);
	}	
	if (status)
	{
		POST_ERROR(f, IT_FAIL);
	}
	SUCCEED(f);
}

int interrupt_test (pfunk f)
{
int i, j, k;
ushort passes;
unsigned int status;
i2eBordStrPtr pB = f->f_pB;
int bits = f->f_interrupts;
int savePOS;
int saveLevels = NO_CHANGE_LEVELS;	/* If we don't save, no restore */
									/* Particularly prevents restoring if
									 * not even on an EISA machine.
									 * If called with NO_CHANGE_LEVELS,
									 * levelsForEISA() does nothing.
									 */
int toTest;
int tbits;
unsigned int missed = 0;

	for (i = 0; i < 16; i++)
	{
		interruptsIssued[i] = interruptsCaught[i] = 0;
	}

	if(busType == PCI_OPT)
	{
		return pci_interrupt_test(f);
	}

	passes = f->f_passes;


	pTestBoard = pB;

	/* link in our interrupt handlers for all potential interrupts */

	if (	!i2LinkEnableIrq(0, &pB, CIR_3,  (void (far *)())i_handler_3)  ||
			!i2LinkEnableIrq(0, &pB, CIR_4,  (void (far *)())i_handler_4)  ||
			!i2LinkEnableIrq(0, &pB, CIR_5,  (void (far *)())i_handler_5)  ||
			!i2LinkEnableIrq(0, &pB, CIR_7,  (void (far *)())i_handler_7)  ||
			!i2LinkEnableIrq(0, &pB, CIR_10, (void (far *)())i_handler_10) ||
			!i2LinkEnableIrq(0, &pB, CIR_11, (void (far *)())i_handler_11) ||
			!i2LinkEnableIrq(0, &pB, CIR_12, (void (far *)())i_handler_12) ||
			!i2LinkEnableIrq(0, &pB, CIR_15, (void (far *)())i_handler_15) )
	{
		i2RestoreIrqs();
		f->f_results[0] = 0x8222;
		POST_ERROR(f, IT_GENERAL);
	}	

	iiEnableMailIrq(pB);	/* Enable mailbox interrupts */

	iiDelay(pB, 3);			/* a Teeny delay to let interrrupts in... */

	/* Before we try to GENERATE any interrupts, we 
	 * allow ourselves to service any "leftover" interrupts
	 * which might have accrued from previous tests, previous
	 * boards, etc...
	 * But there shouldn't be more than just one!
	 */
	
	for (i = 0; i < 16; i++)
	{
		if (interruptsCaught[i] > 1)
		{
			i2RestoreIrqs();
			f->f_results[0] = 0x8226;
			POST_ERROR(f, IT_GENERAL);
		}
		interruptsCaught[i] = 0;
	}

	/* Now, try to generate some interrupts: one for each specified pass. */

	/* Note ctrlexit from this might imply not restoring the selected
	 * interrupt properly.
	 */

	if (hasUsedB && ((busType == MCA_OPT) || (busType == EISA_OPT)) )
	{
		if (busType == MCA_OPT)
		{
			selectIrq = selMCAirq;
			selList = posList;
		}
		else
		{
			selectIrq = selEISAirq;
			selList = eisaList;
			saveLevels = levelsForEISA(convertMask(bits));
		}
			
		savePOS = (*selectIrq)(-1);

		do {
			for(toTest = 0; toTest < 8; toTest++)
			{
				if (bits & bitsList[toTest])
				{
					/* Select the interrupt to use */
					(*selectIrq)(selList[toTest]);

					/* Because of overhead of selecting interrupts,
					 * do four in a row for the same kind now.
					 */
					for (k = 0; k < 4; k++)
					{
						if (do_an_interrupt(pB, xlatIrq[toTest], 
								bitsList[toTest] & bits) )
						{
							(*selectIrq)(savePOS);
							i2RestoreIrqs();
							levelsForEISA(saveLevels);
							f->f_results[0] = 0x8000;
							POST_ERROR(f, IT_GENERAL);
						}	
					}
	
					/* Now see that we got all four */

					i = 4;	/* Time limit for interrupts to arrive */
					while ((status = checkInterrupts()) != 0)
					{
						if (!(i--)) break;	/* Waited plenty long */
						iiDelay(pB, 1);		/* Delay to allow plenty of time 
								 			 * for board to place interrupt
								 		 	 */
						/* Fail safe against a missing interrupt */
						/* If interrupts are o.k., there shouldn't
						 * be anything in the mailbox here.  Reason is,
						 * it should have interrupted us and we would
						 * have read the mailbox already.
						 */
						if (iiGetMail(pB) != NO_MAIL_HERE)
						{
							missed++;
							do {
								iiDelay(pB, 1);
							} while (iiGetMail(pB) != NO_MAIL_HERE);
						}
					}
				}
			}	/* end loop on interrupt type */

			if (status) break;	/* Error ends the test */
		
			f->f_passfail++;
	
		} while (--passes);
		(*selectIrq)(savePOS);
	}
	else	/* ISA card */
	{
		do {
			for (k = 0; k < 4; k++)
			{
				tbits = bits & bitsel[k];
				if (	do_an_interrupt (pB,  3, IRQ_3  & tbits) ||
						do_an_interrupt (pB,  3, IRQ_3  & tbits) ||
						do_an_interrupt (pB,  4, IRQ_4  & tbits) ||
						do_an_interrupt (pB,  4, IRQ_4  & tbits) ||
						do_an_interrupt (pB,  5, IRQ_5  & tbits) ||
						do_an_interrupt (pB,  5, IRQ_5  & tbits) ||
						do_an_interrupt (pB,  7, IRQ_7  & tbits) ||
						do_an_interrupt (pB,  7, IRQ_7  & tbits) ||
						do_an_interrupt (pB, 10, IRQ_10 & tbits) ||
						do_an_interrupt (pB, 10, IRQ_10 & tbits) ||
						do_an_interrupt (pB, 11, IRQ_11 & tbits) ||
						do_an_interrupt (pB, 11, IRQ_11 & tbits) ||
						do_an_interrupt (pB, 12, IRQ_12 & tbits) ||
						do_an_interrupt (pB, 12, IRQ_12 & tbits) ||
						do_an_interrupt (pB, 15, IRQ_15 & tbits) ||
						do_an_interrupt (pB, 15, IRQ_15 & tbits) )
				{
					i2RestoreIrqs();
					f->f_results[0] = 0x8000;
					POST_ERROR(f, IT_GENERAL);
				}	
	
				/* Now all the characters are in the fifo to tell the
		 		* board to issue this and that type of interrupt.
		 		*/
				/* Now ensure the board has read all the data */
		
				i = 8;	/* Enough to flush out any bad interrupts and let the
					 	* good ones come through.
					 	*/
				while ((status = checkInterrupts()) != 0)
				{
					if (!(i--)) break;		/* Waited plenty long */
					iiDelay(pB, 1);		/* Delay to allow plenty of time 
									 	* for board to place interrupt
									 	*/
					if (iiGetMail(pB) != NO_MAIL_HERE)
					{
						missed++;
						iiDelay(pB, 1);
					}					/* If there is any mail here, it must
									 	* be that the interrupt we wanted is
									 	* inoperative.  So we read the mailbox
									 	* to allow the board to continue with
									 	* other interrupts, then insert an
									 	* additional delay to allow the board
									 	* plenty of time for the next int.
									 	*/
				}
				if (status) break;	/* Error ends the test */
		
				f->f_passfail++;
			}
			if (status) break;
	
		} while (--passes);
	}

	f->f_passfail--;	/* So that the pass count will be reported right */

	/* Use library routines to unlink our handlers */

	i2RestoreIrqs();
	levelsForEISA(saveLevels);	/* If was EISA and had levels saved, restore */

	for (i = 0; i < 8; i++)
	{
		j = xlatIrq[i];
		f->f_results[i]   = interruptsIssued[j];
		f->f_results[i+8] = interruptsCaught[j];
	}
	if (missed)
	{
		printf("\nWarning: missed %d interrupts; recovered in software\n", missed);
		POST_ERROR(f, IT_FAIL);
	}	
	if (status)
	{
		POST_ERROR(f, IT_FAIL);
	}
	SUCCEED(f);
}


/*----------------------------------
 * checkInterrupts()
 *
 * returns 0 if interrupts issued matches
 * interrupts caught on all IRQ lines.
 * returns -1 if there is a discrepancy
 *------------------------------------
 */

static int
checkInterrupts()
{
register int i;

	for (i = 0; i < 16; i++)
	{
		if (interruptsIssued[i] != interruptsCaught[i]) return -1;
	}
	return 0;
}

/*------------------------------------
 * do_an_interrupt(pB, num, enable)
 *
 * if enable is true, issue an 
 * interrupt, and return -1 if there
 * was a problem.  If no problem, 
 * or if we weren't enabled, return 0;
 *-------------------------------------
 */
static int do_an_interrupt (i2eBordStrPtr pB, int num, int enable)
{
	if (enable)
	{
		if (byte_data_out(pB, 1, num))
		{
			return -1;
		}
		interruptsIssued[num & 0xf]++;
	}
	return 0;
}



/*---------------------------------
 * All the interrupt handlers,
 *---------------------------------
 */
/*-----------------------------------------------------
 * To properly support both edge and level-triggered
 * interrupts:
 * To properly support edge-triggered interrupts from ISA cards, we must 
 * issue the EOI before we service the card.  Servicing the card enables
 * further interrupts to occur and if we have already issued the EOI we
 * will not miss them.
 * But, if we are using level-triggered interrupts, we must mask the
 * interrupt out before sending the EOI; otherwise another interrupt
 * will be generated (or, worse, a "glitch" IRQ7 or IRQ15).
 * So the robust sequence is:
 * MASK, EOI, SERVICE, UNMASK, IRET
 *-------------------------------------------------------
 */
static ushort maskSaver, maskChange;

void interrupt i_handler_chief ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[chiefInterrupt]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver | (1 << chiefInterrupt);
	WRITE_IRQ_MASKS(maskChange);
	if (chiefInterrupt > 7) 
	{
		SLAVE_EOI
	}
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}
void interrupt i_handler_3 ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[3]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver | (1 << 3);
	WRITE_IRQ_MASKS(maskChange);
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}

void interrupt i_handler_4 ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[4]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver | (1 << 4);
	WRITE_IRQ_MASKS(maskChange);
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}

void interrupt i_handler_5 ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[5]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver | (1 << 5);
	WRITE_IRQ_MASKS(maskChange);
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}

void interrupt i_handler_7 ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[7]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver | (1 << 7);
	WRITE_IRQ_MASKS(maskChange);
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}

void interrupt i_handler_10 ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[10]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver |  (1 << 10) | (1 << 2);
	WRITE_IRQ_MASKS(maskChange);
	SLAVE_EOI
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}

void interrupt i_handler_11 ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[11]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver | (1 << 11) | (1 << 2);
	WRITE_IRQ_MASKS(maskChange);
	SLAVE_EOI
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}

void interrupt i_handler_12 ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[12]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver |  (1 << 12) | (1 << 2);
	WRITE_IRQ_MASKS(maskChange);
	SLAVE_EOI
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}

void interrupt i_handler_15 ()
{
	TO_NEW_STACK(localStack, LS_SIZE);
	interruptsCaught[15]++;
	SAVE_IRQ_MASKS(maskSaver);
	maskChange = maskSaver | (1 << 15) | (1 << 2);
	WRITE_IRQ_MASKS(maskChange);
	SLAVE_EOI
	MASTER_EOI
	iiGetMail(pTestBoard);
	WRITE_IRQ_MASKS(maskSaver);
	TO_OLD_STACK(localStack);
}


/*---------------------------------------------------
 * Interactive LCD diagnostic
 *
 *---------------------------------------------------
 */

/*------------------------------------------------------
 * sendLcdString(pB, pc)
 *
 * Assumes the output fifo at pB has room (no checking):
 * sends the string to the fifo (assumes null terminated),
 * which the downloaded code will send to the LCD
 * Data is always written using only the low half of the
 * data bus.
 *------------------------------------------------------
 */

static void
sendLcdString(i2eBordStrPtr pB, char *pc)
{
int address = pB->i2eData;

	asm	{
		cld;
		mov	si, pc;
		mov	dx, address;
	}
sls:
	asm {
		lodsb;
		test	al, al;
		jz		sls1;
		out	dx, al;
		jmp	sls;
	}
sls1:
	return;
}

static int
waitForAnyKey(i2eBordStrPtr pB)
{
	while (1)
	{
		if (HAS_INPUT(pB))
		{
			return (BYTE_FROM(pB));
		}
	iiDelay(pB, 1);
	}
}
static void
waitForSpecificKey(i2eBordStrPtr pB, int key)
{
int x;		/* To make adding debugging code easier */
	while (1)
	{
		x = waitForAnyKey(pB);
		if (x == key) return;
		iiDelay(pB,1); 
	}
}
			
		

/*----------------------------------------
 * interactive_lcd_one(pB, box)
 * Performs interactive test on LCD on one box
 *----------------------------------------
 */

static char pr0[] = {"\007\201\010\203"};
static char pr1[] = {"\001\200Press And Hold\037*\002"};
static char pr2[] = {"\001\200Press And Hold\037       *\002"};
static char pr3[] = {"\001\200Press And Hold\037              *\002"};
static char rel[] = {"\001\200    Release\002"};
static char brt[] = {"\001\200Bright - Dim...\002"};
static char walk[] = {"\001\300A\002"};
static char adj[] = {"\010\200"};
static char vers[] = {"\001\200\Version...\002\006"};
static char bdone[] = {"\001\200Box 0 LCD\037Test Done\002"};
static char thevers[] ={"\001\300?.? ??-port\002"};

static void
interactive_lcd_one(register i2eBordStrPtr pB, int box)
{
int i;

	iiTrySendMail(pB, box << 4);	/* To select box */
	sendLcdString(pB, pr0);
	sendLcdString(pB, pr1);
	waitForSpecificKey(pB, 0x4);
	sendLcdString(pB, rel);
	waitForSpecificKey(pB, 0x0);

	sendLcdString(pB, pr2);
	waitForSpecificKey(pB, 0x2);
	sendLcdString(pB, rel);
	waitForSpecificKey(pB, 0x0);

	sendLcdString(pB, pr3);
	waitForSpecificKey(pB, 0x1);
	sendLcdString(pB, rel);
	waitForSpecificKey(pB, 0x0);

	walk[1] = 0xc0;
	walk[2] = 'A';
	adj[1] = 0x80;

	sendLcdString(pB, brt);

	for (i = 0; i < 16; i++)
	{
		sendLcdString(pB, adj);
		iiDelay(pB, 250);	/* one-quarter second */
		if (i == 0) iiDelay(pB, 500);	/* Extra time at highest level */
		sendLcdString(pB, walk);
	
		walk[1]++;
		walk[2]++;
		adj[1]++;
	}
	iiDelay(pB, 500);	/* One/half second at lowest level */
	sendLcdString(pB, pr0);
	sendLcdString(pB, vers);
	waitForSpecificKey(pB, 0xe);
	thevers[2] = waitForAnyKey(pB);
	thevers[4] = waitForAnyKey(pB);
	switch (waitForAnyKey(pB))
	{
	case 0x80:
		thevers[6] = ' ';
		thevers[7] = '8';
		break;

	case 0x83:
		thevers[6] = '1';
		thevers[7] = '6';
		break;

	default:
		thevers[6] = '?';
		thevers[7] = '?';
		break;
	}
	waitForSpecificKey(pB, 0x80);
	sendLcdString(pB, thevers);
	
	iiDelay(pB, 1000);	/* One second at dimmest level */
	bdone[6] = '0' + box;
	sendLcdString(pB, pr0);
}

	
	
int lcd_interactive(pfunk f)
{
ushort passes = f->f_passes;
i2eBordStrPtr pB = f->f_pB;
int box;

	do {
		for (box = 0; box < ABS_MAX_BOXES; box++)
		{
			/* If there are any channels here, we assume there is an LCD */
			if (pB->i2eChannelMap[box])
			{
				interactive_lcd_one(pB, box);
			}
		}

	} while (--passes);

	SUCCEED(f);
}	




/*----------------------------------------------------
 * wait_mailbox_in (pB, nsec)
 *
 * pB -- pointer to board structure
 * nsec -- number of seconds to time out
 *		Note, if nsec = 1, an arbitrarily short timeout
 *		might apply since one second granularity applies
 *		to the time() function.
 *
 * Returns the mailbox value, or -1 if times out
 *----------------------------------------------------
 */

static int wait_mailbox_in (i2eBordStrPtr pB, int nsec)
{
time_t old_time;
unsigned short status;

	if ((status = iiGetMail(pB)) != NO_MAIL_HERE)
	{
		return (status & 0xff);
	}
		
	old_time = time(NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if ((status = iiGetMail(pB)) != NO_MAIL_HERE)
		{
			return (status & 0xff);
		}
	}
	return -1;
}



/*---------------------------------------------
 * byte_data_in4i(pB, ch)
 * word_data_read4i(pB, ch, pvalue)
 *
 * These functions return -1 if 
 * the fifo is completely empty.  Otherwise,
 * it performs a few i/o reads to ensure the
 * loadware has had time to loopback all the
 * characters pending, then it tries to read
 * 4 characters (or words) quickly, back-to-back.
 * If any are different, it protests.
 *----------------------------------------------
 */
static int byte_data_in4i (pB, value)
i2eBordStrPtr pB;
int value;
{
int address = pB->i2eData;

	/* The caller has now delayed explicitly to allow the
	 * firmware time to copy the data.
	 */

	if (HAS_NO_INPUT(pB))
	{
		return -1;
	}

	asm {
		push	dx;

		mov	dx, address;
		in	al, dx;
		xchg	ah, al;
		in	al, dx;
		xchg	si, ax;
		in	al, dx;
		xchg	ah, al;
		in	al, dx;
		cmp	al, value;
		jne	retme;
		mov	al, ah;
		cmp	al, value;
		jne	retme;
		mov	ax, si;
		cmp	al, value;
		jne	retme;
		mov	al, ah;
	}
retme:
	asm {
		pop	dx;
		xor	ah, ah;
		mov value, ax;
	}
	return value;
}	
		
static int word_data_read4i (pB, value, pvalue)
i2eBordStrPtr pB;
int value;
ushort *pvalue;
{
int address = pB->i2eData;

	/* The caller has now delayed explicitly to allow the
	 * firmware time to copy the data.
	 */

	if (HAS_NO_INPUT(pB))
	{
		return -1;
	}

	asm {
		push	dx;
		push	cx;

		mov	dx, address;
		in	ax, dx;
		xchg	si, ax;
		in	ax, dx;
		xchg	di, ax;
		in	ax, dx;
		xchg	cx, ax;
		in	ax, dx;

		cmp	ax, value;
		jne	retme;
		xchg	ax, cx;
		cmp	ax, value;
		jne	retme;
		xchg	ax, di;
		cmp	ax, value;
		jne	retme;
		xchg	ax, si;
	}
retme:
	asm {
		pop	cx;
		pop	dx;
		mov value, ax;
	}
	*pvalue = value;
	return 0;
}	
		


/*--------------------------------------------
 * byte_data_out(pB, nsec, kar)
 *
 * Tries to output kar to the fifo data port.
 * If successful, returns 0.
 * Returns -1 if it times out after nsec seconds
 * 
 * Tries once before invoking any time functions:
 * this for speed.
 *
 * if this is used on 16-bit boards, the high bit
 * is lost.
 *----------------------------------------------
 */

static int byte_data_out (pB, nsec, kar)
i2eBordStrPtr pB;
uchar kar;
int nsec;
{
time_t old_time;

	if (HAS_OUTROOM(pB))
	{
		BYTE_TO(pB, kar);
		return 0;
	}
	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_OUTROOM(pB))
		{
			BYTE_TO(pB, kar);
			return 0;
		}
	}
	return -1;
}


/*--------------------------------------------
 * word16_data_out(pB, nsec, kar)
 *
 * Tries to output kar to the fifo data port.
 * If successful, returns 0.
 * Returns -1 if it times out after nsec seconds
 * 
 * Tries once before invoking any time functions:
 * this for speed.
 *
 * Assumes 16 bit-interface
 *----------------------------------------------
 */

static int word16_data_out (pB, nsec, kar)
i2eBordStrPtr pB;
unsigned short kar;
int nsec;
{
time_t old_time;

	if (HAS_OUTROOM(pB))
	{
		WORD_TO(pB, kar);
		return 0;
	}
	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_OUTROOM(pB))
		{
			WORD_TO(pB, kar);
			return 0;
		}
	}
	return -1;
}


/*--------------------------------------------
 * word_data_out(pB, nsec, value)
 *
 * Tries to output value to the fifo data port.
 * If successful, returns 0.
 * Returns -1 if it times out after nsec seconds
 * 
 * Tries once before invoking any time functions:
 * this for speed.
 *
 * This may be used TRANSPARENTLY on either
 * 8 or 16-bit interfaces.
 *----------------------------------------------
 */

static int word_data_out (pB, nsec, value)
i2eBordStrPtr pB;
ushort value;
int nsec;
{
time_t old_time;

	if (HAS_OUTROOM(pB))
	{
		iiWriteWord(pB, value);
		return 0;
	}
	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_OUTROOM(pB))
		{
			iiWriteWord(pB, value);
			return 0;
		}
	}
	return -1;
}



/*--------------------------------------------
 * word_data_read(pB, nsec, pvalue)
 *
 * If data is available, inputs a word from the
 * fifo data port to *pvalue, returns 0.
 *
 * Returns -1 if it times out after nsec seconds
 * 
 * Tries once before invoking any time functions:
 * this for speed.
 *
 * This may be used TRANSPARENTLY on either
 * 8 or 16-bit interfaces.
 *----------------------------------------------
 */

static int word_data_read (pB, nsec, pvalue)
i2eBordStrPtr pB;
ushort *pvalue;
int nsec;
{
time_t old_time;

	if (HAS_INPUT(pB))
	{
		*pvalue = iiReadWord(pB);
		return 0;
	}
	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_INPUT(pB))
		{
			*pvalue = iiReadWord(pB);
			return 0;
		}
	}
	return -1;
}

/*--------------------------------------------
 * word16_data_read(pB, nsec, pvalue)
 *
 * If data is available, inputs a word from the
 * fifo data port to *pvalue, returns 0.
 *
 * Returns -1 if it times out after nsec seconds
 * 
 * Tries once before invoking any time functions:
 * this for speed.
 *
 * This assumes a 16-bit interface
 *----------------------------------------------
 */

static int word16_data_read (pB, nsec, pvalue)
i2eBordStrPtr pB;
ushort *pvalue;
int nsec;
{
time_t old_time;

	if (HAS_INPUT(pB))
	{
		*pvalue = WORD_FROM(pB);
		return 0;
	}
	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_INPUT(pB))
		{
			*pvalue = WORD_FROM(pB);
			return 0;
		}
	}
	return -1;
}



/*--------------------------------------------
 * byte_data_read(pB, nsec, pvalue)
 *
 * If data is available, inputs a byte from the
 * fifo data port to *pvalue, returns 0.
 *
 * Returns -1 if it times out after nsec seconds
 * 
 * Tries once before invoking any time functions:
 * this for speed.
 *
 * If used with 16-bit fifos, the upper byte is
 * lost.
 *----------------------------------------------
 */

static int byte_data_read (pB, nsec, pbyte)
i2eBordStrPtr pB;
unsigned char *pbyte;
int nsec;
{
time_t old_time;

	if (HAS_INPUT(pB))
	{
		*pbyte = BYTE_FROM(pB);
		return 0;
	}
	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_INPUT(pB))
		{
			*pbyte = BYTE_FROM(pB);
			return 0;
		}
	}
	return -1;
}


/*--------------------------------------------
 * byte_data_in(pB, nsec)
 *
 * If data is available, returns a byte from the
 * fifo data port. (zero-extended)
 *
 * Returns -1 if it times out after nsec seconds
 * 
 * Tries once before invoking any time functions:
 * this for speed.
 *
 * If used on 16-bit boards, throws away the 
 * high-order byte.
 *----------------------------------------------
 */

static int byte_data_in (pB, nsec)
i2eBordStrPtr pB;
int nsec;
{
time_t old_time;

	if (HAS_INPUT(pB))
	{
		return (0xff & BYTE_FROM(pB));
	}
	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_INPUT(pB))
		{
			return (0xff & BYTE_FROM(pB));
		}
	}
	return -1;
}


/*------------------------------------------------------------------
 * byte_data_out8e(pB, nsec, value)
 * word_data_out8e(pB, nsec, value)
 *
 * Attempts to output the value to the fifo (8/16 bit mode)
 * returns -1 if timesout after nsec seconds
 * Outputs the value 4 times rapidly, complements, and again 4 times.
 * 
 * Note: we expect the circumstances to guarantee that if there is
 * room for the first character, there is room for the other 8.
 *-------------------------------------------------------------------
 */

static int byte_data_out8e (pB, nsec, kar)
i2eBordStrPtr pB;
uchar kar;
int nsec;
{
time_t old_time;
int address;

	address = pB->i2eData;

	if (HAS_OUTROOM(pB))
	{
		asm {
			push dx;
			mov al,kar;
			mov dx,address;
			out dx,al;
			out dx,al;
			out dx,al;
			out dx,al;
			not al;
			out dx,al;
			out dx,al;
			out dx,al;
			out dx,al;
			pop	dx;
		}
		return 0;
	}

	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_OUTROOM(pB))
		{
			asm {
				push dx;
				mov al,kar;
				mov dx,address;
				out dx,al;
				out dx,al;
				out dx,al;
				out dx,al;
				not al;
				out dx,al;
				out dx,al;
				out dx,al;
				out dx,al;
				pop	dx;
			}
			return 0;
		}
	}
	return -1;
}



static int word_data_out8e (pB, nsec, value)
i2eBordStrPtr pB;
ushort value;
int nsec;
{

time_t old_time;
int address;

	address = pB->i2eData;

	if (HAS_OUTROOM(pB))
	{
		asm {
			push dx;
			mov ax,value;
			mov dx,address;
			out dx,ax;
			out dx,ax;
			out dx,ax;
			out dx,ax;
			not ax;
			out dx,ax;
			out dx,ax;
			out dx,ax;
			out dx,ax;
			pop	dx;
		}
		return 0;
	}

	old_time = time (NULL) + nsec;
	while (time(NULL) < old_time)
	{
		if (HAS_OUTROOM(pB))
		{
			asm {
				push	dx;
				mov ax,value;
				mov dx,address;
				out dx,ax;
				out dx,ax;
				out dx,ax;
				out dx,ax;
				not ax;
				out dx,ax;
				out dx,ax;
				out dx,ax;
				out dx,ax;
				pop	dx;
			}
			return 0;
		}
	}
	return -1;
}


/*-------------------------------
 * PCI support is after this part.
 *-------------------------------
 */

asm .386

static int
getPCIhandle(int index)
{
int retval;

	asm {
		mov ax, 0b102h;
		mov cx, 0291h;
		mov dx, 8e0eh;
		mov si, index;
		int 1ah;
		jc bbpci1;
		cmp ah, 0;
		jne bbpci1;
		mov retval, bx;
	}
	return retval;
bbpci1:
	return -1;
}

static int
getPCIword(int handle, int addr)
{
int retval;

	asm {
		mov ax, 0b109h;
		mov	bx, handle;
		mov di, addr;
		int 1ah;
		mov retval, cx;
	}
	return retval;
}

static unsigned char
getPCIbyte(int handle, int addr)
{
unsigned char retval;

	asm {
		mov ax, 0b108h;
		mov	bx, handle;
		mov di, addr;
		int 1ah;
		mov retval, cl;
	}
	return retval;
}

		
int
biosPCIpresent()
{
	asm {
		mov ax, 0b101h;
		int 1ah;
		jc  bbpci0;
		cmp edx, 20494350h;
		jne bbpci0;
		cmp ah, 0;
		jne bbpci0;
		}
		return 1;
bbpci0:
		return 0;
}
	
int Amcc_Address;

findPCIslot(int * paddress, int * pirq)
{
int handle;
int goodness = 0;
int setup_address[8];
int setup_irq[8];
int i;
int index;

	for(index = 0; index < 8; index++)
	{
		if((handle=getPCIhandle(index)) == -1) break;

		/* We found a PCI card... */

		setup_address[index] = getPCIword(handle, 0x14) & 0xfffe;
		setup_irq[index] = getPCIbyte(handle, 0x3c);

		if (*paddress == 0)
		{
			*paddress = setup_address[index];
			*pirq = setup_irq[index];
		}
		if (*paddress == setup_address[index])
		{
			Amcc_Address = getPCIword(handle, 0x10) & 0xfffe;
			goodness = 1;
		}
		if (setup_irq[index] == *pirq || !*pirq)
		{
			goodness |= 2;
		}
	}
	if(mainVerbosity & V_STARTUP || goodness != 3)
	{
		for(i = 0; i < index; i++)
		{
			printf("Computone PCI board %d at address 0x%x, interrupt %d\n",
				i, setup_address[i], setup_irq[i]);
		}
	}
	if (goodness != 3)
	{
		printf("PCI Bios: no board configured for ");
		if (*paddress)
		{
			printf("address 0x%x, ", *paddress);
		}
		else
		{
			printf("any address, ");
		}
		if (*pirq)
		{
			printf("irq %d\n");
		}
		else
		{
			printf("\n");
		}
	}
	return goodness == 3;	
}

/*--------------------------------------
 * pci_checkout(f)
 * 
 * Checks out some operation registers of
 * the AMCC chip on a PCI card.
 *
 * Input parameters:
 * 		f_passes:	number of passes (0 --> 65536)
 * Output:
 * 		f_error:	OKAY:			no error
 *
 *		f_error:	BAD_POS_2		pos register 2 failed
 *		f_error:	BAD_POS_3		pos register 3 failed
 *		f_wstring[0]	data written
 *		f_rstring[0]	data read back
 *		f_error:	BAD_MC			not seemingly on a microchannel
 *-------------------------------------------------------------
 */
			
int
pci_checkout (pfunk f)
{
int i, j;
int passes;

	/* Note that this test has the interrupts off almost the
	 * whole time: implication is that ctrl-C cannot break us
	 * out.
	 */

	if (!hasUsedB || (busType != PCI_OPT) || !Amcc_Address)
	{
		POST_ERROR(f, BAD_PCI);
	}
	
	passes = f->f_passes;
	f->f_passfail = 0;	//spin through the passes
	do {
			for(i = 0; i < 256; i++)
			{
				/* Yes I know I'm only writing bytes. That is
				 * why my loop stops after 255.
				 */
				OUTB(Amcc_Address +0x0, i);
				OUTB(Amcc_Address +0x4, i ^ 0x55);
				OUTB(Amcc_Address +0x8, i ^ 0xaa);
				OUTB(Amcc_Address +0xc, i ^ 0xff);
				
				j = INB(Amcc_Address + 0x0);
				if (i != j) 
				{	
					f->f_results[0] = Amcc_Address + 0x0;
					f->f_wstring[0] = i;
					f->f_wstring[1] = j;
					POST_ERROR(f, BAD_OMB);
				}
				
				j = INB(Amcc_Address + 0x4);
				if ((i ^ 0x55) != j) 
				{	
					f->f_results[0] = Amcc_Address + 0x4;
					f->f_wstring[0] = i ^ 0x55;
					f->f_wstring[1] = j;
					POST_ERROR(f, BAD_OMB);
				}
				
				j = INB(Amcc_Address + 0x8);
				if ((i ^ 0xaa) != j) 
				{	
					f->f_results[0] = Amcc_Address + 0x8;
					f->f_wstring[0] = i ^ 0xaa;
					f->f_wstring[1] = j;
					POST_ERROR(f, BAD_OMB);
				}
				
				j = INB(Amcc_Address + 0xc);
				if ((i ^ 0xff) != j) 
				{	
					f->f_results[0] = Amcc_Address + 0xc;
					f->f_wstring[0] = i ^ 0xff;
					f->f_wstring[1] = j;
					POST_ERROR(f, BAD_OMB);
				}
		}
	
		f->f_passfail++;
	} while (--passes);
	
	SUCCEED(f);
}
