/*-------------------------------------------------
 * i2dghw.h
 * 
 * Interfaces to diagnostic routines found in i2dghw.c
 * (q.v. for revision notes)
 * 
 * Format note: Set tabstops to 4 for correct appearance.
 * vi:  ts=4
 * pr:  -e4
 *-------------------------------------------------
 */

#ifndef DIAGH_H		/* To prevent multiple includes */
#define DIAGH_H	1

/*---------------------------------------------------------
 * Revision History:
 * 
 * 10/22/91		MAG		Major re-work begun to reflect
 *						changes for expandable boards, etc.
 * 4/1/92		MAG		Additions to support EISA
 * 6/24/93		MAG		Made _results smaller to get more space
 * 1/23/97		MAG		Additions to support 654's
 *---------------------------------------------------------
 */

#include "i2ellis.h"



/*----------------------------------------------------
 * --- FUNK ---
 * The structure  _funk is used to pass options to the
 * individual diagnostic routine, and to return the
 * results from it. 
 *----------------------------------------------------
 */

/* Some sizes used in _funk */

#define NP 8		/* Number of parameters in array */
#define NR 20		/* Number of results in array */
					/* interrupt test relies on this to
					 * be at least 16
					 */
#define NS 100		/* Number of characters in test strings */
					/* Assumed at least 64 */


typedef struct _funk {
							/* Input parameters -- Pass information TO
							 * the test routine 
							 */
	i2eBordStrPtr	f_pB;	/* Pointer to the board structure for the
							 * board to be tested here.
							 * This lets the test routine get to all
							 * sorts of information about the board.
							 */
	ushort f_passes;		/* Number of passes to perform */
	ushort f_interrupts;	/* Bit-map of interrupts to test (if applicable)*/
							/* (See manifests IRQ_3 through IRQ_15) */
	ushort f_mainint;		/* Interrupt to use for regular testing: */
							/* not bitmapped: IRQ3 = 3, IRQ15 = 15 ...*/
	ushort f_channels[4];	/* Bitmap of channels to test (if applicable)*/
							/* Up to 16 channels per box, boxes 0-3, 
							 * on non-expandable, use "box 0".  Bit map
							 * is (1 << channel) for channel 0 - 15
							 */
	ushort f_baud;			/* Single baud rate, if applicable  to test */
	ushort f_params[NP];	/* Any additional, test-specific params */

							/* Output parameters: Return information FROM
							 * the test routine. */
	ushort f_error;			/* gives result of test: OKAY = passed */
							/* otherwise possible error conditions 
							 * depend on the test -- see below */
	ushort f_passfail;		/* Reports on which pass failure occurred */
	ushort f_results[NR];	/* Any additional, test-specific results */
	uchar  f_wstring[NS];	/* Used in some loopback tests: data written */
	uchar  f_rstring[NS];	/* Used in some loopback tests; data read */

}funk, *pfunk;



/*-------------------------------
 * ERROR CODES:
 * Returned in f_error 
 *--------------------------------
 */

#define OKAY 			0 
#define MW_BAD_MEM_SIZE	1 
#define MW_MEM_FAIL		2 
#define MR_BAD_MEM_SIZE	3 
#define MR_MEM_FAIL		4 
#define FS_AF_AE 		5 
#define FS_MASK			6 
#define FS_AF_AE_D		7 
#define FS_MASK_D		8 
#define FS_BYTE_D		9 
#define FS_MAILBOX_D	10 
#define FDMA_XTIMO		11
#define FLL_DATA_ERROR	12 
#define FLL_DATA_ERROR16_EZ	13
#define FLL_DATA_ERROR16	14 
#define FM_TIMEOUT_IN 	15 
#define FM_TIMEOUT_OUT	16 
#define FM_DATA_IN		17 
#define FM_DATA_OUT		18 
#define RT_FAIL 		19 
#define IT_FAIL			20 
#define FDMA_RTIMO		21
#define MW_TIMEOUT1		22 
#define MW_TIMEOUT2		23 
#define MR_TIMEOUT1		24 
#define MR_TIMEOUT2		25 
#define FLL_TIMEOUT2	26 
#define FRL_TIMEOUT2	27 
#define FL_BAD_SIZE		28 
#define FV_BAD_SIZE		29 
#define FLL_GENERAL		30 
#define FRL_GENERAL		31 
#define RT_GENERAL		32 
#define FDMA_MISM		33 
#define IT_GENERAL		34 
#define LOOPBACK_ERR	35 
#define FV_NODRAIN		36
#define FDMA_HDATA		37
#define FDMA_HMAIL		38
#define BAD_BAUD_ERR	39
#define BAD_POS_2		40
#define BAD_POS_3		41
#define BAD_MC			42
#define BAD_PCI			43
#define BAD_OMB			44


/*-------------------------
 * ADDITIONAL RESULTS
 * Returned in f_results[0]
 * Grouped by f_error value
 *--------------------------
 */

/* Possible f_results[0] returns from the fifo sanity check: 
 * Currently, only NON_EXPANDABLE can be returned.  The only way to
 * distinguish expandable from non-expandable product is through 
 * reading the power-up message (which begs the question, in the context
 * of this test.
 * Actually, for future development, the TESTER should specify which
 * kind of product he thinks he is testing.
 */

#define NON_EXPANDABLE 1
#define EXPANDABLE 2


/* f_error = RT_FAIL
 * Indicates on-board device test failure 
 */

#define DEVICE_1400 1
#define DEVICE_8259 2
#define DEVICE_80186 3
#define DEVICE_2692 4
#define DEVICE_MAPPER 5
#define DEVICE_LCD 6
#define	DEVICE_654 7

/* f_results[1] gets this to indicate type of 1400 register that failed */

#define GLOBAL_REG 1
#define CHANNEL_REG 2

/* These may be OR'ed with the Interrupt selection bits
 * given below to indicate interrupt test failure.
 */

#define LOST_IRQ		(1U << 15)
#define TWO_MANY_IRQS	(1U << 14)

/*---------------------------
 * INTERRUPT SELECTION
 * Also returned by r_results[0] to
 * indicate which interrupts failed.
 * Also supplied by f_interrupts
 *---------------------------
 */

#define IRQ_3	(1U << 0)
#define IRQ_4	(1U << 1)
#define IRQ_5	(1U << 2)
#define IRQ_7	(1U << 3)
#define IRQ_10	(1U << 4)
#define IRQ_11	(1U << 5)
#define IRQ_12	(1U << 6)
#define IRQ_15	(1U << 7)

/*-----------------------------
 * Prototypes for functions in
 * i2dghw.c
 * ----------------------------
 */

int memory_walk (pfunk);
int memory_rand (pfunk);
int fifo_sanity (pfunk);
int fifo_verify_size(pfunk);
int fifo_local_stress(pfunk);
int fifo_block_loopback (pfunk);
int fifo_isolate_loop (pfunk);
int fifo_mailbox (pfunk);
int fifo_dma_loopback(pfunk);
int register_test (pfunk);
int register_exercise (pfunk);
int interrupt_test (pfunk);
int lcd_interactive (pfunk);

int pci_checkout(pfunk);
int pos_checkout(pfunk);
int validMCA(void);
int findMCAslot(int *, int *);

int	validEISA(void);
int	findEISAslot(int *, int *);

int biosPCIpresent(void);
int	findPCIslot(int *, int *);

/*------------------------
 * references to i2diag.c
 *------------------------
 */

#define V_FIFO	0x200
#define V_STARTUP 0x80
#define MCA_OPT 1
#define EISA_OPT 2
#define PCI_OPT 3

extern int mainVerbosity;
extern int slotMorE;
extern int hasUsedB;
extern int busType;


#endif /* defined DIAGH_H -- prevents multiple includes */
