 /*------------------------
 * i2diag.c
 * DOS-based diagnostic for 
 * Intelliport-II boards
 *-------------------------
 */

/*-------------------------------------
 * Cosmetic note:
 * Written using tabstops at 4 spaces
 * vi: set ts=4
 * pr: -e4
 *-------------------------------------
 */

/*-------------------------------------------------------
 * Revision History:
 *
 *  6/03/91		MAG	First Draft
 *					Only supports ISA-8/ISA-4 boards
 *  6/05/91		MAG	...continuing development through...
 *  6/15/91		MAG ...continuing...
 *					Removed 115200 rate: board not guaranteed to
 *					not lose characters when doing this rate.
 *
 *  6/19/91		MAG	0.02	Bug fixed: all selected channels will
 *							receive command to enable DSS reporting,
 *							etc.
 *							Separate continuation bits for each loopback type
 *							After data blast test, reads back stuff
 *							During DTR/RTS blasts, output data artificially
 *							paced to prevent lots of commands backing up
 *							on the board. (Since flush only flushes data)
 *
 *  6/19/91		MAG	1.00	Final cosmetic changes before production release
 *							Added an option to suppress operator prompts.
 *				
 *  6/24/91		MAG	1.01	Re-compiled with updated i2lib.c: the
 *							former library fails on 4-port boards
 *							because flow-control packets for non-existant
 *							channels were sent at initialization time.
 * 6/25/91		MAG	1.02	Changes to diagc.c (q.v.).  Closes *.LOD file
 *							after finished with it.
 *							Ensures we wait long enough after commands
 *							that the data writes in loopbacks can work
 * 6/30/91		MAG	1.03	...looking for bugs.
 *							... adding function calls in all loops to ensure
 *							ctrlC/break will be caught...
 *							delay(10) added to syncCommand(), bypassCommand()
 *							
 * 7/08/91		MAG 1.10	New, more thorough fifo test added,
 *							tests re-numbered to put a RAM test earlier,
 *							and keep the FIFO tests together.
 *							(LOOK OUT, *.BAT FILES -- tests renumbered)
 *							"Firmware" now called "loadware".
 *							Sync commands combined for fewer calls.
 *							
 * 7/17/91		MAG	1.10.1	Changed load file INTT.LOD and in DIAGC.C
 *							on host side).  Previous version would not
 *							check mailbox before writing it.  Yet host might
 *							send data asking for multiple interrupts before
 *							the board manages to issue one... Result, 
 *							missing interrupts. Change to DIAGC.C is to
 *							give just a little extra time for the last
 *							interrupt to be processed.
 * 7/19/91		MAG 1.10.2	Changed diagc.c to fix interrupt tester:
 *							IT_FAIL too soon created spurious failures.
 * 9/12/91		MAG	1.10.3	Increased amount of time delay after reset before
 *							attempting to load standard loadware.
 * 9/22/91		MAG	1.10.4	Fixed bug whereby we were selecting the wrong
 *							interrupt (always irq15) for the high-level tests.
 *							(ISA-4/8 worked ok since command to board was also
 *							wrong.
 * 10/22/91		MAG	2.00.0	Major modifications begun for the expandable board.
 * 12/12/91		MAG 2.00.4	MicroChannel support given as well.
 * 12/13/91		MAG 2.01.0	POS test added for MicroChannel
 *							more kbhit() calls added for better break response
 *							if break=on isn't in config.sys
 * 12/16/91		MAG 2.01.1	DTR/DSR flow control test was always skipped.
 *							Mis-spelling fixed.
 * 12/17/91		MAG 2.01.2	Help-screen info about loopback plugs updated.
 * 12/19/91		MAG 2.01.3	Loopback select slower baud rates now.
 *							DTR/DSR flow control test removed.  There is
 *							no such thing as DSRFLOW.
 * 12/20/91		MAG 2.01.4	Loopback tests can have reasonably fast rates
 *							unless running in torture mode.  Then we slow
 *							them down.
 * 2/12/91		MAG 2.01.5	',' command line argument also pauses after
 *							DSS signal error reports. (bug: 2.01.4 didn't)
 *							Bug fixed in DSS loopback diagnostics: former
 *							versions did not ensure the DSS under test 
 *							was stable before pulsing it, therefore we
 *							could think DSS had gone high as a result of
 *							the command we just did, but really it was
 *							ALREADY high: we haven't reacted to the command
 *							to lower it, even. 
 *							-- To speed up DSS testing, new method is used.
 *							For both high and low values of the "other"
 *							output, we set all the test outputs to a
 *							a random set of values (assert/negate).  Then we
 *							check all the inputs to see they match.  Then
 *							we complement all the test outputs, and re-test
 *							inputs.
 *							This reduces the number of round trips, and
 *							very much speeds testing time.
 *							See genericDss... for details.
 * 3/04/92	MAG	2.01.6		New argument: each Z doubles the time of delays
 * 3/09/92	MAG	2.01.7		Additional 1/2 second delay added after downloading
 *							standard (only the standard) loadware, to allow
 *							initial commands to be drained.  Otherwise, the
 *							DSS loopback test (for example) might not allow 
 *							enough time for the commands to be processed.
 * 3/24/92	MAG	2.01.8		The FIFO Mailbox test is changed:  The board 
 *							now complements the data before writing it back.
 *							Then we know the board handled it, and we're
 *							not just looking at our own outbound mailbox.
 * 3/24/92	MAG	2.01.9		If POS test fails, restores normal context
 *							before returning from test.
 * 4/01/92	MAG 2.02.0		EISA support needed.
 * 4/03/92	MAG	2.02.1		Continuation: ready to use now...
 *							If 'be' argument used, will scan slots for
 *							a board matching the specs.  
 *							Interrupt test now issues (4 * passes) interrupts
 *							to each selected line.  Test for ISA made 
 *							stronger as well.
 *							Also a major change: When testing EISA or
 *							micro channel, we don't need to specify the
 *							address.  If we do, it will be verified.  If
 *							not, the first board found will be tested.
 *							If interrupt test is used, must use I; for
 *							high level tests, if A is omitted then I may be.
 *							New limits for interrupt test: default 1000,
 *							limit 15000 (because of reporting irqs).
 * 6/02/92	MAG	2.02.2		If 'L' and ',' options both appear, we pause
 *							for a keystroke after downloading the loadware,
 *							then continue.
 * 6/05/92	MAG 2.02.3		Data echo test added.
 * 6/11/92	MAG	2.02.5		Error log option added(.4), minor cosmetic
 *							adjustments added.  G{file}
 * 7/23/92	MAG	2.02.6		Additional test added (#7: register exerciser)
 * 11/16/92 MAG ---			Brackets added to test initializations.
 * 3/15/93	MAG	2.02.07		Interrupt test fixed so it will fail when
 *							interrupts have been missed and "fail-safed".
 *							(Still untested for robustness as of 3/15...)
 * 5/03/93	MAG	2.02.08		If you use the '+' option, the data for the 
 *							FIFO block command is forced.  (Entered as a
 *							16-bit word).
 * 6/24/93	MAG 2.02.09		Some arrays made smaller to save space.
 *							Bug (in i2lib.c) fixed, were getting failures
 *							when using "L ," options with high-level tests.
 *							Wasn't bumping number of boards before interrupt
 *							was enabled.
 * 6/25/93	MAG	2.03.0		Changes to fatal-error detection include
 *							a sanity test in i2lib for too-big counts.
 * 7/08/93	MAG	2.03.1		Changes to recognize the EISA-CEX version "282",
 *							and "283" (for EISA-8 with asic).
 *							which support also edge-triggered interrupts.
 * 8/06/93	MAG 2.03.2		Re-linked with dev-kit changes (see i2ellis.c)
 *							to support ISA-4 (asic).  Former versions do
 *							not understand that the board is supposed to
 *							have only 4 ports, because of Power-up message
 *							change for this board.
 * 8/20/93	MAG	2.03.3		The W option was added: on resetting the
 *							board, if no channels are reported, you 
 *							must reset and try again.
 *							The number of (separate!) W arguments determines
 *							how many boxes need to be reported.
 *							ALSO, if W option is used, it will pause 
 *							after each major pass completed for a CR.
 * 8/26.93	MAG	2.03.4		Needed because of W option: If W option is used,
 *							then in between the major passes we will clear
 *							out the continuation flags for each channel. These
 *							would normally be set when a channel fails to
 *							prevent further passes from retesting, and failing,
 *							and so on.
 * 12/22/93	MAG	2.03.5		Additional info printed if the power-on reset mess.
 *							isn't found: to rule out s/w bugs...
 * 01/20/94 MAG 2.04.1		Used the clock() routine which returns time in
 *							approximately 55 mS increments to validate the
 *							delay() routine, to prevent it from returning
 *							too soon.
 * 02/17/94 MAG 2.04.2		Timeout limits increased for the memory tests 
 *							(see i2dghw.c) ISA-8 ASIC iff 512K might take 
 *							a long time at 16MHz processor.
 * 03/31/95	MAG	2.04.3		Memory diagnostics were reporting incorrect
 *							values read back.
 * 04/11/95 MAG 2.04.4		Removed requirement that address <= 3ff
 *							(A temporary expedient: PCI support needed soon)
 * 06/15/95	MAG	2.05.0		Support for PCI bus added for real.
 * 06/22/95 MAG 2.05.1		Special interrupt test added for PCI: on the
 *							host side tests only the interrupt that the
 *							board is configured for, but the board must
 *							generate them all by "trying" to generate
 *							each kind: that proves the CEX asic is
 *							wired properly.
 * 08/03/95 MAG 2.05.2		If you use the I argument on PCI interrupt
 *							test, only that interrupt line from the ASIC
 *							will be used.  If you don't use it, all
 *							lines will be tested. (In either case, the
 *							actual interrupts are serviced on the single
 *							IRQ for which PCI is configured.)
 *							
 * 08/10/95 MAG 2.05.3		Added test 27, like test 26 but
 *							DTR and RTS are held low.
 * 08/11/95 MAG 2.05.4		Added test 31, a "leak test" which
 *							fails IFF any port under test _has
 *							data looped back to it.
 * 04/16/96 MAG 2.05.5		X option = persist in major passes
 *							even if there is an error.
 * 01/23/97	MAG	3.00.0		Support for ST654's added.
 * 03/10/98 MAG 3.01.0		Released with loadware to support the
 *							ISA-8 using 654's (new ARR values needed).
 * 06/24/99 MAG 3.02.0		Released with loadware 2.1.1 to support
 *							Intelliport Plus ISA-4 (654's). New loadware
 *							for register test also included.
 *-------------------------------------------------------
 */

#define VERSION "3.02.0"

#include <stdio.h>
#include <stdlib.h>
#include <alloc.h>
#include <dos.h>
#include <mem.h>
#include <conio.h>
#include <sys\stat.h>
#include <io.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;



#include "i2dghw.h"	/* struct funk definitions */
#include "i2lib.h"	/* basic library routines, commands, etc */
#include "i2file.h"	/* routine to download from file */


/*==========================================
 * Structures used locally
 *==========================================
 */

/*-----------------------------------
 * Test-Definition structure
 *-----------------------------------
 */

typedef struct _testdef {
	int (* pTest[2])(pfunk);	/* Two pointers to test functions:
								 * [0] for non-expandable,
								 * [1] for expandable.
								 * NULL -- this test does not apply to
								 * this kind of board.
								 */
	

	void (* pResults[2])(pfunk,struct _testdef *);
								/* Pointers to function to display results:
								 * [0] for non-expandable
								 * [1] for expandable
								 */

	char *loadName[2];			/* Name of loadware to download:
								 * [0] for non-expandable,
								 * [1] for expandable.
								 * NULL -- no loadware used
								 */
 
	char *testName;				/* Name of the test */
	char *testDesc;				/* Description of test */
	ushort sequence;			/* Order listed on command line */
								/* 0 = not selected */
	ushort defPasses;			/* Default passes if not specified */ 
	ushort maxPasses;			/* Maximum number of passes (silently) */
	
	ushort flags;				/* Option flags */
} testdef, *ptestdef;


/* ptestdef->flags */

#define E_FATAL		1		/* If any error, stop testing */
#define NEED_ARG_E	2		/* Is too primative a test to "sniff" whether
							 * this is an expandable or non-expandable board.
							 * Therefore cannot be run unless an argument
							 * was given telling which to assume.
							 */
#define E_RESERVED	4		/* This test number is reserved: skip it */
#define REAL_LW		8		/* Uses standard loadware: no need to 
							 * re-load between tests.
							 */

#define DIAG_LW		0x10	/* Uses "diagnostic" loadware which must be
							 * re-loaded before each test.
							 */

#define PLUG1		0x20	/* Needs TX-RX , DTR-DCD, RTS-CTS */
#define PLUG2		0x40	/* Needs DTR-DSR, RTS-RI */
#define PLUG3		0x80	/* Needs TX-RX, DTR-CTS, RTS-DCD */
#define PLUG_BITS	(PLUG1 | PLUG2 | PLUG3)

#define PERCHANNEL	0x100	/* tests individual channels */
#define DEBUG_NEED	0x200	/* Requires the debug port to have been connected */
#define PERINTERRUPT 0x400	/* Uses the interrupt bitmap to select interrupts */
#define TIMED_TEST	0x800	/* Passes are seconds */
#define NEEDS_INPUT	0x1000	/* Needs operator input */
#define DIAG_RESET	0x2000	/* Needs us to reset the board first */
#define CANNOT_FAIL	0x4000	/* Indicates an exercise test: no checking */




/* -----------------------------------------
 * Defines for per-channel error conditions
 * -----------------------------------------
 */

/* These inhabit f_rstring[channel] when f_error == LOOPBACK_ERR */

#define L_ERR_NOXMT		1
#define L_ERR_LATE		2
#define L_ERR_WRONG		3
#define L_ERR_DCDDN		4
#define L_ERR_DCDUP		5
#define L_ERR_CTSDN		6
#define L_ERR_CTSUP		7
#define L_ERR_BREAKS	8
#define L_ERR_DSRDN		9
#define L_ERR_DSRUP		10
#define L_ERR_RIDN		11
#define L_ERR_RIUP		12
#define L_ERR_GOTDATA	13


/*--------------------------------------
 * Structure for controlling channel- 
 * based tests, and related
 *--------------------------------------
 */

typedef struct _chControl {
	ushort flags;			/* See below */
	i2ChanStrPtr c_pCh;		/* Pointer to channel structure */
	ushort nOut;			/* Number of characters to send */
	ushort iOut;			/* Index to next character to send */
	uchar charsOut[128];	/* String of characters to send out */ 
	ushort nIn;				/* Number of characters read in */
	uchar charsIn[128];		/* String of characters read in */ 
	ushort parity;			/* Index to parity selection table */
	ushort stops;			/* Index to stop-bit selection table */
	ushort baud;			/* Index to baud rate selection table */
	ushort csize;			/* Index to character size selection table */
	ushort flow;			/* Index to flow-control array */
	ushort passes;			/* Total no. passes completed on this channel */

} chControl, far *fpchControl;

#define thisChan c_pCh->infl.hd.i2sChannel

/* Bits for flags */

#define CONTINUE_DATA_1 1
#define CONTINUE_DATA_2 2
#define CONTINUE_DATA_3 4
#define CONTINUE_DATA_4 8
#define CONTINUE_CTS 0x10
#define CONTINUE_DCD 0x20
#define CONTINUE_DSR 0x40
#define CONTINUE_RI 0x80
#define CONTINUE_BREAKS 0x100
#define CONTINUE_DCD_X	0x200
#define CONTINUE_CTS_X	0x400
#define CONTINUE_DATA_5	0x800
#define CONTINUE_DATA_6	0x1000

#define CONTINUE_ANY (CONTINUE_DATA_1 | CONTINUE_DATA_2 | CONTINUE_DATA_3 | \
					 CONTINUE_DATA_4 | CONTINUE_CTS | CONTINUE_DCD | \
					 CONTINUE_DCD_X | CONTINUE_CTS_X | \
					 CONTINUE_DSR | CONTINUE_RI | CONTINUE_BREAKS | \
					 CONTINUE_DATA_5 | CONTINUE_DATA_6)

#define PROVISIONAL 0x2000	/* Set if channel failed first half of dss loopback */

#define SPECULATION 0x4000	/* Set if channel was not selected, but we
							 * want to test this channel provided it
							 * exists.
							 * Should we determine later that such a
							 * channel doesn't exist, no sweat.
							 */

/*---------------------------------------------------------
 * Structure to select protocol parameters
 *---------------------------------------------------------
 */

typedef struct _proTable {
	char *proRecog;	/* Code to use on command line, etc... */
	char *proDesc;	/* Description of protocol */
	char proParm;	/* The protocol parameter to use */
} proTable, *pproTable;





/*-------------------------
 * PROTOTYPES
 *--------------------------
 */

/* Declarations for all tests in this module */

int interpret_po(pfunk);
int internal_loop(pfunk);
int dtr_dcd_loop(pfunk);
int rts_cts_loop(pfunk);
int dtr_cts_loop(pfunk);
int rts_dcd_loop(pfunk);
int dtr_dsr_loop(pfunk);
int rts_ri_loop(pfunk);
int data_loop1(pfunk);
int data_loop2(pfunk);
int data_loop3(pfunk);
int data_loop4(pfunk);
int data_loop5(pfunk);
int data_blast(pfunk);
int data_echo(pfunk);
int data_rate_test(pfunk);
int wy60_blast(pfunk);
int link_mc6_blast(pfunk);
int variable_blast(pfunk, int);
int dtr_blast(pfunk);
int rts_blast(pfunk);
int break_loop(pfunk);

/* Declarations for other routines here */

void usage(void);
void usage_bad(char *);
void usage_h(void);
int getNumeric(char *);
int irqToBit(int);
int findCharacter(char *, char);
void runTimeInitialize(void);
void parseArguments(int, char **);
void ltrError(char *, char *);
void operatorHit(int);
void holdIfNeeded(void);
int breakHandler(void);
void rip(int, char *);
void allocChannelTable(void);
void printChanMap(ushort *);
void endSpeculation(void);
void validateChannels(void);
void assertProtocol(fpchControl, i2ChanStrPtr);
void printProtocol(fpchControl);
void printProtocolGlobal(fpchControl);
void chooseProtocol(fpchControl,ushort);
int stringToIndex(int, pproTable, char *);
void printOutputHelps(void);
void printProtocolHelps(int, pproTable, char *);
void showDiscreprencies(fpchControl, int, char *, int, uchar far *, int, uchar far *);
int farCompare(int, uchar far *, uchar far *);
int mapSpeedToSize(int);
int loopBack(pfunk, ushort, ushort);
int leekTest(pfunk);
void sendEnables(void);

void commandToAll(int, cmdSyntaxPtr, int);
void commandToOne(int, i2ChanStrPtr, cmdSyntaxPtr, int);
void sensitiveDelay(int);
void flushAllInput(void);

int eprintf(char *,...);

static int genericDssLoop(pfunk, cmdSyntaxPtr, cmdSyntaxPtr,
		int, char *, char *, int, int, int);
static int waitForDssInput(int, int, i2ChanStrPtr);
static void polyDrain(void);
static void initChannelTable(unsigned short);

/* Two Unpublished routines from i2lib.c */

void iiSendPendingMail(i2eBordStrPtr);
void serviceOutgoingFifo(i2eBordStrPtr);	


/* Declarations for all the reporting functions */

void r_goodnul(pfunk,ptestdef);
void r_goodfsize(pfunk,ptestdef);
void r_goodmem(pfunk,ptestdef);
void r_goodint(pfunk,ptestdef);

void reportFailure(pfunk, int, char *);

/* Declarations for error functions */

void erReportPass(pfunk);
void erReportByteError(pfunk);
void erReportWordError(pfunk);
void erReportFsize(pfunk);
void erReportFVsize(pfunk);

void erOKEY(pfunk);
void erDUMB(pfunk);
void erMEMVAL(pfunk);
void erMEMFAIL(pfunk);
void erPCIFAIL(pfunk);
void erFIFO(pfunk);
void erFTIME(pfunk);
void erFDATA(pfunk);
void erFDATA16(pfunk);
void erFTDATA(pfunk);
void erFMDATA(pfunk);
void erREGFAIL(pfunk);
void erINTERRUPT(pfunk);
void erSTRANGE(pfunk);
void erFSTRANGE(pfunk);
void erLOOPBACK(pfunk);

/* Declarations for new error-reporting functions */

void boardErrorReport(i2eBordStrPtr);
void boardMessReport(i2eBordStrPtr);
void fullDownloadFromFile(i2eBordStrPtr, char *, int);
void printIrqMap(pfunk pf);
void xlatePlugCode(int);
void fatalTrap(i2eBordStrPtr pB);



/*-------------------------------------
 * GLOBALS
 *-------------------------------------
 */


#define VALID_VECTORS 8

int validIrqList[VALID_VECTORS] = {3, 4, 5, 7, 10, 11, 12, 15};


ushort nfpchTable = 0;			/* Number of entries in the following: */
ushort speculate = 0;			/* Are we speculating on channels? */
ushort validated = 0;			/* Have we validated selected channels yet? */

ushort Persist = 0;

fpchControl fpchTable = (fpchControl) 0; /* Points to an array of channel-test
										  * control structures.
										  */


#define N_CHANNELS 64

i2eBordStr	boardInfo[1];	/* Control structure for board */
i2ChanStr	chanInfo[N_CHANNELS];	/* Channel info, for up to 64 channels */



/*-------------------------------------
 * Table of test definitions
 *-------------------------------------
 */

#define NUMBER_OF_TESTS 53	/* Update if many more tests are added */

/* Here is an array of parameter structures for each test */

funk argumentList[NUMBER_OF_TESTS];


/* Here is an array for each test there can be */

testdef testList[NUMBER_OF_TESTS + 1] = {

/*--------------- TESTS 1 - 9 -------------------- */
/* are misc ones not normally used in checkout */

/* Test 1 */
{
{ fifo_sanity,	NULL},
{ r_goodnul,		r_goodnul},
{ NULL,			NULL},
"Fifo Sanity",
"Host checks fifo register reliability",
0,
1,1,
E_FATAL | NEED_ARG_E | DIAG_RESET
},

/* Test 2 */
{
{ interpret_po,	interpret_po},
{ r_goodnul,		r_goodnul},
{ NULL,			NULL},
"Power-on Reset message",
"Host interprets power-on reset message",
0,
1,1,
DIAG_RESET | CANNOT_FAIL
},

/* Test 3 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"reserved",
"reserved",
0,
1,1,
E_RESERVED
},

/* Test 4 */
{
{ NULL,					lcd_interactive},
{ r_goodnul,				r_goodnul},
{ "regl.lod",				"regl.lod"},
"Interactive LCD Controller",
"Operator validates correct LCD controller operation",
0,
1,1,
E_FATAL | DIAG_LW | CANNOT_FAIL
},

/* Test 5 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"reserved",
"reserved",
0,
1,1,
E_RESERVED
},

/* Test 6 */
{
{ fifo_verify_size,	fifo_verify_size},
{ r_goodfsize,	r_goodfsize},
{ "null.lod",		"null.lod"},
"Fifo Measurement",
"Host measures fifo size",
0,
1,1,
E_FATAL | DIAG_LW | CANNOT_FAIL
},

/* Test 7 */
{
{ register_exercise,	register_exercise},
{ r_goodnul,		r_goodnul},
{ "regt3.lod",	"regt3.lod"},
"External Bus Exerciser",
"Writes & reads devices on external bus",
0,
2, 30000,
DIAG_LW | TIMED_TEST | CANNOT_FAIL
},

/* Test 8 */
{
{ register_test,		register_test},
{ r_goodnul,			r_goodnul},
{ "regt2.lod",		"regt2.lod"},
"Debug-Port Register",
"Verifies access to debug port registers",
0,
20,30000,
DIAG_LW | DEBUG_NEED
},

/* Test 9 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"reserved",
"reserved",
0,
1,1,
E_RESERVED
},


/*----------- Tests 10-24: Low level tests ----------- */

/* Test 10 */
{
{ fifo_mailbox,	fifo_mailbox},
{ r_goodnul,		r_goodnul},
{ "fifom.lod",	"fifom.lod"},
"Fifo Mailbox",
"Host checks mailbox data from board & vice versa",
0,
40,30000,
E_FATAL | DIAG_LW
},

/* Test 11 */
{
{ interrupt_test,		interrupt_test},
{ r_goodint,			r_goodint},
{ "intt.lod",			"intt.lod"},
"Interrupt",
"Board generates selected interrupts to host",
0,
1000,15000,
E_FATAL | DIAG_LW | PERINTERRUPT
},

/* Test 12 */
{
{ memory_rand,	memory_rand},
{ r_goodmem,		r_goodmem},
{ "memr.lod",		"memr.lod"},
"Random-pattern Board Memory",
"Board performs random-pattern test on its local memory",
0,
5,30000,
E_FATAL | DIAG_LW
},

/* Test 13 */
{
{ fifo_isolate_loop,	fifo_isolate_loop},
{ r_goodnul,				r_goodnul},
{ "fifoi.lod",			"fifoi.lod"},
"FIFO Bi-Directional Loopback",	
"Host writes, reads data; board loops it back; both check integrity",
0,
250,30000,
E_FATAL | DIAG_LW
},

/* Test 14 */
{
{ fifo_local_stress,		fifo_local_stress},
{ r_goodnul,				r_goodnul},
{ "fifol.lod",			"fifol.lod"},
"Host Interface Timing Stress FIFO Loopback",
"Host writes, reads, checks fifo data; (board loops it back)",
0,
200,30000,
E_FATAL | DIAG_LW
},

/* Test 15 */
{
{ fifo_block_loopback,	fifo_block_loopback},
{ r_goodnul,				r_goodnul},
{ "fifob.lod",			"fifob.lod"},
"FIFO Block-Message Loopback",
"Host writes, reads, checks variable-length blocks of data",
0,
300,30000,
E_FATAL | DIAG_LW
},

/* Test 16 */
{
{ register_test,		register_test},
{ r_goodnul,			r_goodnul},
{ "regt.lod",			"regt.lod"},
"Board/Box Hardware Register",
"Board checks Hardware Register reliability",
0,
5,30000,
E_FATAL | DIAG_LW
},

/* Test 17 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 18 */
{
{ NULL,			fifo_dma_loopback},	
{ r_goodnul,		r_goodnul},
{ NULL,			"fifod.lod"},
"FIFO Loopback Using DMA",
"Board writes and reads FIFO using DMA, compares data",
0,
15,30000,
E_FATAL | DIAG_LW
},

/* Test 19 */
{
{ memory_walk,		memory_walk},
{ r_goodmem,			r_goodmem},
{ "memw.lod",			"memw.lod"},
"Walking-Bit Board Memory",
"Board performs walking-bit test on its local memory",
0,
1,30000,
E_FATAL | DIAG_LW
},

/* Test 20 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 21 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 22 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 23 */
{
{ pci_checkout,			pci_checkout},
{ r_goodnul,			r_goodnul},
{ NULL,			NULL},
"AMCC operation register checkout",
"Host ensures it can write and read AMCC interface registers",
0,
5,30000,
E_FATAL
},

/* Test 24 */
{
{ pos_checkout,	pos_checkout},
{ r_goodnul,		r_goodnul},
{ NULL,			NULL},
"POS register checkout",
"Host ensures it can write and read Micro Channel POS registers",
0,
5,30000,
E_FATAL
},


/* ---- Tests 25 - 34 ---- */
/* ---- Data Loopback tests --- */

/* Test 25 */
{
{ internal_loop,			internal_loop},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"Internal Data Loopback",
"Sends data to uart, receives, compares",
0,
3,30000,
REAL_LW | PERCHANNEL
},

/* Test 26 */
{
{ data_loop1,				data_loop1},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"External Data Loopback (DSS up)",
"Sends data to uart, through loopback plug, receives, compares",
0,
10,30000,
REAL_LW | PLUG1 | PERCHANNEL
},


/* Test 27 */
{
{ data_loop2,				data_loop2},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"External Data Loopback (DSS down)",
"Sends data to uart, through loopback plug, receives, compares",
0,
10,30000,
REAL_LW | PLUG1 | PERCHANNEL
},

/* Test 28 */
{
{ data_loop3,				data_loop3},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"External Data Loopback w/Data Set Exercise",
"Sends data to uart, through loopback, receives, compares, moves DTR & RTS",
0,
10,30000,
REAL_LW | PLUG1 | PERCHANNEL
},

/* Test 29 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 30 */
{
{ data_loop4,				data_loop4},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"External Data Loopback (RTS/CTS Flow Ctrl)",
"Sends data to uart, through loopback plug, receives, compares",
0,
20,30000,
REAL_LW | PLUG1 | PERCHANNEL
},

/* Test 31 */
{
{ leekTest,			leekTest},
{ r_goodnul,		r_goodnul},
{ "ff.lod",			"ff.lod"},
"Data Leak Test",
"Sends some data to UART, ensures that _none_ returns",
0,
1, 1,
REAL_LW | PLUG1 | PERCHANNEL
},

/* Test 32 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 33 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 34 */
{
{ data_loop5,				data_loop5},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"External Data Loopback (DTR/CTS Flow Ctrl)",
"Sends data to uart, through loopback plug, receives, compares",
0,
20,30000,
REAL_LW | PLUG3 | PERCHANNEL
},

/*------ Tests 35-44 ----- */
/* Other loop-back tests */

/* Test 35 */
{
{ break_loop,				break_loop},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"Break Generation/Detection",
"Sends breaks & tries to receive them",
0,
10,30000,
REAL_LW | PLUG1 | PERCHANNEL
},


/* Test 36 */
{
{ dtr_dcd_loop,			dtr_dcd_loop},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"DTR/DCD Loopback",	
"Ensures DCD tracks DTR changes on each channel",
0,
10,30000,
REAL_LW | PLUG1 | PERCHANNEL
},

/* Test 37 */
{
{ rts_cts_loop,			rts_cts_loop},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"RTS/CTS Loopback",
"Ensures CTS tracks RTS changes on each channel",
0,
10,30000,
REAL_LW | PLUG1 | PERCHANNEL
},

/* Test 38 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 39 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},


/* Test 40 */
{
{ NULL,					dtr_dsr_loop},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"DTR/DSR Loopback",	
"Ensures DSR tracks DTR changes on each channel",
0,
10,30000,
REAL_LW | PLUG2 | PERCHANNEL
},

/* Test 41 */
{
{ NULL,					rts_ri_loop},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"RTS/RI  Loopback",
"Ensures RI  tracks RTS changes on each channel",
0,
10,30000,
REAL_LW | PLUG2 | PERCHANNEL
},

/* Test 42 */
{
{ NULL,			NULL},
{ NULL,			NULL},
{ NULL,			NULL},
"res",
"res",
0,
1,1,
E_RESERVED
},

/* Test 43 */
{
{ dtr_cts_loop,			dtr_cts_loop},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"DTR/CTS Loopback",	
"Ensures CTS tracks DTR changes on each channel",
0,
10,30000,
REAL_LW | PLUG3 | PERCHANNEL
},

/* Test 44 */
{
{ rts_dcd_loop,			rts_dcd_loop},
{  r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"RTS/DCD Loopback",
"Ensures DCD tracks RTS changes on each channel",
0,
10,30000,
REAL_LW | PLUG3 | PERCHANNEL
},

/*------- Tests 45 - 54 ---------- */
/* Transmission only, tracing tests */

/* Test 45 */
{
{ data_blast,				data_blast},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"Data Transmission",
"Sends a string over and over to each selected channel",
0,
2,30000,
REAL_LW | PERCHANNEL | TIMED_TEST | NEEDS_INPUT | CANNOT_FAIL
},

/* Test 46 */
{
{ dtr_blast,				dtr_blast},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"DTR Tracing",
"Raises/Lowers DTR repeatedly on selected channels",
0,
2,30000,
REAL_LW | PERCHANNEL | TIMED_TEST | CANNOT_FAIL
},

/* Test 47 */
{
{ rts_blast,				rts_blast},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"RTS Tracing",
"Raises/Lowers RTS repeatedly on selected channels",
0,
2,30000,
REAL_LW | PERCHANNEL | TIMED_TEST | CANNOT_FAIL
},

/* Test 48 */
{
{ data_echo,		data_echo},
{ r_goodnul,		r_goodnul},
{ "ff.lod",		"ff.lod"},
"Data Echo",
"Reads and echos incoming characters",
0,
2,30000,
REAL_LW | PERCHANNEL | TIMED_TEST | CANNOT_FAIL
},

/* Test 49 */
{
{ data_rate_test,			data_rate_test},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"Baud Rate Accuracy",
"Sends 60 characters at 50 baud and times them",
0,
1, 1,
REAL_LW | PERCHANNEL
},

/* Test 50 */
{
{ wy60_blast,				wy60_blast},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"Variable Rate (Wy60) Data Transmission",
"Sends messages at 2400, 9600, 19200, 38400 baud",
0,
2,30000,
REAL_LW | PERCHANNEL | TIMED_TEST | CANNOT_FAIL
},

/* Test 51 */
{
{ link_mc6_blast,			link_mc6_blast},
{ r_goodnul,				r_goodnul},
{ "ff.lod",				"ff.lod"},
"Variable Rate (Link MC6) Data Transmission",
"Sends messages at 2400, 9600, 19200, 38400, 76800, 115200 baud",
0,
2,30000,
REAL_LW | PERCHANNEL | TIMED_TEST | CANNOT_FAIL
},

/* This indicates the end of the list of tests */
/* Test 52: */
{
{ 0, 0}, 
{ 0, 0}, 
{ 0, 0}, 
0, 0, 0, 0, 0, 0 
},
};



/*--------------------------------
 * Tables for protocol selection
 *--------------------------------
 */

/*-------------------------------------------------
 * For flow control there are two tables, since it
 * takes two sequences to handle both directions.
 * Also, other code relies on entry 0 being for no
 * flow control.
 *--------------------------------------------------
 */

#define N_FLOW 2
proTable p_flow1[N_FLOW] = {	/* CMD_IXON_OPT */
	"o", "      no", CIX_NONE,
	"x", "xon/xoff", CIX_XON,
};

proTable p_flow2[N_FLOW] = {	/* CMD_OXON_OPT */
	"o", "      no", COX_NONE,
	"x", "xon/xoff", COX_XON,
};

#define N_SIZE 2
proTable p_size[N_SIZE] = {		/* CMD_SETBITS */
	"7", "7 bits", CSZ_7,
	"8", "8 bits", CSZ_8,
};

#define N_STOPS 3
proTable p_stops[N_STOPS] = {	/* CMD_SETSTOP */
	"1",   "1   stop", CST_1,
	"1.5", "1.5 stop", CST_15,
	"2",   "2   stop", CST_2,
};

#define N_PARITY 5
proTable p_parity[N_PARITY] = {	/* CMD_SETPAR */
	"none",  "   no", CSP_NP,
	"odd",   "  odd", CSP_OD,
	"even",  " even", CSP_EV,
	"space", "space", CSP_SP,
	"mark",  " mark", CSP_MK,
};

#define N_UNIQ_BAUD 28
/* This table only used when hand-picking rates */
proTable p_baud[N_UNIQ_BAUD] = {	/* CMD_SETBAUD */
	"50",		"     50 ",	CBR_50,		/* 0 */
	"75",		"     75 ",	CBR_75,		/* 1 */
	"110",		"    110 ", CBR_110,	/* 2 */
	"134.5",	"    134+",	CBR_134,	/* 3 */
	"150",		"    150 ",	CBR_150,	/* 4 */
	"200",		"    200 ",	CBR_200,	/* 5 */
	"300",		"    300 ",	CBR_300,	/* 6 */
	"600",		"    600 ",	CBR_600,	/* 7 */
	"1200",		"   1200 ",	CBR_1200,	/* 8 */
	"1800",		"   1800 ",	CBR_1800,	/* 9 */
	"2000",		"   2000 ", CBR_2000,	/* 10 */
	"2400",		"   2400 ",	CBR_2400,	/* 11 */
	"3600",		"   3600 ", CBR_3600,	/* 12 */
	"4800",		"   4800 ",	CBR_4800,	/* 13 */
	"7200",		"   7200 ", CBR_7200,	/* 14 */
	"9600",		"   9600 ",	CBR_9600,	/* 15 */
	"19200",	" 19,200 ",	CBR_19200,	/* 16 */
	"38400",	" 38,400 ",	CBR_38400,	/* 17 */
	"56000",	" 56,000 ", CBR_56000,	/* 18 */
	"57600",	" 57,600 ", CBR_57600,	/* 19 */
	"64000",	" 64,000 ", CBR_64000,	/* 20 */
	"76800",	" 76,800 ", CBR_76800,	/* 21 */
	"115200",	"115,200 ", CBR_115200,	/* 22 */
	"153600",	"153,600 ", CBR_153600,	/* 23 */
	"230400",	"230,400 ", CBR_230400,	/* 24 */
	"307200",	"307,200 ", CBR_307200,	/* 25 */
	"460800",	"460,800 ", CBR_460800,	/* 26 */
	"921600",	"921,600 ", CBR_921600	/* 27 */
};

/* These are used when selecting random baud rates.
 * Two separate tables of indices are used to prevent
 * our randomly choosing too many channels running
 * at too high a rate.
 */

#define N_ULTRA_INDICES 8
unsigned char ultra_rates[N_ULTRA_INDICES] = {
22, 21, 21, 20, 19, 18, 17, 17 };

#define N_HIGH_INDICES 16
unsigned char high_rates[N_HIGH_INDICES] = {
17, 16, 16, 16, 16, 16, 16, 15,
15, 15, 15, 15, 15, 15, 14, 13 };

#define N_LOW_INDICES 32
unsigned char low_rates[N_LOW_INDICES] = {
 0,  1,  2,  3,  4,  5,  6,  7, 
 8,  9,  9, 10, 11, 11, 12, 12, 
12, 12, 13, 13, 13, 13, 13, 14,
14, 14, 14, 14, 15, 15, 15, 15 };

#define NO_HIGH_LIMIT	1000
#define LOOP_HIGH_LIMIT	7

int moreFromHigh = 0;		/* How many more can we have from the high list? */



/*--------------------------------------------------------------
 * Statics to store options more global than for specific tests 
 *--------------------------------------------------------------
 */

char *executableName = 0;	/* Pointer to 0th argument */
int mainPasses = 1;		/* Number of total passes */
int testSequence = 0;	/* Running count of test arguments */

int realTests = 0;		/* Initialized by run-time initialization */
int allFlags = 0;		/* The OR of all flags on all selected tests */
int plugPrompt = 0;		/* Set if the test must prompt for plug changes */
int mainInterrupts = 0;
int chiefInterrupt = 0;
int freezeFlow = -1;	/* -1 default indicates option not set */
int freezeBaud = -1;	/* -1 default indicates option not set */
int freezeParity = -1;	/* -1 default indicates option not set */
int freezeStops = -1;	/* -1 default indicates option not set */
int freezeSize = -1;	/* -1 default indicates option not set */
int highBaudOnly = 0;	/* set if we are doing HW flow control tests */


int interaction = 1;	/* Set indicates we want operator interaction */

/* These are to support the '+' argument */
int forceStringLength = 0;
#define MAX_FORCE 70
char argString[MAX_FORCE+2];
int argInt;				/* If the argument is a number, store it here */

#define	EXPAND_TYPE		1
#define NONEXP_TYPE		0

int isExpandable = -1;	/* -1 = Doesn't know whether expandable or not */
						/* EXPANDABLE_TYPE, NONEXP_TYPE */
int	is16bit = -1;		/* -1 = Doesn't care whether 16 or 8-bit interface. */
						/* 0 = 8-bit interface, 1 = 16-bit interface. */


int hasUsedE = 0;		/* true if Ex/En command-line argument appeared*/

int	testAddress = 0x308;	/* Default address to test */
int hasUsedA = 0;			/* True if the A command-line argument appeared */

int	logFd = -1;				/* If non-negative, the G option used and
							 * this is the file descriptor of the log file
							 */
int delayMult = 0;		/* Without Z argument, delay not shifted */
int	waitForChannels = 0;	/* With W argument, wait until more than 0 boxes */

#define ISA_OPT		0
#define MCA_OPT		1
#define EISA_OPT	2
#define PCI_OPT		3

int busType = ISA_OPT;	/* This is the default */
int hasUsedB = 0;		/* true if B command-line argument appeared */

int loadOnly = 0;		/* Load only, then stop! */

int slotMorE = 0;		/* Slot number information for EISA or
						 * Micro Channel.
						 * For Micro Channel:
						 * Bit-mapped: bits 2-0 are slot number 
						 * Bit 3: 1 (for setup enable) 
						 * Other bits: 0 
						 * This is the value which would be written to
						 * the microChannel setup register 0x96.
						 * For EISA:
						 * The base slot-dependant address.
						 */


time_t startTime;



/* Verbosity and output control */

#define V_NORMAL	1		/* Normal levels of output */
#define V_ERROR		2		/* Detailed error descriptions */
#define V_VERSION	4		/* Display version numbers */
#define V_PROTOCOL	8		/* Display protocol information */
#define V_TIME		0x10	/* Display testing time */
#define V_IRQ		0x20	/* Display IRQ map */
#define V_CHANNELS	0x40	/* Map of selected channels */
#define V_STARTUP	0x80	/* Display information at startup */
#define V_DEBUG		0x100	/* Debugging use */
#ifndef V_FIFO
#define V_FIFO	0x200	/* More detail in low-level routines */
#endif

struct vTypes {
	int vFlag;
	char *explan;
	char c1;
	char c2;
} vTarray[17] = {
	V_NORMAL,				"Normal messages",			'n', 'N',
	V_ERROR,		"Detailed Loopback Errors",	'e', 'E', 
	V_VERSION,		"Version numbers",			'v', 'V', 
	V_PROTOCOL,		"Protocols",				'p', 'P',
	V_TIME,			"Testing Time",				't', 'T', 
	V_IRQ,			"IRQ Map",					'i', 'I', 
	V_CHANNELS,		"Channel Map",				'c', 'C',
	V_STARTUP,		"Test information at startup",
														's', 'S',
	V_DEBUG,				"~Debugging",				'd', 'D',
	V_FIFO,			"Detailed FIFO errors",		'f', 'F',
	0,			"",					' ', ' ', 
};

int mainVerbosity = V_NORMAL;

int isHolding = 0;
int isFatal = 0;
int statParity = 0;		/* The MM option was used */
int checkParityOption = CSE_NOREP;



/* Additional statics */

#define POOLSIZE 256

/* Pools of "ramdom" test data */

uchar randomPool7[POOLSIZE];
uchar randomPool8[POOLSIZE];


/*-----------------------------------------------------
 * runTimeInitialize()
 * 
 * Some things are more practical to initialize this way
 * than at compile time
 *------------------------------------------------------
 */

void
runTimeInitialize()
{
register int i;
register ptestdef ptd;
register pfunk pf;

	for (i = 0, ptd = testList, pf = argumentList; 
		ptd->testName; i++, ptd++, pf++)
	{
		pf->f_pB = boardInfo;
		pf->f_passes = ptd->defPasses;
		pf->f_interrupts = 0;
		pf->f_mainint = 0;	/* Means, uninitialized yet */
		pf->f_channels[0] = pf->f_channels[1] =
			pf->f_channels[2] = pf->f_channels[3] = 0;
	}
	realTests = i;
	

	ctrlbrk(breakHandler);

	randomize();	/* Initialize random number generator */

	/* Initialize random character pool */
	for (i = 0; i < POOLSIZE; i++)
	{
		randomPool7[i] = random('z' - ' ') + ' ';
		randomPool8[i] = (rand() & 1) ?
			(random('z' - ' ') + ' ') :
			((random('z' - ' ') + ' ') | 0x80);
	}
}

/*------------------------------------
 * eprintf
 * A bit of a hack, I'm afraid, to support
 * an error log whose file descriptor might
 * be in logFd.  
 * This is a function that can be used in place of
 * printf for all displays that might go to the error log.
 * The catch is, it's hard to wrap printf because it is written\
 * for a variable number of arguments, depending on parameters in
 * the first argument.  Here we do not bother to decide how many
 * arguments we really have.  We just pass up to 8 parameters 
 * WHETHER THEY WERE PUSHED ON THE STACK OR NOT!  This means
 * garbage is being pushed sometimes, but the value will only
 * be used (by sprintf and printf) when it is good.
 *--------------------------------------
 */

int
eprintf(char *cp, ...)
{
char buffer[100];
int count;
register int *pi = (int *) &cp;

	count = sprintf(buffer, cp, 
			pi[1], pi[2], pi[3], pi[4], pi[5], pi[6], pi[7], pi[8]);
	if (logFd >= 0)
	{
		write(logFd, buffer, count);
	}
	return printf(buffer);
}

int
breakHandler()
{
	eprintf("\nOperator terminated by ^C\n");

	/* Any other conditional cleanup? */
	
	i2RestoreIrqs();

	return 0;	/* Means, we should abort */
}
		


/* This is the way to make unexpected exits rather than
 * using exit directly: guarantees to clean up after you
 */

void rip(exitValue, message)
int exitValue;
char *message;
{
time_t endTime;

	if (message != (char *) 0)
	{
		eprintf("%s\n",message);
	}
	i2RestoreIrqs();

	if (mainVerbosity & V_TIME)
	{
		printf("Total Testing time: ");
		endTime = time(NULL) - startTime;
		if (endTime >= 3600)
		{
			printf("%ld hour(s), ", endTime / 3600);
			endTime %= 3600;
		}

		if (endTime >= 60)
		{
			printf("%ld minute(s), ", endTime / 60);
			endTime %= 60;
		}
		printf("%ld second(s).\n", endTime);
	}	
		
	exit (exitValue);
}

/* Vanilla usage message prints very little so screen not lost */
/* Advises piping to 'more' to break into screens */

void
usage()	
{
	printf("For a list of command-line options, type\n");
	printf("%s help | more\n", executableName);
	printf("             or\n");
	printf("%s help > prn\n", executableName);
	printf("             to re-direct to the printer\n");
	rip(1, 0);
}

void
usage_bad(message)
char *message;
{
	printf("\n%s: %s\n\n",executableName, message);
	usage();
}

void
usage_h()
{
register ptestdef ptd; 
int i;

	printf("%s [options] [protocols] t{n} t{n}...\n", executableName);
	printf("[options] are one or more of the following:\n");
	printf("     P{n}    -- {n} is number of whole passes to do.\n");
	printf("                default = 1\n");
	printf("     X       -- continue P-option passes even if error occurs\n");
	printf("     A{n}    -- {n} is the base address of board (normally in hex).\n");
	printf("                default = 0x308\n");
	printf("     I{n}    -- {n} is the interrupt to test:\n");
	printf("		3, 4, 5, 7, 10, 11, 12, or 15\n");
	printf("                If several interrupts are listed, \n");
	printf("                the first is used in normal operation;\n");
	printf("                the remainder only by the 'interrupt test'.\n");
	printf("                Ranges are allowed, e.g., I4-7 I10-15\n");
	printf("     C{n}   --  {n} is the selected channel; 0 through 63\n");
	printf("                Ranges are allowed, e.g., C0-7 C16-31\n");
	printf("                Applies to all tests which operate on a per-channel\n");
	printf("                basis.  If not supplied, tests will be performed on all.\n");
	printf("     Y     --   Suppresses all operator prompts unless essential\n");
	printf("     E{x}  --   {x} specifies the type of board you are testing:\n");
	printf("                X  = expandable\n");
	printf("                X6 = expandable using 16-bit data interface\n");
	printf("                X8 = expandable using  8-bit data interface\n");
	printf("                N  = non-expandable\n");
	printf("                 (default: determined from power-on reset message)\n");
	printf("     B{x}  --  {x} specifies the type of bus\n");
	printf("                I = ISA (default)\n");
	printf("                M = Micro Channel\n");
	printf("                E = EISA \n");
	printf("                P = PCI\n");
	printf("     F     --   Stop test after the first error\n");
	printf("     Z     --   Double all delays\n");
	printf("     ,     --   Pause for operator after any error\n");
	printf("     M     --   Mark parity/framing/overruns with %c\n", 0x80);
	printf("     MM    --   Report parity/framing/overruns using status packets\n");
	printf("     G{file} -- Log errors to specified file\n");
	printf("     O{x...} -- Turn on/off output options\n");
	printf("                {x...} is a list of letters; each one turns on an option:\n");
	printOutputHelps();
	printf("\f\n");
	printf("[protocols] may be given to restrict the tests to a specific\n");
	printf("     baud rate, character size, parity, etc.  If not specified,\n");
	printf("     protocols will be chosen randomly\n");
	printf("Protocols and descriptions:\n");

	printProtocolHelps(N_SIZE, p_size,"per character");
	printProtocolHelps(N_STOPS, p_stops,"stop bit(s)");
	printProtocolHelps(N_PARITY, p_parity,"parity");
	printProtocolHelps(N_UNIQ_BAUD, p_baud,"baud");
	printProtocolHelps(N_FLOW, p_flow1, "flow control");

	printf("\nTests are selected by number; e.g., t1 t2 t3\n");
	printf("     Ranges are allowed, e.g., t1-10 \n");
	printf("     A number following a decimal point overrides the default\n");
	printf("     number of passes for that test, e.g., t5.7 runs seven passes of test 5.\n");
	printf("     Do not specify passes in the same argument as a range.\n");
	printf("     Tests are performed in the order first listed.  If listed\n");
	printf("     multiple times, they only performed the first time, but\n");
	printf("     the last number of passes specified is used.\n");
	printf("     For example, t4-8 t5.3 performs tests 4 through 8 in order\n");
	printf("     performing only three passes of test 5.\n");
	printf("     To eliminate selected tests from a range, select it afterward\n");
	printf("     with zero passes.  For example t4-8 t5.0 performs tests 4, 6, and 8\n");
	printf("\f\n");
	printf("Tests & Descriptions:\n");

	/* Spin through the tables here...! */

	for(ptd = testList, i = 1; ptd->testName; i++ , ptd++)
	{
		if (ptd->flags & E_RESERVED) continue;

		printf("t%-2d  %s test \n       (default %d, maximum %d %s)\n",
				i, ptd->testName, ptd->defPasses, ptd->maxPasses,
				(ptd->flags & TIMED_TEST) ? 
					(ptd->defPasses == 1 ? "second" : "seconds" ) :
					(ptd->defPasses == 1 ? "pass" : "passes" ) );

		printf("     %s\n",
			ptd->testDesc);
		printf("     Comments:\n");
		if (ptd->flags & E_FATAL) 
			printf("       Any error stops test.\n");
		if (ptd->pTest[1] == NULL)
			printf("       Non-expandable only.\n");
		if (ptd->pTest[0] == NULL)
			printf("       Expandable only.\n");
		if (ptd->flags & DIAG_LW)
			printf("       Uses special loadware.\n");
		if (ptd->flags & PERCHANNEL)
			printf("       Channels may be selected on command line.\n");
		if (ptd->flags & PLUG_BITS)
		{
			printf("       Requires loopback plug:\n          ");
			xlatePlugCode(ptd->flags);
		}
		if (ptd->flags & DEBUG_NEED)
			printf("       Needs debug port (power off before attaching.)\n");
		if (ptd->flags & NEED_ARG_E)
			printf("       Requires EX or EN argument on command line \n");
	}
	printf("\f\n");
	printf("Numbers used in options or to identify tests may be \n");
	printf("given in octal, decimal, or hexadecimal notation.\n\n");
	printf("Numbers beginning with '0x' are taken as hexadecimal.\n");
	printf("This is the natural way to express board addresses, e.g. 0x308.\n");
	printf("Numbers beginning with '0' are taken as octal.\n");
	printf("Numbers beginning with other digits are taken as decimal.\n");
	rip(7, 0);
}


/*--------------------------------------------
 * getNumeric(buffer)
 * Converts a string to a numeric value
 *--------------------------------------------
 */
int
getNumeric(buffer)
register char *buffer;
{
register char c;
register int value = 0;

	if (*buffer == '0')
	{
		buffer++;
		if (*buffer == 'x')
		{
			buffer++;
			/* Convert from hex */
			while ((c = *buffer++) != 0)
			{
				/* Check for incipient overflow and max out */
				if (value & 0xf000) return 0x7fff;

				if (c >= '0' && c <= '9')
				{
					value = (value << 4) + c - '0';
				}
				else if (c >= 'a' && c <= 'f')
				{
					value = (value << 4) + c + 10 - 'a';
				}
				else if (c >= 'A' && c <= 'F')
				{
					value = (value << 4) + c + 10 - 'A';
				}
				else return -1;
			}
			return value;
		}
		else
		{
			/* Convert from octal */
			while ((c = *buffer++) != 0)
			{
				if (value >= 0x1000) return 0x7fff;
				if (c >= '0' && c <= '7')
				{
					value = (value << 3) + c - '0';
				}
				else return -1;
			}
			return value;
		}
	}
	else
	{
		/* Convert from decimal */
		while ((c = *buffer++) != 0)
		{
			if (value > 3276  || (value == 3276 && c > '7')) return 0x7fff;
			if (c >= '0' && c <= '9')
			{
				value = (value * 10) + c - '0';
			}
			else return -1;
		}
		return value;
	}
}
	

/*--------------------------------------------------------------
 * findCharacter(buffer, c)
 * returns -1 if the given character isn't in the string,
 * returns the position of same (0 - length-1) if it does appear
 *--------------------------------------------------------------
 */

int
findCharacter(buffer,c)
register char *buffer;
register char c;
{
char *oldBuffer = buffer;

	while (*buffer)
	{
		if (*buffer == c)
		{
			return buffer-oldBuffer;
		}
		buffer++;
	}
	return -1;
}

int
irqToBit(irq)
int irq;
{
	switch(irq)
	{
	case 3:	 return IRQ_3;
	case 4:	 return IRQ_4;
	case 5:	 return IRQ_5;
	case 7:	 return IRQ_7;
	case 10: return IRQ_10;
	case 11: return IRQ_11;
	case 12: return IRQ_12;
	case 15: return IRQ_15;
	default: return -1;
	}
}



/*-------------------------------------
 * printOutputHelps()
 *
 * Prints the list of arguments for the
 * O option and their explanations
 *-------------------------------------
 */

void
printOutputHelps()
{
register struct vTypes *pvt = vTarray;

	while(pvt->vFlag)
	{
		if (pvt->explan[0] != '~')
		{
			printf("         %c -- Display %s.\n", pvt->c2, pvt->explan);
		}
		pvt++;
	}
}



/*----------------------------------------------
 * printProtocolHelps(npro, ppro, description)
 * npro = number of entries to search,
 * ppro = pointer to the first entry
 * description = additional text to print
 *
 * Prints help information for this type of table.
 *----------------------------------------------
 */

void
printProtocolHelps(npro, ppro, description)
register int npro;
register pproTable ppro;
char *description;
{
	while(npro--)
	{
		printf("     -%-10s --  %s %s\n",
			ppro->proRecog, ppro->proDesc, description);
		ppro++;
	}
}


/*-------------------------------
 * stringToIndex(npro, ppro,string)
 * npro = number of entries to search,
 * ppro = pointer to the first entry
 * string = string to search for
 * 
 * Returns -1 if string isn't found,
 * otherwise returns index.
 *-------------------------------
 */

int
stringToIndex(npro, pproSave, string)
register int npro;
pproTable pproSave;
char *string;
{
register pproTable ppro = pproSave;

	while(npro--)
	{
		if (strcmp (ppro->proRecog, string) == 0)
		{
			return ppro - pproSave;
		}
		ppro++;
	}
	return -1;
}
	

/*-------------------------
 * parseArguments(argc,argv)
 *
 * Sets up test control structures accodring
 * to tests and options selected
 * (see also usage, above)
 *-------------------------
 */

void
parseArguments(argc,argv)
int argc;
char **argv;
{
register char *pc;
int i,j,k;
pfunk pf;
ptestdef ptd;
int dash, dot, bit, min, max, box, pass;
struct vTypes *pvt;

	/* Store the program name for all to find*/
	/* We aren't using arg[0] because it returns the whole path name, 
	 * including the drive letter and .EXE extension!
	 */
	executableName = "i2diag";

	if (argc < 2) usage();	/* No arguments at all */

	while(--argc)
	{
		++argv;
		pc = *argv;
		switch (*pc)
		{
		/* Only check the first letter: assume to be help */
		case 'X':
		case 'x':
			Persist = 1;
			break;

		case 'H':
		case 'h':
			usage_h();
			break;

		/* Specify number of passes */
		case 'P':
		case 'p':
			i = getNumeric(&pc[1]);
			if (i < 0) usage_bad("Number expected after 'P'");
			mainPasses = i;
			break;

		/* Specify whether to test Expandable or Non-expandable */
		case 'E':
		case 'e':
			hasUsedE = 1;
			switch(pc[1])
			{
			case 'X':
			case 'x':
				isExpandable = EXPAND_TYPE;
				switch(pc[2])
				{
				case '6':
					is16bit = 1;
					break;
				case '8':
					is16bit = 0;
					break;
				default:
					is16bit = -1;
					break;
				}
				break;	

			case 'N':
			case 'n':
				isExpandable = NONEXP_TYPE;
				break;

			default:
				usage_bad("X or N expected after E");
			}
			break;
				
		/* Specify whether Microchannel, or what... */
		case 'B':
		case 'b':
			hasUsedB = 1;
			switch(pc[1])
			{
			case 'I':
			case 'i':
				busType = ISA_OPT;
				break;

			case 'M':
			case 'm':
				busType = MCA_OPT;
				break;

			case 'E':
			case 'e':
				busType = EISA_OPT;
				break;

			case 'P':
			case 'p':
				busType = PCI_OPT;
				break;

			default:
				usage_bad("I, M, E, or P expected after B");
			}
			break;


		/* Specify base address */
		case 'A':
		case 'a':
			hasUsedA = 1;
			i = getNumeric(&pc[1]);
			if ( (i & 7) || !(i & 0xff00))
			{
				usage_bad("Invalid board address");
			}
			testAddress = i;
			break;

		/* Specify a log file */
		case 'G':
		case 'g':
			logFd = open(&pc[1], O_RDWR | O_CREAT | O_APPEND, 
				S_IREAD | S_IWRITE);
			break;
		

		/* Specify output level */
		case 'O':
		case 'o':
			mainVerbosity = 0;
			while ((i = *(++pc)) != 0)
			{
				for (pvt = vTarray; pvt->vFlag; pvt++)
				{
					if (i == pvt->c1 || i == pvt->c2)
					{
						mainVerbosity |= pvt->vFlag;
						break;
					}
				}
				if (!(pvt->vFlag))
				{
					usage_bad("Unexpected characters after 'O' option");
				}
			}
			break;

		/* Specify interrupt level(s) to test */
		case 'I':
		case 'i':
			dash = findCharacter(&pc[1], '-');
			if (dash < 0)
			{
				i = getNumeric(&pc[1]);
				if ((bit = irqToBit(i)) == -1) usage_bad("Invalid IRQ given");
				for (pf = argumentList, j = realTests;
					j; j--, pf++)
				{
					pf->f_interrupts |= bit;
					if (pf->f_mainint == 0)
					{
						pf->f_mainint = i;
					}
				}
				mainInterrupts |= bit;
				if (chiefInterrupt == 0) chiefInterrupt = i;
			}
			else
			{
				if(dash == 0) usage_bad("'-' used without a range");
				pc[1+dash] = 0;	
				min = getNumeric(&pc[1]);
				if ((bit = irqToBit(min)) == -1) usage_bad("Invalid IRQ given");
				max = getNumeric(&pc[2+dash]);
				if ((bit = irqToBit(max)) == -1) usage_bad("Invalid IRQ given");
				if (min > max) usage_bad("IRQ range backward");
				for(k = 0; k < 8; k++)
				{
					i = validIrqList[k];
					if (i < min || i > max) continue;

					bit = irqToBit(i);
					for (pf = argumentList, j = realTests;
						j; j--, pf++)
					{
						pf->f_interrupts |= bit;
						if (pf->f_mainint == 0) 
							pf->f_mainint = i;
					}
					mainInterrupts |= bit;
					if (chiefInterrupt == 0) chiefInterrupt = i;
				}
			}
			break;
	
		/* Specify Channels to test */
		case 'C':
		case 'c':
			dash = findCharacter(&pc[1], '-');
			if (dash < 0)
			{
				i = getNumeric(&pc[1]);
				if (i < 0 || i > 63) usage_bad("Invalid channel number given");
				bit = 1 << (i & 0xf);	/* low order is channel */
				box = i >> 4;			/* high order is box number */
				for (pf = argumentList, j = realTests;
					j; j--, pf++)
				{
					pf->f_channels[box] |= bit;
				}
			}
			else
			{
				if(dash == 0) usage_bad("'-' used without a range");
				pc[1+dash] = 0;	
				min = getNumeric(&pc[1]);
				if (min < 0 || min > 63) 
					usage_bad("Invalid channel number given");
				max = getNumeric(&pc[2+dash]);
				if (max < 0 || max > 63) 
					usage_bad("Invalid channel number given");
				if (min > max) usage_bad("range of channels backward");
				for(i = min; i <= max; i++)
				{
					bit = 1 << (i & 0xf);	/* low order is channel */
					box = i >> 4;	/* high order is box number */

					for (pf = argumentList, j = realTests;
						j; j--, pf++)
					{
						pf->f_channels[box] |= bit;
					}
				}
			}
			break;

		/* Specify test.passes or range of tests */
		/* Note on reserved test numbers:
		 * You cannot list them alone on the command line,
		 * but they may appear in a range (so the "holes" get
		 * skipped)
		 */
		case 'T':
		case 't':
			dash = findCharacter(&pc[1], '-');
			if (dash < 0)
			{
				dot = findCharacter(&pc[1], '.');
				if (dot >= 0)
				{
					if (dot == 0) usage_bad("'.' used without pass count");
					pc[1+dot] = 0;
					pass = getNumeric(&pc[2+dot]);
					if (pass < 0) usage_bad("Invalid number of passes");
				}
				/* Get the number of the test */
				i = getNumeric(&pc[1]);
				if (i < 1 || i > realTests) usage_bad("Invalid test number given");
				ptd = &testList[--i];	/* Change to 0-based */
				if (ptd->flags & E_RESERVED) usage_bad("Reserved test number given");
				if (ptd->sequence == 0)
				{
					ptd->sequence = ++testSequence;
				}
				if (dot >= 0) 
				{
					argumentList[i].f_passes = (pass < ptd->maxPasses) ?
						pass : ptd->maxPasses;
				}
			}
			else
			{
				if(dash == 0) usage_bad("'-' used without a range");
				pc[1+dash] = 0;	
				min = getNumeric(&pc[1]);
				if (min < 1 || min > realTests) 
					usage_bad("Invalid Test in range given");
				max = getNumeric(&pc[2+dash]);
				if (max < 1 || max > realTests) 
					usage_bad("Invalid Test in range given");
				if (min > max) usage_bad("range of tests backward");

				/* Note boundaries are adjusted for 0-based array 
				 * indexing, where test numbers begin at 1
				 */
				--min;
				for(i = min, ptd = &testList[min]; i < max; i++, ptd++)
				{
					if (!(ptd->flags & E_RESERVED) && ptd->sequence == 0)
					{
						ptd->sequence = ++testSequence;
					}
				}
			}
			break;

		case '-':
			if ( (i = stringToIndex(N_SIZE, p_size, &pc[1]) ) >= 0)
			{
				freezeSize = i;
			}
			else if ( (i = stringToIndex(N_STOPS, p_stops, &pc[1]) ) >= 0)
			{
				freezeStops = i;
			}
			else if ( (i = stringToIndex(N_PARITY, p_parity, &pc[1]) ) >= 0)
			{
				freezeParity = i;
			}
			else if ( (i = stringToIndex(N_UNIQ_BAUD, p_baud, &pc[1]) ) >= 0)
			{
				freezeBaud = i;
			}
			else if ( (i = stringToIndex(N_FLOW, p_flow1, &pc[1]) ) >= 0)
			{
				freezeFlow = i;
			}
			else
			{
				usage_bad("Invalid protocol specified");
			}
			break;

		case 'Y':
		case 'y':
			interaction = 0;
			break;
		default:
			usage_bad("Bad command-line option");
			break;

		case 'L':		/* Un-documented: to just load the file */
			loadOnly = 1;
			break;

		case 'F':
		case 'f':
			isFatal = 1;	/* All errors are fatal */
			break;

		case 'M':
		case 'm':
			if (pc[1] == 'M' || pc[1] == 'm')
			{
				statParity = 1;	/* Errors are by status */
				checkParityOption = CSE_ESTAT;
			}
			else
			{
				statParity = 0;
				checkParityOption = CSE_NULL;	/* Errors are by NULL */
			}
			break;

		case ',':
			isHolding = 1;	/* Pause after each error */
			break;

		case '+':
			for (i = 0; i < MAX_FORCE; i++)
			{
				if(!pc[1+i]) break;
				argString[i] = pc[1+i];
			}
			/* We haven't copied a 0 yet! */
			forceStringLength = i;
			argString[i] = 0;
			argInt = getNumeric(argString);
			break;

		case 'W':
		case 'w':
			/* When resetting board, if it reports no channels
			 * found, reset it and try again.  This is used
			 * for common-i/o board test-beds.
			 */
			waitForChannels++;
			break;

		case 'Z':
		case 'z':
			delayMult++;
			break;
					
		} /* End switch */
	} /* End of while(argc) */
}


/*----------------------
 * ERROR REPORTING 
 *----------------------
 */

/* Table of handlers for each type of error message */
/* The handler consists of an error message string and a function
 * to display the parameters.
 * 
 * These must remain harmonious with the manifests defined in i2dghw.h 
 */

typedef struct _errTable {
	char *errString;
	void (*errorHandler)(pfunk);
} errTable, *perrTable;
	
char badTime[] = "Timeout while starting test";
char badTPass[] = "Timeout waiting for pass to complete";
char badMSize[] = "Invalid on-board memory size";
char badStrange[] ="Unexpected error encountered";

errTable eList[] = {
	"Test completed successfully",			/* OKEY  0 */
		erOKEY,
	badMSize, 								/* MW_BAD_MEM_SIZE	1 */
		erMEMVAL,
	"On-board memory test failed",			/* MW_MEM_FAIL		2 */
		erMEMFAIL,
	badMSize, 								/* MR_BAD_MEM_SIZE	3 */
		erMEMVAL,
	"On-board memory test failed",			/* MR_MEM_FAIL		4 */
		erMEMFAIL,
	"Initial AF/AE register error",			/* FS_AF_AE 		5 */
		erFIFO,
	"Initial mask register error",			/* FS_MASK			6 */
		erFIFO,
	"AF/AE register test failed",			/* FS_AF_AE_D		7 */
		erFIFO,
	"Mask register test failed",			/* FS_MASK_D		8 */
		erFIFO,
	"Byte-detect register test failed",		/* FS_BYTE_D		9 */
		erFIFO,
	"Outbound mailbox test failed",			/* FS_MAILBOX_D		10 */
		erFIFO,
	"Board timed out sending data to host",	/* FDMA_XTIMO		11 */
		erFTDATA,
	"Data mismatch error",					/* FLL_DATA_ERROR	12 */
		erFDATA,
	"Data mismatch error (single access)",	/* FLL_DATA_ERROR16_EZ	13	*/
		erFDATA16,
	"Data mismatch error (repeat access)",	/* FLL_DATA_ERROR16	14 */
		erFDATA16,
	"Timeout receiving mailbox data",		/* FM_TIMEOUT_IN 	15 */
		erFTDATA,
	"Timeout sending mailbox data",			/* FM_TIMEOUT_OUT	16 */
		erFTDATA,
	"Host Received bad mailbox data",		/* FM_DATA_IN		17 */
		erFMDATA,
	"Board Received bad mailbox data",		/* FM_DATA_OUT		18 */
		erFMDATA,
	"Board hardware register test failed",	/* RT_FAIL 			19 */
		erREGFAIL,
	"Interrupt test failed",				/* IT_FAIL			20 */
		erINTERRUPT,
	"Board timed out reading data from host",	/* FDMA_RTIMO	21 */
		erFTDATA,
	badTime,								/* MW_TIMEOUT1		22 */
		erDUMB,
	badTPass,								/* MW_TIMEOUT2		23 */ 
		erFTDATA,
	badTime,								/* MR_TIMEOUT1		24 */
		erDUMB,
	badTPass, 								/* MR_TIMEOUT2		25 */
		erFTDATA,
	"Timeout looping data back",			/* FLL_TIMEOUT2		26 */
		erFTIME,
	"Timeout getting data",					/* FRL_TIMEOUT2		27 */
		erFTIME,
	"Invalid FIFO size reported",			/* FL_BAD_SIZE		28 */
		erReportFsize,
	"FIFO size not as reported",			/* FV_BAD_SIZE		29 */
		erReportFVsize,
	badStrange,								/* FLL_GENERAL		30 */
		erFSTRANGE,
	badStrange,								/* FRL_GENERAL		31 */
		erFSTRANGE,
	badStrange,								/* RT_GENERAL		32 */
		erSTRANGE,
	"Board received mis-matching data",		/* FDMA_MISM		33 */
		erFTDATA,
	badStrange,								/* IT_GENERAL		34 */
		erSTRANGE,
	"Error(s) in loopback test",			/* LOOPBACK_ERR		35 */
		erLOOPBACK,
	"Output FIFO will not drain",			/* FV_NODRAIN		36 */
		erDUMB,
	"Host timed out waiting for fifo data",	/* FDMA_HDATA		37 */
		erFTDATA,
	"Host timed out reading/writing mailbox",	/* FDMA_HMAIL	38 */
		erFTDATA,
	"Incorrect Baud Rate",					/* BAD_BAUD_ERR		39 */
		erDUMB,
	"Micro Channel POS[2] error", 			/* BAD_POS_2		40 */
		erFMDATA,
	"Micro Channel POS[3] error",			/* BAD_POS_3		41 */
		erFMDATA,
	"This test only for Micro Channel (use BM argument)",
		erDUMB,								/* BAD_MC			42 */
	"This test only for PCI (use BP argument)",
		erDUMB,								/* BAD_PCI			43 */
	"AMCC error (OMB)",
		erPCIFAIL,							/* BAD_OMB			44 */
};


/* Helper routines used by parameter-reporting routines */

void erReportPass(pf)
pfunk pf;
{
	eprintf("     Failed on pass %d.\n", pf->f_passfail + 1);
}

void erReportFsize(pf)
pfunk pf;
{
	eprintf("     %d-byte fifo size reported:\n", pf->f_results[0]);
}

void erReportFVsize(pf)
pfunk pf;
{
	eprintf("     %d-byte fifo size reported:\n", pf->f_results[0]);
	eprintf("     %d-byte fifo size mesaured:\n", pf->f_results[1]);
}


void erReportByteError(pf)
pfunk pf;
{
	eprintf("     Wrote data %02x, read back %02x\n",
		pf->f_wstring[0], pf->f_rstring[0]);
}


void erReportWordError(pf)
pfunk pf;
{
	eprintf("     Wrote data %04x, read back %04x\n",
		*((unsigned short *)pf->f_wstring),
		*((unsigned short *)pf->f_rstring));
}


/* Individual parameter-reporting routines */
	
void erOKEY(pf)
pfunk pf;
{
	pf++;		/* My way of telling the compiler, yes I know */
			/* I am not using pf */
	return;
}


void erDUMB(pf)
pfunk pf;
{
	pf++;	/* So compiler doesn't complain we are ignoring args */
	return;
}

void erMEMVAL(pf)
pfunk pf;
{
	eprintf("     Board reports %dk.\n", 
		pf->f_results[0]);
}


void erMEMFAIL(pf)
register pfunk pf;
{
	erReportPass(pf);
	eprintf("     %dk on-board memory reported\n", pf->f_results[0]);
	eprintf("     Address %04x:%04x -- Wrote %02x%02x, Read %02x%02x\n",
		pf->f_results[1], pf->f_results[2],
		pf->f_wstring[1], pf->f_wstring[0],
		pf->f_wstring[3], pf->f_wstring[2]);
}

void erPCIFAIL(pf)
register pfunk pf;
{
	erReportPass(pf);
	eprintf("     Address %04x -- Wrote %02x, Read %02x\n",
		pf->f_results[0], pf->f_wstring[0], pf->f_wstring[1]);
}

void erFIFO(pf)
pfunk pf;
{

	erReportByteError(pf);
}


void erFTIME(pf)
pfunk pf;
{
	erReportPass(pf);
}


void erFDATA(pf)
pfunk pf;
{
int hit = 0;
int word = 0;

	erReportPass(pf);
	switch(pf->f_results[1])
	{
	case 1:
		eprintf("       Low Byte, ");
		hit = 1;
		break;
	case 2:
		eprintf("       High Byte, ");
		hit = 1;
		break;
	case 3:
		eprintf("       High & Low Bytes, ");
		hit = 1;
		word = 1;
		break;
	default:
		break;
	}
	switch(pf->f_results[2])
	{
	case 1:
		eprintf("       Writing to board");
		hit = 1;
		break;
	case 2:
		eprintf("       Reading from board");
		hit = 1;
		break;
	case 0:
		break;
	}
	if (hit) eprintf("\n");
	
	if (word || pf->f_results[3])
	{
		erReportWordError(pf);
	}
	else
	{
		erReportByteError(pf);
	}
}

void erFDATA16(pf)
pfunk pf;
{
	erReportPass(pf);
	erReportWordError(pf);
}

void erFTDATA(pf)
pfunk pf;
{
	erReportPass(pf);
}


void erFMDATA(pf)
pfunk pf;
{
	erReportPass(pf);
	erReportByteError(pf);
}


void erREGFAIL(pf)
pfunk pf;
{
	erReportPass(pf);
	switch(pf->f_results[0])
	{
	case DEVICE_1400:
		eprintf("     UXART %d failure, %s register\n",
			pf->f_results[1], 
			pf->f_results[2] == GLOBAL_REG ? "global" : "channel");
		break;
	case DEVICE_8259:
		eprintf("     Interrupt controller %d failure\n",
			pf->f_results[1]);
		break;
	case DEVICE_80186:
		eprintf("     Processor failure");
		eprintf("     Register address 0xff%x \n", pf->f_results[3]);
		pf->f_wstring[1] = pf->f_results[1];
		pf->f_rstring[1] = pf->f_results[2];
		erReportWordError(pf);
		return;
	
	case DEVICE_2692:
		eprintf("     Debug port failure");
		break;
	case DEVICE_MAPPER:
		eprintf("     Chip-mapper failure");
		break;
	case DEVICE_LCD:
		eprintf("     LCD controller %d: ", pf->f_results[1]);
		switch(pf->f_results[2])
		{
		case 0:
			eprintf("loopback error");
			break;
		case 1:
			eprintf("write timeout\n");
			return;		/* The other info is irrelevant */
		case 2:
			eprintf("read timeout\n");
			return;	
		}
		break;
	case DEVICE_654:
		eprintf("     ST654 %d failure\n",
			pf->f_results[1]);
		break;
	default:
		eprintf("     Software error: unexpected device code %d\n",
				pf->f_results[0]);
		break;
	}
	eprintf("     Register address 0x%x \n", pf->f_results[3]);
	erReportByteError(pf);
}


void
printIrqMap(pfunk pf)
{
register ushort *pRes = pf->f_results;

	eprintf("                   3     4     5     7    10    11    12    15\n");
	eprintf("       issued:%6u%6u%6u%6u%6u%6u%6u%6u\n",
		pRes[0], pRes[1], pRes[2], pRes[3],
		pRes[4], pRes[5], pRes[6], pRes[7]);
		
	eprintf("       caught:%6u%6u%6u%6u%6u%6u%6u%6u\n",
		pRes[8], pRes[9], pRes[10], pRes[11],
		pRes[12], pRes[13], pRes[14], pRes[15]);
}


void erINTERRUPT(pf)
pfunk pf;
{
register ushort *pRes;
int flags, i;

	erReportPass(pf);
	if (mainVerbosity & (V_ERROR | V_IRQ))
	{
		printIrqMap(pf);
	}
	flags = 0;
	for (pRes = pf->f_results, i = 0;  i < 8; i++, pRes++)
	{
		if		(pRes[0] < pRes[8]) flags |= 1;
		else if	(pRes[0] > pRes[8]) flags |= 2;
	}
	switch(flags)
	{
		case 1:
			eprintf("    Extra Interrupts");
			break;
		case 2:
			eprintf("    Missing Interrupts");
			break;
		case 3:
			eprintf("    Extra & Missing Interrupts");
			break;
		case 0:
			eprintf("    No discrepancies");
			break;
	}
	eprintf(" detected\n");
		
}

void erFSTRANGE(pf)
pfunk pf;
{
	erReportPass(pf);
	eprintf("     Error code 0x%x\n", pf->f_results[1]);
}

void erSTRANGE(pf)
pfunk pf;
{
	erReportPass(pf);
	eprintf("     Error code 0x%x\n", pf->f_results[0]);
}


/* Note that this need do NOTHING at higher verbosities,
 * because we have already printed errors
 */

void erLOOPBACK(pf)
pfunk pf;
{
register uchar *cp;
uchar *ppass;
int i;
int width = 0;

	if (!(mainVerbosity & V_NORMAL)) return;

	eprintf(" ");
	for (i = 0, cp = pf->f_rstring, ppass = pf->f_wstring ; 
			i < 64; i++, cp++, ppass++)
	{
		if (width > 55)
		{
			eprintf("\n ");
			width = 0;
		}
		switch (*cp)
		{
		case L_ERR_NOXMT:
			width += eprintf("ch %2d pass %d: no xmit.  ", i, *ppass);
			break;
		case L_ERR_LATE:
			width += eprintf("ch %2d pass %d: missing.  ", i, *ppass);
			break;
		case L_ERR_WRONG:
			width += eprintf("ch %2d pass %d: bad.      ", i, *ppass);
			break;
		case L_ERR_DCDDN:
			width += eprintf("ch %2d pass %d: DCD low.  ", i, *ppass);
			break;
		case L_ERR_DCDUP:
			width += eprintf("ch %2d pass %d: DCD up.   ", i, *ppass);
			break;
		case L_ERR_CTSDN:
			width += eprintf("ch %2d pass %d: CTS low.  ", i, *ppass);
			break;
		case L_ERR_CTSUP:
			width += eprintf("ch %2d pass %d: CTS up.   ", i, *ppass);
			break;
		case L_ERR_DSRDN:
			width += eprintf("ch %2d pass %d: DSR low.  ", i, *ppass);
			break;
		case L_ERR_DSRUP:
			width += eprintf("ch %2d pass %d: DSR up.   ", i, *ppass);
			break;
		case L_ERR_RIDN:
			width += eprintf("ch %2d pass %d: RI low.  ", i, *ppass);
			break;
		case L_ERR_RIUP:
			width += eprintf("ch %2d pass %d: RI up.   ", i, *ppass);
			break;
		case L_ERR_BREAKS:
			width += eprintf("ch %2d pass %d: # breaks. ", i, *ppass);
			break;
		case L_ERR_GOTDATA:
			width += eprintf("ch %2d pass %d: LEAKING. ", i, *ppass);
			break;
		default:
			break;
		}
	}
	if (width) eprintf("\n");
}



/* Helper-routine for the main-level reporting functions. */

void
reportFailure(pf, i, pc)
pfunk pf;	/* The test name */
int i;		/* The error code (Note! no bounds check here) */
char *pc;	/* Pointer to test name (in case we need it) */
{
register perrTable pert;

	/* If we were only reporting failures, we havn't printed
	 * the test name yet.
	 * If we have a log, the log will never have gotten the
	 * test name yet.
	 */
	if (!(mainVerbosity & V_NORMAL) || logFd >= 0)
	{
		/* If printing because there is a log, display an extra
		 * newline to prevent screen from being confused.
		 * The log file does not need the extra \n.
		 */

		if(mainVerbosity & V_NORMAL) printf("\n");
		eprintf("t(%d): %s Test Failed\n", (pf - argumentList) + 1, pc);
	}
	else
	{
		eprintf("\n");
	}

	pert = &eList[i];
	eprintf("   %s\n", pert->errString);
	(*(pert->errorHandler))(pf);
	holdIfNeeded();
}

/* The outer-level reporting functions */

void
r_goodnul(pf,ptd)
pfunk pf;
ptestdef ptd;
{
int i;

	i = pf->f_error;

	if (i == OKAY)
	{
		if (mainVerbosity & V_NORMAL)
		{
			printf( (ptd->flags & CANNOT_FAIL) ?
				"\n" : "Passed\n");
		}
	}
	else
	{
		reportFailure(pf, i, ptd->testName);
	}
}

void
r_goodint(pf,ptd)
pfunk pf;
ptestdef ptd;
{
int i;

	i = pf->f_error;

	if (i == OKAY)
	{
		if (mainVerbosity & V_NORMAL)
		{
			printf( (ptd->flags & CANNOT_FAIL) ?
				"\n" : "Passed\n");
			if (mainVerbosity & V_IRQ)
			{
				printIrqMap(pf);
			}
		}
	}
	else
	{
		reportFailure(pf, i, ptd->testName);
	}
}

void
r_goodfsize(pf,ptd)
pfunk pf;
ptestdef ptd;
{
int i;

	i = pf->f_error;

	if (i == OKAY)
	{
		if (mainVerbosity & V_NORMAL)
		{
			printf(" %d-byte fifo\n", pf->f_results[0]);
		}
	}
	else
	{
		reportFailure(pf, i, ptd->testName);
	}
}
void
r_goodmem(pf,ptd)
pfunk pf;
ptestdef ptd;
{
int i;

	i = pf->f_error;

	if (i == OKAY)
	{
		if (mainVerbosity & V_NORMAL)
		{
			printf("Passed:  %dk board memory\n", pf->f_results[0]);
		}
	}
	else
	{
		reportFailure(pf, i, ptd->testName);
	}
}

/*------------------------------------------------------
 * Helper routine for reporting loadware loading errors
 *------------------------------------------------------
 */

/* Helper routine for same... */

void ltrError(char *pcName, char *pc)
{
	eprintf("\n   Error attempting to write loadware %s:\n     %s\n",
		pcName, pc);
	rip(5,"--- Terminating test --- \n");
}



/*-----------------------------------------------------------------
 * sendEnables();
 *
 * Enable all data-set signal reporting 
 * Even though, DSR and RI exist only on the expandable 
 * This to make sure the non-expandable doesn't get up to mischief 
 *----------------------------------------------------------------
 */

void
sendEnables()
{
char buffer[16];

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 5,
		CMD_CTS_REP, CMD_DCD_REP, CMD_DSR_REP, CMD_RI_REP,
		CMD_BRK_REP(CBK_STAT) );
	
	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);

}


/* ------------------------------
 * Functions which perform tests
 *-------------------------------
 */

int internal_loop(pf)
pfunk pf;
{
char buffer[16];

	/* Initialize all the active channels to 
	 * internal loopback mode, possible error-reporting
	 */

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 3,
		CMD_SET_ERROR(checkParityOption),
		CMD_INLOOP(CIN_ENABLE), CMD_PARCHK(CPK_ENAB));

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);

	/* Perform the loopback test */
	return loopBack(pf, 0, CONTINUE_DATA_1);
}

int data_loop1(pf)
pfunk pf;
{
char buffer[16];

	/* Initialize all the active channels to 
	 * external loopback mode, possible error reporting.
	 */

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 4,
		CMD_DTRRTS_UP,
		CMD_SET_ERROR(checkParityOption),
		CMD_INLOOP(CIN_DISABLE), CMD_PARCHK(CPK_ENAB));

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);


	/* Perform the loopback test */
	return loopBack(pf, 0, CONTINUE_DATA_2);
}

int data_loop2(pf)
pfunk pf;
{
char buffer[16];

	/* Initialize all the active channels to 
	 * external loopback mode, possible error reporting.
	 */

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 4,
		CMD_DTRRTS_DN,
		CMD_SET_ERROR(checkParityOption),
		CMD_INLOOP(CIN_DISABLE), CMD_PARCHK(CPK_ENAB));

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);


	/* Perform the loopback test */
	return loopBack(pf, 0, CONTINUE_DATA_2);
}


int data_loop3(pf)
pfunk pf;
{
char buffer[16];

	/* Initialize all the active channels to 
	 * external loopback mode, possible error reporting.
	 */

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 4,
		CMD_DTRRTS_UP,
		CMD_SET_ERROR(checkParityOption),
		CMD_INLOOP(CIN_DISABLE), CMD_PARCHK(CPK_ENAB));

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);

	/* Perform the loopback test in torture mode */
	return loopBack(pf, 1, CONTINUE_DATA_4);
}


int data_loop4(pf)
pfunk pf;
{
char buffer[16];
int status;

	/* Initialize all the active channels to 
	 * external loopback mode, possible error reporting.
	 * also enable CTSflow and RTSflow, disable CTS reporting.
	 */

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 7,
		CMD_DTRRTS_UP,
		CMD_SET_ERROR(checkParityOption),
		CMD_INLOOP(CIN_DISABLE), CMD_PARCHK(CPK_ENAB),
		CMD_CTS_NREP, CMD_CTSFL_ENAB, CMD_RTSFL_ENAB);

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);

	/* Perform the loopback test */
	/* Since RTS flow is loadware not hardware, do not run ultra-fast */
	status = loopBack(pf, 0, CONTINUE_DATA_5);

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 3,
		CMD_CTSFL_DSAB, CMD_RTSFL_DSAB, CMD_CTS_REP);

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);

	return status;
}

int data_loop5(pf)
pfunk pf;
{
char buffer[16];
int status;

	/* Initialize all the active channels to 
	 * external loopback mode, possible error reporting.
	 * also enable CTSflow and DTRflow, disable CTS reporting.
	 */

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 7,
		CMD_DTRRTS_UP,
		CMD_SET_ERROR(checkParityOption),
		CMD_INLOOP(CIN_DISABLE), CMD_PARCHK(CPK_ENAB),
		CMD_CTS_NREP, CMD_CTSFL_ENAB, CMD_DTRFL_ENAB);

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);

	/* Perform the loopback test */
	/* Both CTSFLOW and DTRFLOW are hardware! go for it! */
	highBaudOnly = 1;		/* Force random choices to be high */
	status = loopBack(pf, 0, CONTINUE_DATA_3);
	highBaudOnly = 0;		/* Restore normal protocol selection */

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 3,
		CMD_CTSFL_DSAB, CMD_DTRFL_DSAB, CMD_CTS_REP);

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);

	return status;
}


/* ---------------------------------------
 * break_loop()
 * 
 * Test break-generation & detection
 *
 * If verbose > 1, displays results per
 * channel as progresses.
 * Returns 0 if no error, else returns -1
 * Puts results for each tested channel in
 * the pf structure:
 * Note, once an error is detected on a particular
 * channel, it "drops out";  The test progresses as
 * long as there is a channel left to test.
 *
 *	f_passfail:   The pass it totally failed on
 *  f_wstring[]: Number of passes successful on each channel.
 *					which failed (otherwise, went the distance)
 *  f_rstring[]: Results per channel:
 *		0 = successful
 *		L_ERR_???	- error condition
 *----------------------------
 */

int break_loop(pf)
pfunk pf;
{
int pass, i, anytested;
fpchControl fpch;
int retries;
int perfect = 1;
int lfyet = 0;
int anymissing;
int onlyOne;
i2ChanStrPtr pCh;
int status;
int okChange;
i2eBordStrPtr pB = pf->f_pB;

	onlyOne = (freezeBaud >= 0 && freezeStops >= 0 &&
	    	freezeParity >= 0 && freezeSize >= 0 && freezeFlow >= 0);

	/* If all protocols are frozen, we do not need to re-select 
	 * protocols between passes, therefore should do it
	 * only once here.
	 */

	if (onlyOne)
	{
		moreFromHigh = NO_HIGH_LIMIT;	/* Used when choosing protocol */
		for (fpch = fpchTable, i = nfpchTable;
			i; fpch++, i--)
		{
			if (fpch->flags & CONTINUE_BREAKS)
			{
				chooseProtocol(fpch, 1);
				assertProtocol(fpch, fpch->c_pCh);
			}
		}
		if (mainVerbosity & V_PROTOCOL)
		{
			if (!lfyet)
			{
				printf("\n");
				lfyet = 1;
			}
			printProtocolGlobal(fpchTable);
		}
	}


	okChange = 0;
	for (pass = 1; pass <= pf->f_passes; pass++)
	{
		moreFromHigh = NO_HIGH_LIMIT;	/* Used when choosing protocol */
		if (!okChange--) okChange = 4;	/* Change protocol only every 4th pass */
		/* For each channel, clear break count, send breaks */

		for(fpch = fpchTable, i = nfpchTable, anytested = 0; 
			i; fpch++, i--)
		{
			if (!(fpch->flags & CONTINUE_BREAKS)) continue;

			/* Select specified (or random) protocol */

			pCh = fpch->c_pCh;

			if (!onlyOne && okChange)
			{
				chooseProtocol(fpch, 0);
				assertProtocol(fpch, pCh);
				if (mainVerbosity & V_PROTOCOL) 
				{
					if (!lfyet)
					{
						printf("\n");
						lfyet = 1;
					}
					printProtocol(fpch);
				}
			}

			anytested = 1;

			/* Clear any latched break condition */
			i2GetStatus(pCh, I2_BRK);

			/* Send a single break, followed by a 10ms delay */

			status = i2QueueCommands(PTYPE_INLINE, pCh, 10, 
				2, CMD_SEND_BRK(250), CMD_PAUSE(1) );

			if (status == 0)
			{
				rip(10, "Timeout trying to send break");
			}
		}

		if (!anytested) break;	/* Stop when no one left to test */

		/* One second should be plenty long */
		retries = 100;

		/* Wait until all breaks received */
		/* Keep track of the total per channel detected */

		while(--retries)
		{
			for(fpch = fpchTable, i = nfpchTable , anymissing = 0,
				anytested = 0; i; fpch++, i--)
			{
				if (!(fpch->flags & CONTINUE_BREAKS)) continue;

				anytested = 1;

				if (!(i2GetStatus(fpch->c_pCh, 0) & I2_BRK))
				{
					anymissing = 1;
				}
			}
			if (!anymissing) break;

			iiDelay(pB,10);
		}

		/* See if the number of breaks we have counted, match! */
		for(fpch = fpchTable, i = nfpchTable ;
			i; fpch++, i--)
		{
			if (!(fpch->flags & CONTINUE_BREAKS)) continue;
			if(!(i2GetStatus(fpch->c_pCh, 0) & I2_BRK) )
			{
				if (mainVerbosity & V_ERROR) 
				{
					if (!lfyet)
					{
						printf("\n");
						lfyet = 1;
					}
					eprintf("Break loopback channel %2d at pass %d:  1 generated, 0 detected\n",
						fpch->thisChan, pass);	
				}
				pf->f_wstring[fpch->thisChan] = pass;
				pf->f_rstring[fpch->thisChan] = L_ERR_BREAKS;
				fpch->flags &= ~CONTINUE_BREAKS;
				perfect = 0;
			}
		}
		if (!anytested) break;

	} /* End loop on passes */

	/* We've completed all the passes or broken out */
	if (perfect)
	{
		pf->f_error = OKAY;
	}
	else
	{
		pf->f_error = LOOPBACK_ERR;	
		pf->f_passfail = pass - 1;	/* if > f_passes, never totally failed */
	}
	return (pf->f_error == OKAY) ? 0 : 1;
}



/*------------------------------------------------------------
 * waitForDssInput(statusBit, goal, pCh)
 * 
 * This is a helper routine for genericDssLoop;
 * Waits up to 0.6 second for the status from channel at pCh
 * to become set to the value of goal.  
 * Returns true or false according to whether the input
 * came in in the allotted time.
 *------------------------------------------------------------
 */

static int
waitForDssInput(register int statusBit, int goal, i2ChanStrPtr pCh)
{
register int retries;
i2eBordStrPtr pB = boardInfo;

	/* Wait until input asserted */
	retries = 60;
	while(--retries)
	{
		if ((i2GetStatus(pCh, 0) & statusBit) == goal)
		{
			break;
		}
		iiDelay(pB, 10);
	}
	return retries;
}


/*-------------------------------------------------------------------------
 * genericDssLoop(pf, pAssert, pNegate, statusBit, 
 *			outputName, inputName, continueFlag, errorCode1, errorCode2)
 * 
 * pf
 *		pointer to test parameter block
 * pAssert, pNegage
 *		pointers to commands to assert and negate the output signal
 * statusBit
 *		the status bit corresponding to the input signal
 * outputName
 *		the name of the output signal
 * inputName
 *		the name of the input signal 
 * continueFlag
 *		the bit to clear if we aren't continuing
 * errorCode1
 *		the error code to report if there is a problem with part 1
 * errorCode2
 *		the error code to report if there is a problem with part 2
 *
 * Data-set loopback test:
 *
 * If verbose > 1, displays results per
 * channel as progresses.
 * Returns 0 if no error, else returns -1
 * Puts results for each tested channel in
 * the pf structure:
 * Note, once an error is detected on a particular
 * channel, it "drops out";  The test progresses as
 * long as there is a channel left to test.
 *
 *	f_passfail:   The pass it totally failed on
 *  f_wstring[]: Number of passes successful on each channel.
 *					which failed (otherwise, went the distance)
 *  f_rstring[]: Results per channel:
 *		0 = successful
 *		L_ERR_???	- error condition
 *-----------------------------------------------------------------------
 */

static int 
genericDssLoop(pfunk pf, cmdSyntaxPtr pAssert,cmdSyntaxPtr pNegate,
	int statusBit, char *outputName, char *inputName,
	int continueFlag, int errorCode1, int errorCode2)
{
int pass, i, anytested;
fpchControl fpch;
int multiplier;
int perfect = 1;
int channelNo;
int lfyet = 0;
i2eBordStrPtr pB = pf->f_pB;
unsigned char assert[N_CHANNELS+2];	/*  Maximum number of channels to test */
									/* Because of how index is used, the 
									 * elements 1 through N_CHANNELS are used.
									 */
cmdSyntaxPtr pOtherAssert;
cmdSyntaxPtr pOtherNegate;


	/* Which output is the "other" one? */
	if (pAssert == CMD_DTRUP)
	{
		pOtherAssert = CMD_RTSUP;
		pOtherNegate = CMD_RTSDN;
	}
	else
	{
		pOtherAssert = CMD_DTRUP;
		pOtherNegate = CMD_DTRDN;
	}

	/* Initialize an array which determines whether we will set or
	 * clear the output under test.  To start with, we will negate
	 * all the signals.
	 */
	for (i = 0; i < sizeof (assert); i++) assert[i] = 0;

	for (pass = 1; pass <= pf->f_passes; pass++)
	{
		/* Important to use 8 for multiplier max, because we look
		 * at the bits.
		 */
		for (multiplier = 0; multiplier < 8; multiplier++)
		{
			/* On some passes, toggle the outputs that aren't
			 * under test
			 */

			if (! (multiplier & 1) )
			{
				commandToAll(PTYPE_INLINE, 
					(multiplier & 2) ? pOtherAssert : pOtherNegate, 0);
			}

			/* Now set or clear each test signal according to the list */

			for(fpch = fpchTable, i = nfpchTable, anytested = 0;
				i; fpch++, i--)
			{
				if (!(fpch->flags & (continueFlag | PROVISIONAL))) continue;

				commandToOne(PTYPE_INLINE, fpch->c_pCh,
					assert[i] ? pAssert : pNegate, 0);
			}

			/* That's all we will send until we get some input.
			 * so flush it all immediately!
			 */
			serviceOutgoingFifo(pB);	/* That's all there will be */
			iiSendPendingMail(pB);		/* Flush it all to the board */
	
			for(fpch = fpchTable, i = nfpchTable, anytested = 0;
				i; fpch++, i--)
			{
				if (!(fpch->flags & (continueFlag | PROVISIONAL))) continue;

				if (waitForDssInput(statusBit, 
					assert[i] ? statusBit : 0, fpch->c_pCh))
				{
					anytested = 1;
				}
				else
				{
					channelNo = fpch->c_pCh->infl.hd.i2sChannel;
					if (mainVerbosity & V_ERROR) 
					{
						if (!lfyet)
						{
							printf("\n");
							lfyet = 1;
						}	
						if (assert[i])
						{
								eprintf("%s-%s channel %2d at pass %d -- %s high but %s low\n",
								outputName, inputName, channelNo, 
								pass, outputName, inputName);
						}
						else
						{
							eprintf("%s-%s channel %2d at pass %d -- %s low but %s high\n",
								outputName, inputName, channelNo, 
								pass, outputName, inputName);
						}
						holdIfNeeded();
					}
					pf->f_wstring[channelNo] = pass;
					if (assert[i])
					{
						pf->f_rstring[channelNo] = errorCode1;
					}
					else
					{
						pf->f_rstring[channelNo] = errorCode2;
					}

					/* Once we've failed twice, we don't try anymore */
					if (fpch->flags & PROVISIONAL)
					{
						fpch->flags &= ~(continueFlag | PROVISIONAL);
					}
					else
					{	
						fpch->flags &= ~continueFlag;
						fpch->flags |= PROVISIONAL;
					}
					perfect = 0;
				}
			}

			if (!anytested) break;

			/* Now we have tested all the inputs, try another pass */
					
			if ((multiplier & 3) == 3)
			{
				for (i = 0; i < sizeof (assert); i++) assert[i] = random(2);
			}
			else 
			{
				for (i = 0; i < sizeof (assert); i++) assert[i] ^= 1;
			}

		}	/* End loop on multiplier */

		if (!anytested) break;
	} /* End loop on passes */

	/* We've completed all the passes or broken out */
	if (perfect)
	{
		pf->f_error = OKAY;
	}
	else
	{
		pf->f_error = LOOPBACK_ERR;	
		pf->f_passfail = pass - 1;	/* if > f_passes, never totally failed */
	}
	return (pf->f_error == OKAY) ? 0 : 1;
}



/*----------------------------------------
 * dtr_dcd_loop()
 * 
 * Performs "generic data-set loopback test
 * on the selected pair of lines.
 *----------------------------------------
 */

int dtr_dcd_loop(pf)
pfunk pf;
{

	return genericDssLoop(pf, CMD_DTRUP, CMD_DTRDN, I2_DCD,
		"DTR", "DCD", CONTINUE_DCD, L_ERR_DCDDN, L_ERR_DCDUP);

}

/*----------------------------------------
 * rts_cts_loop()
 * 
 * Performs "generic data-set loopback test
 * on the selected pair of lines.
 *----------------------------------------
 */

int rts_cts_loop(pf)
pfunk pf;
{

	return genericDssLoop(pf, CMD_RTSUP, CMD_RTSDN, I2_CTS,
		"RTS", "CTS", CONTINUE_CTS, L_ERR_CTSDN, L_ERR_CTSUP);

}

/*----------------------------------------
 * dtr_dsr_loop()
 * 
 * Performs "generic data-set loopback test
 * on the selected pair of lines.
 *----------------------------------------
 */

int dtr_dsr_loop(pf)
pfunk pf;
{

	return genericDssLoop(pf, CMD_DTRUP, CMD_DTRDN, I2_DSR,
		"DTR", "DSR", CONTINUE_DSR, L_ERR_DSRDN, L_ERR_DSRUP);

}

/*----------------------------------------
 * rts_ri_loop()
 * 
 * Performs "generic data-set loopback test
 * on the selected pair of lines.
 *----------------------------------------
 */

int rts_ri_loop(pf)
pfunk pf;
{

	return genericDssLoop(pf, CMD_RTSUP, CMD_RTSDN, I2_RI,
		"RTS", "RI ", CONTINUE_RI, L_ERR_RIDN, L_ERR_RIUP);

}


/*----------------------------------------
 * dtr_cts_loop()
 * 
 * Performs "generic data-set loopback test
 * on the selected pair of lines.
 *----------------------------------------
 */

int dtr_cts_loop(pf)
pfunk pf;
{

	return genericDssLoop(pf, CMD_DTRUP, CMD_DTRDN, I2_CTS,
		"DTR", "CTS", CONTINUE_CTS_X, L_ERR_CTSDN, L_ERR_CTSUP);

}
/*----------------------------------------
 * rts_dcd_loop()
 * 
 * Performs "generic data-set loopback test
 * on the selected pair of lines.
 *----------------------------------------
 */

int rts_dcd_loop(pf)
pfunk pf;
{

	return genericDssLoop(pf, CMD_RTSUP, CMD_RTSDN, I2_DCD,
		"RTS", "DCD", CONTINUE_DCD_X, L_ERR_DCDDN, L_ERR_DCDUP);

}



/*---------------------------------------
 * Sends bookmark requests for all selected
 * channels, then waits (forever) for them
 * to drain.
 * No timeout on this, but ctrl-C will break us
 * out.
 *----------------------------------------
 */
static void
polyDrain()
{
fpchControl fpch;
int i;
int stillMore;

	for (fpch = fpchTable, i = nfpchTable; 
		i; i--, fpch++)
	{
		i2QueueCommands(PTYPE_INLINE, fpch->c_pCh, 1, 1, CMD_BMARK_REQ);
	}

	do{
		stillMore = 0;
		kbhit();
		for (fpch = fpchTable, i = nfpchTable; 
			i; i--, fpch++)
		{
			if (fpch->c_pCh->bookMarks >0) stillMore = 1;
			i2InputFlush(fpch->c_pCh);
		}
	} while(stillMore);
}


/*-----------------------------
 * Blasts data continuously to 
 * specified channels using random or
 * stated protocol.
 * Number of passes indicates number of
 * seconds 
 *------------------------------
 */

static uchar bbuffer[200] = "Operator entered no data!\r\n";

int data_blast(pf)
pfunk pf;
{
uchar far *fcp;
register uchar *cp;
int i, j;
fpchControl fpch;
time_t donetime, starttime, now, showtime;
long total = 0;
int lfyet = 0;
i2ChanStrPtr pCh;
int stillMore;
i2eBordStrPtr pB = pf->f_pB;

	/* Copy output data to each channel in use */

	moreFromHigh = NO_HIGH_LIMIT;
	for (fpch = fpchTable, i = nfpchTable; 
		i; i--, fpch++)
	{
		for (cp = bbuffer, fcp = fpch->charsOut; *cp; )
		{
			*fcp++ = *cp++;
		}
		*fcp++ = '\r';
		*fcp++ = '\n';
		fpch->nOut = 2 + (cp - bbuffer);
		fpch->iOut = 0;
		*fcp++ = 0;

		/* Select specified (or random) protocol */

		pCh = fpch->c_pCh;

		chooseProtocol(fpch, 0);
		assertProtocol(fpch, pCh);
		if (mainVerbosity & V_PROTOCOL) 
		{
			if (!lfyet)
			{
				printf("\n");
				lfyet = 1;
			}
			printProtocol(fpch);
		}
		i2SetOption(pCh, CO_NBLOCK_WRITE);	/* Select non-blocking writes */
	}

	/* Raise Data set signals to make any terminal happy */
	commandToAll(PTYPE_INLINE, CMD_DTRRTS_UP, 0);

	starttime = time(NULL);
	donetime = starttime + pf->f_passes;
	showtime = starttime + 5;

	while (donetime > time(NULL))
	{
		for (fpch = fpchTable, i = nfpchTable; 
			i; i--, fpch++)
		{
			/* Send as much of string as we can */

			pCh = fpch->c_pCh;

			fpch->iOut += i2Output(pCh, 
				(char far*) &(fpch->charsOut[fpch->iOut]),
				fpch->nOut - fpch->iOut);

			if (fpch->iOut == fpch->nOut)
			{
				fpch->iOut = 0;
				total += fpch->nOut;
				if (mainVerbosity & V_TIME)
				{
					now = time(NULL);
					if (showtime < now)
					{
						if (!lfyet)
						{
							printf("\n");
							lfyet = 1;
						}
						showtime = now + 5;
						printf("%ld sent...\r", total);
					}
				}

				/* In case there is a loop-back connected, try to
				 * keep input drained
				 */
				i2InputFlush(pCh);
				kbhit();
			}
		}
	}

	/* If we are reporting a time, we will wait for the output
	 * buffers to drain before reporting.
	 */

	if (mainVerbosity & V_TIME)
	{
		printf("\nDraining buffers...");
		do{
			stillMore = 0;
			for (fpch = fpchTable, i = nfpchTable; 
				i; i--, fpch++)
			{
				pCh = fpch->c_pCh;
				if (i2OutputFree(pCh) < 10) stillMore = 1;
			}
			kbhit();
		}
		while(stillMore);

		/* Drain the board buffers */
		printf("board...");
		polyDrain();

		printf("Total %ld sent in %ld seconds.\n", total, time(NULL) - starttime);
		
	}

	/* In case there is data at a low baud rate accumulated, flush it! */
	/* Also flush any input the drivers know about in case there were
	 * loopback plugs in place.*/


	for (fpch = fpchTable, i = nfpchTable; 
		i; i--, fpch++)
	{
		pCh = fpch->c_pCh;
		i2FlushOutput(pCh); 
		i2InputFlush(pCh);
	}	
		
	/* Continuing flushing input for 5.5 seconds!,
	 * DON'T REDUCE THIS BY MUCH!
	 * which should give any data which had looped back time
	 * to clear the board-buffers (which don't get flushed)
	 * and be transferred to the drivers.  This is done
	 * using several flushes rather than a long wait since
	 * the board input buffers might (?) be a lot larger
	 * than the buffers in the driver.
	 * If the board has N times as much space as the driver
	 * buffers, we need to do the flush N times.
	 */

	if (mainVerbosity & V_NORMAL)
	{
		printf("...draining input...");
	}

	for (j = 500; j ; j--)
	{
		flushAllInput();
		iiDelay(pB,11);	/* Gives board time to transfer more */
	}
	pf->f_error = OKAY;
	return 0;
}


/*----------------------------------
 * data_rate_test(pf)
 * Sends exactly 60 characters at
 * 50 baud to each selected channel.
 *----------------------------------
 */

char *rateData = 
"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ34567890";

int data_rate_test(pf)
pfunk pf;
{
int i, j;
fpchControl fpch;
time_t starttime, now;
register i2ChanStrPtr pCh;
int stillMore;
i2eBordStrPtr pB = pf->f_pB;

	/* Initialize the procotol in the tables*/
	for (fpch = fpchTable, i = nfpchTable; 
		i; i--, fpch++)
	{
		fpch->csize  = 0;	/* SEVEN-BIT CHARACTERS */
		fpch->flow   = 0;	/* No flow control! */
		fpch->parity = 2;	/* Even parity */
		fpch->stops  = 0;	/* 0 stop bits */

		pCh = fpch->c_pCh;
		i2ClrOption(pCh, CO_NBLOCK_WRITE);	/* Select blocking writes */
		fpch->baud   = 0; 			/* 50 baud */
		if (mainVerbosity & V_PROTOCOL && i == nfpchTable) 
		{
			printf("\n");
			printProtocolGlobal(fpch);
		}
		assertProtocol(fpch, pCh);	/* The initial baud rate needs to be right */
		i2InputFlush(pCh);	/* In case there is any input that would clog us */
	}
	polyDrain();	/* Wait for the commands(!) to drain */

	starttime = time(NULL);

	/* Send the characters to every selected port */
	for (fpch = fpchTable, i = nfpchTable; 
		i; i--, fpch++)
	{
		pCh = fpch->c_pCh;

		if (i2Output(pCh, (char far*) rateData, 60) != 60)
		{
			printf("\n--Internal Logic Error ---\n");
		}
	}

	/* Ensure the host buffers are mostly clear */
	do{
		stillMore = 0;
		for (fpch = fpchTable, i = nfpchTable; 
			i; i--, fpch++)
		{
			pCh = fpch->c_pCh;
			if (i2OutputFree(pCh) < 10) stillMore = 1;
			i2InputFlush(pCh);	
		}
		kbhit();
	}
	while(stillMore);

	polyDrain();	/* Wait for all the data to drain, even from board */

	now = time(NULL) - starttime;

	/* Continuing flushing input to guarantee board is clear! */

	if (mainVerbosity & V_NORMAL)
	{
		printf("...draining input...");
	}

	for (j = 100; j ; j--)
	{
		flushAllInput();
		iiDelay(pB,11);	/* Gives board time to transfer more */
	}
	if (now > 13 || now < 11)
	{
		if (mainVerbosity & (V_TIME | V_ERROR))
		{
			eprintf("\n -- Error: baud rate is too %s: %d sent in %d seconds\n",
				now > 13 ? "slow" : "fast",
				60, now);
			holdIfNeeded();
		}
		pf->f_error = BAD_BAUD_ERR;
		return -1;
	}

	/* Drop through == no error */
	if (mainVerbosity & V_TIME)
	{
		printf("\n -- %d characters sent to each channel in %d seconds\n",
			60, now);
	}	
	
	pf->f_error = OKAY;
	return 0;
}



char assert1[] =	{0x1b, 'c', '0', 	/* Select the operating parameters */
					'4',				/* 2400 baud */
					'0',				/* 1 stop bit */
					'3',				/* Even parity */
					'0',				/* 7-bit characters */
};

char assert2[] =	{0x1b, 'c', '0', 	/* Select the operating parameters */
					'2',				/* 9600 baud */
					'0',				/* 1 stop bit */
					'3',				/* Even parity */
					'0',				/* 7-bit characters */
};

char assert3[] =	{0x1b, 'c', '0', 	/* Select the operating parameters */
					'1',				/* 19200 baud */
					'0',				/* 1 stop bit */
					'3',				/* Even parity */
					'0',				/* 7-bit characters */
};

char assert4[] =	{0x1b, 'c', '0', 	/* Select the operating parameters */
					'0',				/* 38400 baud */
					'0',				/* 1 stop bit */
					'3',				/* Even parity */
					'0',				/* 7-bit characters */
};

/* -- These are supported on the Link MC6, which also supports the above --*/
char assert5[] =	{0x1b, 'c', '0', 	/* Select the operating parameters */
					'B',				/* 76800 baud */
					'0',				/* 1 stop bit */
					'3',				/* Even parity */
					'0',				/* 7-bit characters */
};

char assert6[] =	{0x1b, 'c', '0', 	/* Select the operating parameters */
					'D',				/* 115200 baud */
					'0',				/* 1 stop bit */
					'3',				/* Even parity */
					'0',				/* 7-bit characters */
};

char *assertions[] = {assert1, assert2, assert3, assert4, assert5, assert6};

char *displayStrings[] = {
	"\r\nThese characters were sent at   2400 baud\r\n",
	"\r\nThese characters were sent at   9600 baud\r\n",
	"\r\nThese characters were sent at  19200 baud\r\n",
	"\r\nThese characters were sent at  38400 baud\r\n",
	"\r\nThese characters were sent at  76800 baud\r\n",
	"\r\nThese characters were sent at 115200 baud\r\n",
};

char barbaraPole[]="ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789abcdefghijklmnopqrstuvwxyz                ";

int baudList[] = {11, 15, 16, 17, 21, 22};	/* Indices into p_baud[] */

int cyclesList[] = {4, 12, 16, 20, 22, 22};

/*-----------------------------
 * wy60_blast()
 * link_mc6_blast()
 * 		variable_blast()
 *
 * Blasts data continuously to
 * specified channels at special
 * selected baud rates.
 * Sends Wy-60 (link MC6) compatible messages to
 * change the terminal baud rate.
 * Number of passes indicates number of
 * seconds.
 * The Terminal should start out at 2400 baud.
 * The test changes it from there...
 *------------------------------
 */

int wy60_blast(pfunk pf)
{
	return variable_blast(pf, 4);
}


int link_mc6_blast(pfunk pf)
{
	return variable_blast(pf, 6);
}


int variable_blast(pf, nrates)
pfunk pf;
int nrates;		/* Will be 4 for WY60, 6 for link MC6 */
{
int i, j;
fpchControl fpch;
time_t donetime, starttime;
int lfyet = 0;
i2ChanStrPtr pCh;
i2eBordStrPtr pB = pf->f_pB;
int whichProtocol = 0;

	/* Initialize the procotol in the tables*/
	for (fpch = fpchTable, i = nfpchTable; 
		i; i--, fpch++)
	{
		fpch->csize  = 0;	/* SEVEN-BIT CHARACTERS */
		fpch->flow   = 1;	/* Xon/Xoff */
		fpch->parity = 2;	/* Even parity */
		fpch->stops  = 0;	/* 0 stop bits */

		pCh = fpch->c_pCh;
		i2ClrOption(pCh, CO_NBLOCK_WRITE);	/* Select blocking writes */
		fpch->baud   = baudList[whichProtocol];
		assertProtocol(fpch, pCh);	/* The initial baud rate needs to be right */
	}

	/* Raise Data set signals to make any terminal happy */
	commandToAll(PTYPE_INLINE, CMD_DTRRTS_UP, 0);

	starttime = time(NULL);
	donetime = starttime + pf->f_passes;

	/* Go until time elapses, but stay in until
	 * we've set the protocol back to the starting point.
	 * In this way, unattended multiple passes (P argument)
	 * will work properly.
	 */

	while (donetime > time(NULL) || whichProtocol != 1)
	{
		/* Make sure all the old data is all sent
		 * before continuing.  This prevents our not having
		 * enough buffers for commands, etc.
		 * Note: this implies if one terminal flow-controls,
		 * all will wait here for him!  
		 */

		polyDrain(); 

		/* Assert the baud rate and send b/r-changing string
		 * for each channel.
		 */

		for (fpch = fpchTable, i = nfpchTable; 
			i; i--, fpch++)
		{
			fpch->baud   = baudList[whichProtocol];
		
			pCh = fpch->c_pCh;

			/* Send the string to put us in the new protocol,
			 * before we change the protocol.  We put a little
			 * pause before and after to guarantee recovery.
			 */

			while (i2OutputFree(pCh) < 20)
			{
				i2InputFlush(pCh);	/* Drain any input that might be there */
			}
				
			i2Output(pCh, (char far *)assertions[whichProtocol], 7);

			/* After we tell the terminal what the new rate is to be,
			 * we give him time to actually change his rate before
			 * sending him any more data.  The pause here is to ensure
			 * he has plenty of time.
			 * Then we will change our own baud rate and start sending
			 * him the new data immediately
			 */
			i2QueueCommands(PTYPE_INLINE, pCh, 1, 1, CMD_PAUSE(20) );

			/* The firmware will wait for the last character
			 * to drain before really asserting the protocol.
			 */
			assertProtocol(fpch, pCh);
			if (mainVerbosity & V_PROTOCOL && i == nfpchTable) 
			{
				if (!lfyet)
				{
					printf("\n");
					lfyet = 1;
				}
				printProtocolGlobal(fpch);
			}
		}

		flushAllInput();
		for (fpch = fpchTable, i = nfpchTable; 
			i; i--, fpch++)
		{
			pCh = fpch->c_pCh;

			while (i2OutputFree(pCh) < 60)
			{
				i2InputFlush(pCh);	/* Drain any input that might be there */
			}
				
			i2Output(pCh, (char far *)displayStrings[whichProtocol], 45);
		}

		/* Some barber-poles to all channels */
		for (j = cyclesList[whichProtocol]; --j;)
		{
			flushAllInput();
			for (fpch = fpchTable, i = nfpchTable; 
				i; i--, fpch++)
			{
				pCh = fpch->c_pCh;
				while (i2OutputFree(pCh) < 86)
				{
					i2InputFlush(pCh);	/* Drain input that might be there */
				}
				i2Output(pCh, (char far *)barbaraPole, 79);
			}
			flushAllInput();
		}

		/* Select next protocol... */
		if (++whichProtocol == nrates) whichProtocol = 0;
	}

	if (mainVerbosity & V_NORMAL)
	{
		printf("...draining input...");
	}

	for (j = 500; j ; j--)
	{
		flushAllInput();
		iiDelay(pB,11);	/* Gives board time to transfer more */
	}
	pf->f_error = OKAY;
	return 0;
}

	

/*--------------------------
 * Continuously toggles DTR 
 * for tracing purposes.
 *--------------------------
 */

int dtr_blast(pf)
pfunk pf;
{
time_t donetime;
int i;
i2eBordStrPtr pB = pf->f_pB;

	donetime = time(NULL) + pf->f_passes;

	while (donetime > time(NULL))
	{
		for (i = 10; i; i--)
		{
			/* Raise DTR, 10 mS , Drop DTR, 30 mS */

			commandToAll(PTYPE_INLINE,CMD_DTRUP, 1);
			commandToAll(PTYPE_INLINE,CMD_PAUSE(1), 1);
			commandToAll(PTYPE_INLINE,CMD_DTRDN, 1);
			commandToAll(PTYPE_INLINE,CMD_PAUSE(3), 1);

		}
		iiDelay (pB,410);	/* Should be enough delay for the above to clear */
						/* 10 * (10 + 30) = 400 */
						/* So our delay is slightly longer to allow for
						 * tolerances
						 */
	}
	pf->f_error = OKAY;
	return 0;
}


/*--------------------------
 * Continuously toggles RTS 
 * for tracing purposes.
 *--------------------------
 */

int rts_blast(pf)
pfunk pf;
{
time_t donetime;
int i;
i2eBordStrPtr pB = pf->f_pB;

	donetime = time(NULL) + pf->f_passes;

	while (donetime > time(NULL))
	{
		for (i = 10; i; i--)
		{
			/* Raise RTS, 10mS, Drop RTS, 20mS */

			commandToAll(PTYPE_INLINE,CMD_RTSUP, 1);
			commandToAll(PTYPE_INLINE,CMD_PAUSE(1), 1);
			commandToAll(PTYPE_INLINE,CMD_RTSDN, 1);
			commandToAll(PTYPE_INLINE,CMD_PAUSE(2), 1);
		}
		iiDelay (pB,310);	/* Should be enough delay for the above to clear */
						/* 10 * (10 + 20) = 300 */
						/* So our delay is slightly longer to allow for
						 * tolerances.
						 * Note that the raising/lowering of the signal
						 * itself should take on the order of microseconds
						 * to process from a full queue.
						 */
	}
	pf->f_error = OKAY;
	return 0;
}	


/*------------------------------------
 * Reads data from input, echos it.
 *------------------------------------
 */

int data_echo(pf)
pfunk pf;
{
int i, j;
fpchControl fpch;
time_t donetime, starttime, now, showtime;
long total = 0;
int lfyet = 0;
i2ChanStrPtr pCh;

	/* Select specified (or random) protocol */

	for (fpch = fpchTable, i = nfpchTable; 
		i; i--, fpch++)
	{
		pCh = fpch->c_pCh;

		chooseProtocol(fpch, 0);
		assertProtocol(fpch, pCh);
		if (mainVerbosity & V_PROTOCOL) 
		{
			if (!lfyet)
			{
				printf("\n");
				lfyet = 1;
			}
			printProtocol(fpch);
		}
		i2SetOption(pCh, CO_NBLOCK_WRITE);	/* Select non-blocking writes */
		fpch->nOut = 0;		/* Keeps track of input characters not yet written */
	}

	/* Raise Data set signals to make any terminal happy */
	commandToAll(PTYPE_INLINE, CMD_DTRRTS_UP, 0);

	starttime = time(NULL);
	donetime = starttime + pf->f_passes;
	showtime = starttime + 5;

	while (donetime > time(NULL))
	{
		for (fpch = fpchTable, i = nfpchTable; 
			i; i--, fpch++)
		{

			pCh = fpch->c_pCh;

			/* Get more data if all sent */

			if (fpch->nOut == 0)
			{
				j = i2Input(pCh, (char far *) fpch->charsOut,
					sizeof(fpch->charsOut));
				fpch->nOut = (j < 0) ? 0 : j;
				fpch->iOut = 0;
			}

			/* Send as much of string as we can */

			if (fpch->nOut)
			{
				fpch->iOut += i2Output(pCh, 
					(char far*) &(fpch->charsOut[fpch->iOut]),
					fpch->nOut - fpch->iOut);
			}

			if (fpch->iOut == fpch->nOut)
			{
				total += fpch->nOut;
				fpch->iOut = 0;
				fpch->nOut = 0;
				if (mainVerbosity & V_TIME)
				{
					now = time(NULL);
					if (showtime < now)
					{
						if (!lfyet)
						{
							printf("\n");
							lfyet = 1;
						}
						showtime = now + 5;
						printf("%ld echoed...\r", total);
					}
				}
			}
		}
		kbhit();
	}

	pf->f_error = OKAY;
	return 0;
}



/*-----------------------------------------------
 * Allocates and creates a channel control table
 * for use by any tests which work "by channel"
 *-----------------------------------------------
 */

static void
initChannelTable(unsigned short flag)
{
pfunk pf = argumentList;
register fpchControl fpch;
i2ChanStrPtr pCh;
int i,j;
ushort chan;

	fpch = fpchTable;
	pCh = chanInfo;
	for (i = 0; i < 4; i++)
	{
		chan = pf->f_channels[i];
		for (j = 16; j ; pCh++, j--, chan >>=1)
		{
			if (chan & 1) 
			{
				fpch->flags = flag;
				fpch->c_pCh = pCh;
				fpch++;
			}	
		}
	}
}

void
allocChannelTable()
{
register pfunk pf = argumentList;
int i, j; 
ushort chan;
ushort flag = CONTINUE_ANY;

	/* Determine how many channels were selected */
	j = 0;
	for (i = 0; i < 4; i++)
	{
		for (chan = pf->f_channels[i]; chan ; chan >>=1)
		{
			if (chan & 1) j++;
		}
	}
	if (j)
	{
		nfpchTable = j;		/* This determines size of table */
		speculate = 0;
		validated = 0;
	}
	else
	{
		nfpchTable = 64;
		speculate = 1;
		validated = 1;	/* If we are speculating, no validation needed */
		flag |= SPECULATION;
	}

	/* If table already exists, kill it and start over:
	 * maybe the size is different?
	 */

	if (fpchTable != (fpchControl) 0)
	{
		farfree((char far *)fpchTable);
	}

	/* Allocate the table */

	fpchTable = (fpchControl) farmalloc(nfpchTable * sizeof (chControl));

	if (fpchTable == (fpchControl) 0)
	{
		rip(10, "Unable to allocate memory for tables");
	}

	initChannelTable(flag);
}

	

/*------------------------------------------------------------
 * If we are speculating on the number of channels there are,
 * endSpeculation() determines this and (possibly) prints the
 * map of what we are testing.
 *------------------------------------------------------------
 */
void
endSpeculation()
{
register pfunk pf;
fpchControl fpch ;
int i, j; 
ushort chan;
ushort *channelArray;
i2ChanStrPtr pCh;

	if (!speculate) return;
		
	speculate = 0;	/* Don't need to speculate twice! */

	pf = argumentList;
	channelArray = pf->f_pB->i2eChannelMap;

	for (i = realTests; i; i--, pf++)
	{
		for (j = 0; j < 4; j++)
		{
			pf->f_channels[j] = channelArray[j];
		}
	}

	pCh = chanInfo;
	fpch = fpchTable;
	pf = argumentList;
	for (i = 0; i < 4; i++)
	{
		chan = pf->f_channels[i];
		for (j = 16; j ; pCh++, j--, chan >>=1)
		{
			if (chan & 1) 
			{
				fpch->flags = CONTINUE_ANY;
				fpch->c_pCh = pCh;
				fpch++;
			}	
		}
	}
	nfpchTable = fpch - fpchTable;

	if (nfpchTable == 0) rip(16,"Board reports it has no channels!\n");

	if (mainVerbosity & V_NORMAL)
	{
		printf("\n -- Board reports %d channels to test --\n", nfpchTable);
		if (mainVerbosity & V_CHANNELS)
		{
			printChanMap(argumentList[0].f_channels);
		}
	}
}
	

/*------------------------------------------------------------
 * validateChannels()
 *
 * If we have specified some channels, we need to verify 
 * that all the specified channels are on the board,
 * or else we will have an error
 * 
 *------------------------------------------------------------
 */
void
validateChannels()
{
register pfunk pf = argumentList;
int i, j, k, m; 
ushort chan;
ushort *channelArray;
int first = 1;

	if (validated) return;
		
	validated = 1;	/* Don't need to validate twise */

	/* -- In future we will use the library to obtain the proper map -- */
	/*    for now, we will pertend we have found an 8-port board */

	channelArray = pf->f_pB->i2eChannelMap;

	chan = 0;
	for (i = 0; i < 4; i++)
	{
		j = pf->f_channels[i];
		k = channelArray[i];
		for (m = 16; m; 
				m--, j >>=1, k>>=1, chan++)
		{
			if ( (j & 1) && !(k & 1) )
			{
				if (first)
				{
					first = 0;
					eprintf("\n --Channels specified but not reported present by hardware:\n");
				}
				printf("%d, ",chan);
			}
		}
	}
	if (!first)
	{
		rip(12,"\n\n -- Terminated test\n");
	}	
}



/*--------------------------------------
 * Routines to choose, assert, and print
 * protocols for a given channel
 *--------------------------------------
 */

/*-------------------------------
 * printProtocol
 *
 * given a pointer to a chControl, 
 * prints the protocol stored
 *--------------------------------
 */

void
printProtocol(fpch)
fpchControl fpch;
{

	printf("Channel %2d: %s baud %s %s parity %s %s flow ctrl.\n",
		fpch->thisChan,
		p_baud[fpch->baud].proDesc,
		p_size[fpch->csize].proDesc,
		p_parity[fpch->parity].proDesc,
		p_stops[fpch->stops].proDesc,
		p_flow1[fpch->flow].proDesc );
}

/*-------------------------------
 * printProtocolGlobal
 *
 * given a pointer to a chControl, 
 * prints the protocol stored
 * But does not indicate channel.
 *--------------------------------
 */

void
printProtocolGlobal(fpch)
fpchControl fpch;
{
	printf("All using %s baud %s %s parity %s %s flow ctrl.\n",
		p_baud[fpch->baud].proDesc,
		p_size[fpch->csize].proDesc,
		p_parity[fpch->parity].proDesc,
		p_stops[fpch->stops].proDesc,
		p_flow1[fpch->flow].proDesc );
}

/*---------------------------------
 * assertProtocol(fpch, pCh)
 * Given a pointer to a chControl,
 * asserts the protocol stored.
 *---------------------------------
 */


void
assertProtocol(fpchControl fpch, i2ChanStrPtr pCh)
{
int status;


	status = i2QueueCommands(PTYPE_INLINE, pCh, 100, 6, 
		CMD_SETBAUD(p_baud[fpch->baud].proParm),
		CMD_SETBITS(p_size[fpch->csize].proParm),
		CMD_SETPAR(p_parity[fpch->parity].proParm),
		CMD_SETSTOP(p_stops[fpch->stops].proParm),
		CMD_IXON_OPT(p_flow1[fpch->flow].proParm),
		CMD_OXON_OPT(p_flow2[fpch->flow].proParm) );

	if (status == 0)
	{
		rip(10, "Timeout asserting protocol");
	}
	if (status < 0)
	{
		rip(10, "Error asserting protocol");
	}

}

void
chooseProtocol(fpch,canflow)
fpchControl fpch;
ushort canflow;		/* clear if we must not use flow control */
{
int bias;

	/* Formerly only chose high baud rates beyond the second pass,
	 * but this is not needed now:  At lower baud rates, we will
	 * send fewer characters to make up for things
	 * Now, instead, there is logic to prevent us from selecting
	 * too many fast baud rates at once.
	 */

	if (freezeBaud >= 0)
	{
		fpch->baud = freezeBaud;
	}
	else if (highBaudOnly)
	{
		fpch->baud = ultra_rates[random(N_ULTRA_INDICES)];
	}
	else
	{
		if (	nfpchTable < moreFromHigh) bias = 2;
		else if (nfpchTable < (moreFromHigh << 1)) bias = 3; 
		else if (nfpchTable < (moreFromHigh << 2)) bias = 4;
		else bias = 8;

		if (moreFromHigh <= 0 || random(bias) )
		{
			fpch->baud = low_rates[random(N_LOW_INDICES)];
		}
		else
		{
			fpch->baud = high_rates[random(N_HIGH_INDICES)];
			moreFromHigh--;
		}
	}
		
	fpch->csize = freezeSize < 0 ?
		random(N_SIZE) : freezeSize;
	fpch->parity = freezeParity < 0 ?
		random(N_PARITY) : freezeParity;
	fpch->stops = freezeStops < 0 ?
		random(N_STOPS) : freezeStops;
	if (freezeFlow >= 0)
	{
		fpch->flow = freezeFlow;
	}
	else if (canflow)
	{
		fpch->flow = random(N_FLOW);
	}
	else
	{
		fpch->flow = 0;
	}
}

	
/*--------------------------------
 * commandToAll(type, pCs, forever)
 *
 * pCs 
 *		pointer to a command syntax structure
 * type
 *		PTYPE_BYPASS or PTYPE_INLINE
 * forever
 *		if false, will time out after 1 second.
 *
 * Sends the command represented by
 * count, string to all the active channels
 *---------------------------------
 */

void
commandToAll(int type, cmdSyntaxPtr pCs, int forever)
{
fpchControl fpch;
int i;

	for (fpch = fpchTable, i = nfpchTable;
		i; fpch++, i--)
	{
		if (fpch->flags & CONTINUE_ANY)
		{
			commandToOne(type, fpch->c_pCh, pCs, forever);
		}
	}
}

/*-------------------------------------------------
 * flushAllInput()
 *
 * Flushes all input from any channels under test.
 *-------------------------------------------------
 */

void
flushAllInput()
{
register fpchControl fpch;
register int i;

	for (fpch = fpchTable, i = nfpchTable; 
		i; i--, fpch++)
	{
		i2InputFlush(fpch->c_pCh);	/* Drain any input that might be there */
	}
}
	
/*--------------------------------
 * commandToOne(type, pCh, pCs, forever)
 *
 * pCs 
 *		pointer to a command syntax structure
 * type
 *		PTYPE_BYPASS or PTYPE_INLINE
 * pCh
 *		pointer to a channel structure
 * forever
 *		if false, will time out after 1 second.
 *
 * Sends the command represented by
 * count, string to the selected channel
 *---------------------------------
 */

void
commandToOne(int type, i2ChanStrPtr pCh, cmdSyntaxPtr pCs, int forever)
{
int status;
char buffer[128];

	status = i2QueueCommands(type,  pCh, 
		(forever ? -1 : 100), 1, pCs);

	if (status <= 0)
	{
		if (status == 0)
		{
			sprintf(buffer,"Timeout sending %s command.\n",
				(type == PTYPE_BYPASS) ? "bypass" : "in-line");
		}
		else
		{
			sprintf(buffer,"Error sending %s command.\n",
				(type == PTYPE_BYPASS) ? "bypass" : "in-line");
		}
		rip(10,buffer);
	}
}



void
printChanMap(unsigned short *pu)
{
int k, j, i;

	eprintf("0       8        0       8        0       8        0       8\n");
	for(i = 0; i < 4; i++)
	{
		for(k = 16, j = *pu++;
			k; k--, j >>= 1)
		{
			eprintf((j & 1) ? "Y" :"n");
		}
		eprintf(" ");
	}
	eprintf("\n");
}
	

/*-------------------------------------------
 * showDiscreprencies()
 * Displays data written, and data received
 * for the loopback test.
 *--------------------------------------------
 */
void 
showDiscreprencies(fpch, channel, message,
		nout, charsout, nin, charsin)
fpchControl fpch;
register uchar far *charsout;
register uchar far *charsin;
int channel;
char *message;
int nout, nin;
{
uchar c;

	if (!(mainVerbosity & V_PROTOCOL)) 
	{
		printProtocol(fpch);
	}
	eprintf("%s on channel %d\n",
		message, channel);
	
	charsout[nout] = 0;
	eprintf("Sent %2d: %Fs\n", nout, (char far *)charsout);
	charsin[nin] = 0;

	/* Print the characters coming back this way so any nulls or
	 * control characters received will be mapped into printables
	 */
	eprintf("Got  %2d: ", nin);
	while(nin--) 
	{
		c = *charsin++;
		eprintf("%c", c < ' ' ? (c | 0x80) : c);
	}
	eprintf("\n");
	holdIfNeeded();
}
	
/*-------------------------------------------------------
 * farCompare(count, source1, source2)
 *
 * returns 1 if strings are different (up to the count)
 * returns 0 otherwise.
 *------------------------------------------------------
 */

int farCompare(count, source1, source2)
register uchar far *source1;
register uchar far *source2;
register int count;
{
	while(count--)
	{
		if (*source1++ != *source2++) return 1;
	}
	return 0;
}


#define TEST_STRING_SIZE 70
/*--------------------------------
 * mapSpeedToSize(rateIndex)
 *
 * Returns the proper number of
 * characters to send, based on the
 * rate (in p_baud) indexed by
 * rateIndex
 *---------------------------------
 */

int
mapSpeedToSize(int rateIndex)
{
	switch (rateIndex)
	{
	case 0:		return 4;	/* 50 baud */
	case 1:		return 6;	/* 75 baud */
	case 2:		return 8;	/* 110 */
	case 3:		return 8;	/* 134.5 */
	case 4:		return 10;	/* 150 */
	case 5:		return 10;	/* 200 */
	case 6:		return 15;	/* 300 */
	case 7:		return 30;	/* 600 */
	case 8:		return 60;	/* 1200 */
	default:	return TEST_STRING_SIZE;
	}
}

/*----------------------------
 * loopBack(pf,torture)
 * pf = argument structure
 * 
 * Master Loop-back test
 * (Used by a number of tests)
 *
 * If verbose > 1, displays results per
 * channel as progresses.
 * Returns 0 if no error, else returns -1
 * Puts results for each tested channel in
 * the pf structure:
 * Note, once an error is detected on a particular
 * channel, it "drops out";  The test progresses as
 * long as there is a channel left to test.
 *
 *	f_passfail:   The pass it totally failed on
 *  f_wstring[]: Number of passes successful on each channel.
 *					which failed (otherwise, went the distance)
 *  f_rstring[]: Results per channel:
 *		0 = successful
 *		L_ERR_???	- error condition
 *----------------------------
 */

int 
loopBack(pf, torture, whichContinue)
pfunk pf;
ushort torture;
ushort whichContinue;
{
fpchControl fpch;
int i, anytested, anyincome;
int pass;
int onlyOne;
uchar *cp;
uchar far *fcp;
int j, anymissing;
time_t donetime = 0;
int perfect = 1;
int high = 0;
int retries;
int lfyet = 0;
int lf2yet = 0;
i2ChanStrPtr pCh;
int okChange;
i2eBordStrPtr pB = pf->f_pB;
int corrupt = 0;

	for (cp = pf->f_rstring, i = 64;
		i ; i--)
	{
		*cp++ = 0;
	}
	
	/* If all four protocol elements are specified, be are frozen */

	onlyOne = (freezeBaud >= 0 && freezeStops >= 0 &&
	    	freezeParity >= 0 && freezeSize >= 0 && freezeFlow >= 0);

	/* If all protocols are frozen, we do not need to re-select 
	 * protocols between passes, therefore should do it
	 * only once here.
	 */

	if (onlyOne)
	{
		moreFromHigh = torture ? 0 : LOOP_HIGH_LIMIT;
		for (fpch = fpchTable, i = nfpchTable;
			i; fpch++, i--)
		{
			pCh = fpch->c_pCh;
			if (fpch->flags & whichContinue)
			{
				chooseProtocol(fpch, 1);
				assertProtocol(fpch, pCh);
				i2SetOption(pCh, CO_NBLOCK_WRITE);
			}
		}
		if (mainVerbosity & V_PROTOCOL)
		{
			printf("\n");
			printProtocolGlobal(fpchTable);
		}
	}
		
	/* If MM option was used, clear any error-statuses before sending data */
	if (statParity)
	{
		for(fpch = fpchTable, i = nfpchTable, anytested = 0;
			i; fpch++, i--)
		{
				i2GetStatus(fpch->c_pCh, (I2_OVR | I2_PAR | I2_FRA) );
		}
	}
	okChange = 0;
	for (pass = 1; pass <= pf->f_passes; pass++)
	{
		if (!okChange--) okChange = 4;	/* Change protocol only every 4th pass */
		moreFromHigh = torture ? 0 : LOOP_HIGH_LIMIT;

		/* Assert protocols (if need be) for each channel */
		for(fpch = fpchTable, i = nfpchTable;
			i; fpch++, i--)
		{
			if(!(fpch->flags & whichContinue)) continue;

			/* If protocols are not all frozen, we pick a new
			 * one for each pass
			 */

			pCh = fpch->c_pCh;
				
			if (!onlyOne && okChange)
			{
				chooseProtocol(fpch, 1);
				assertProtocol(fpch, pCh);
				i2SetOption(pCh, CO_NBLOCK_WRITE);
				if (mainVerbosity & V_PROTOCOL) 
				{
					if (!lfyet)
					{
						lfyet = 1;
						printf("\n");
					}
					printProtocol(fpch);
				}
			}
		}

		/* Generate test data for all channels */

		for(fpch = fpchTable, i = nfpchTable ;
			i; fpch++, i--)
		{
			if(!(fpch->flags & whichContinue)) continue;

			if (forceStringLength)
			{
				cp = (unsigned char *)argString;
				for (j = forceStringLength, fcp = fpch->charsOut;
					j; j--)
				{
					*fcp++ = *cp++;
				}
				*fcp = 0;

				fpch->nOut = forceStringLength;
			}
			else
			{
				cp = fpch->csize ?
					&randomPool8[random(POOLSIZE - TEST_STRING_SIZE)] :
					&randomPool7[random(POOLSIZE - TEST_STRING_SIZE)] ;
				for (j = TEST_STRING_SIZE, fcp = fpch->charsOut;  	
					j; j--)
				{
					*fcp++ = *cp++;	
				}
				*fcp = 0;

				fpch->nOut = mapSpeedToSize(fpch->baud);
			}
			fpch->iOut = 0;
			fpch->nIn = 0;
		}


		/* TRANSMIT THE DATA: */
		/* Because of synchronous commands which may have backed up in
		 * the buffers, we may not immediately have room to send the data
		 * However, there should be room within a very short time, say 1/10
		 * second, except for when the loadware is just starting up.  (The
		 * worst case would be when a debugging port is installed).
		 * To be safe, let's try for up to 1 second.  Then something
		 * catastrophic must have occurred.
		 */

		for (retries = 100; retries; retries--)
		{
			for(fpch = fpchTable, i = nfpchTable, anytested = 0;
				i; fpch++, i--)
			{
				if(!(fpch->flags & whichContinue)) continue;
	
				pCh = fpch->c_pCh;

				if (fpch->iOut < fpch->nOut)
				{
					anytested = 1;
	
					/* Transmit the data */

					fpch->iOut += i2Output(pCh,
						(char far*) &(fpch->charsOut[fpch->iOut]),
						fpch->nOut - fpch->iOut);
				}
			}

			if (!anytested) break;	/* If we didn't need to try to send
									 * anything, we must be all sent,
									 * so stop trying.
									 */
			iiDelay(pB,10);
		}

		/* ENSURE ALL DATA SENT (OR AT LEAST BUFFERED): */

		for(fpch = fpchTable, i = nfpchTable, anytested = 0;
			i; fpch++, i--)
		{
			if(!(fpch->flags & whichContinue)) continue;

			anytested = 1;
			if (fpch->iOut < fpch->nOut)
			{
				if(mainVerbosity & V_ERROR)
				{
					if (!(mainVerbosity  & V_PROTOCOL))
					{
						if (!lfyet)
						{
							lfyet = 1;
							printf("\n");
						}
						printProtocol(fpch);
					}
					eprintf("Error transmitting on channel %d: sent %d out of %d characters.\n",
						fpch->thisChan, fpch->iOut, fpch->nOut);
				}

				pf->f_wstring[fpch->thisChan] = pass;
				pf->f_rstring[fpch->thisChan] = L_ERR_NOXMT;
				perfect = 0;
				if(Persist) break;	/* If pass persistance has been
									 * selected, stop this test 
									 * immediately on the first error,
									 * and DON'T turn of future testing.
									 */

				fpch->flags &= ~whichContinue;
			}
		}
		if (!anytested)break;	/* Indicates whichContinue not set anywhere */


		/* READ THE DATA:
		 * Continue a minumum of 3 seconds, as long as some characters 
		 * are coming in somewhere, and 3 seconds thereafter
		 */

		/* At least one channel must be still going, then */

		donetime = time(NULL) + 3;

		for (anyincome = 1; anyincome || (donetime > time(NULL));)
		{
			for(fpch = fpchTable, i = nfpchTable, anyincome = 0, anymissing = 0;
				i; fpch++, i--)
			{
				if(!(fpch->flags & whichContinue)) continue;

				if (fpch->nIn == fpch->nOut) continue;

				/* Toggle the data set signals while waiting for input */
				/* Do this BEFORE we look for any incoming characters,
				 * to give us the benefit of any extra long delays there
				 * may be here.
				 */

				pCh = fpch->c_pCh;

				if (torture)
				{
					if (high)
					{
						high = 0;
						commandToAll(PTYPE_BYPASS, CMD_DTRRTS_UP, 1);
						commandToAll(PTYPE_BYPASS, CMD_DTRDN, 1);
						commandToAll(PTYPE_BYPASS, CMD_DTRUP, 1);
					}
					else
					{
						high = 1;
						commandToAll(PTYPE_BYPASS, CMD_DTRRTS_DN, 1);
						commandToAll(PTYPE_BYPASS, CMD_RTSUP, 1);
						commandToAll(PTYPE_BYPASS, CMD_RTSDN, 1);
					}
				}

				anymissing = 1;	/* Still need input on something */

				j = i2Input(pCh,
					(char far *)&(fpch->charsIn[fpch->nIn]),
					fpch->nOut - fpch->nIn);

				if (j)
				{
					fpch->nIn += j;
					anyincome = 1;
				}

			}
			if (!anymissing) break;

			/* if anything has come in, bump the clock */
			if (anyincome) 
			{
				donetime = time(NULL) + 3;
			}
		}

		/* Now check out the data! */

		for(fpch = fpchTable, i = nfpchTable, anytested = 0;
			i; fpch++, i--)
		{
			if(!(fpch->flags & whichContinue)) continue;

			/* check the counts, data */
			if (fpch->nIn != fpch->nOut)
			{
				if (mainVerbosity & V_ERROR) 
				{
					if (!lfyet)
					{
						lfyet = 1;
						printf("\n");
					}
					showDiscreprencies(fpch, fpch->thisChan, 
						(fpch->nIn < fpch->nOut ? "Data missing/late" :
							"More data received than sent"),
						fpch->nOut, fpch->charsOut,
						fpch->nIn, fpch->charsIn);
				}
				pf->f_wstring[fpch->thisChan] = pass;
				pf->f_rstring[fpch->thisChan] = L_ERR_LATE;

				perfect = 0;
				/* If pass persistance has been selected,
				 * we stop this test after the current pass,
				 * but do not disable further testing on this
				 * port.
				 */
				if(Persist)
				{
					corrupt = 1;
				}
				else
				{
					fpch->flags &= ~whichContinue;
				}
			}
			else if (farCompare(fpch->nIn, fpch->charsOut, fpch->charsIn))
			{
				if (mainVerbosity & V_ERROR) 
				{
					if (!lfyet)
					{
						lfyet = 1;
						printf("\n");
					}
					showDiscreprencies(fpch, fpch->thisChan, 
						"Data incorrect",
						fpch->nOut, fpch->charsOut,
						fpch->nIn, fpch->charsIn);
				}
				pf->f_wstring[fpch->thisChan] = pass;
				pf->f_rstring[fpch->thisChan] = L_ERR_WRONG;
				perfect = 0;
				/* If pass persistance has been selected,
				 * we stop this test after the current pass,
				 * but do not disable further testing on this
				 * port.
				 */
				if(Persist)
				{
					corrupt = 1;

					/* Attempt to flush out any additional incoming
					 * data so that future passes will not be
					 * bogussed.
					 */
					i2InputFlush(fpch->c_pCh);
				}
				else
				{
					fpch->flags &= ~whichContinue;
				}
			}
			else
			{
				anytested = 1;
			}
		}
		if (corrupt || !anytested) break;
	}
	/* We Do this once, after all passes, to get accumulated statuses */
	/* If MM option was used, check for any error-statuses */
	if (statParity)
	{
		for(fpch = fpchTable, i = nfpchTable, anytested = 0;
			i; fpch++, i--)
		{
			j = i2GetStatus(fpch->c_pCh, (I2_OVR | I2_PAR | I2_FRA) );
			if (j & (I2_OVR | I2_PAR | I2_FRA))
			{
				if (!lfyet)
				{
					lfyet = 1;
					printf("\n");
				}
				if (!lf2yet)
				{
					lf2yet = 1;
					printf("\n");
				}
				eprintf("Receiver Error(s) from channel %d: ", fpch->thisChan);
				if (j & I2_OVR) eprintf("OVERRUN ");
				if (j & I2_PAR) eprintf("PARITY ");
				if (j & I2_FRA) eprintf("FRAMING ");
				eprintf("\n");
			}
		}
	}
	/* We've completed all the passes or broken out */
	if (perfect)
	{
		pf->f_error = OKAY;
	}
	else
	{
		pf->f_error = LOOPBACK_ERR;	
		pf->f_passfail = pass - 1;	/* if > f_passes, never totally failed */
	}

	return (pf->f_error == OKAY) ? 0 : 1;
}


/*----------------------------
 * leekTest(pf)
 * pf = argument structure
 * 
 * Cloned from the loopBack()
 * test, but is designed to
 * ensure that NO data has
 * looped back.
 *
 *	f_passfail:   The pass it totally failed on
 *  f_wstring[]: Number of passes successful on each channel.
 *					which failed (otherwise, went the distance)
 *  f_rstring[]: Results per channel:
 *		0 = successful
 *		L_ERR_???	- error condition
 *----------------------------
 */

int 
leekTest(pf)
pfunk pf;
{
fpchControl fpch;
int i, anytested;
int pass;
uchar *cp;
uchar far *fcp;
int j;
int perfect = 1;
int retries;
i2eBordStrPtr pB = pf->f_pB;
char buffer[16];

	/* Initialize all the active channels to 
	 * external loopback mode, possible error reporting.
	 */

	i2CombineCommands(PTYPE_INLINE, (cmdSyntaxPtr) buffer, 4,
		CMD_DTRRTS_DN,
		CMD_SET_ERROR(checkParityOption),
		CMD_INLOOP(CIN_DISABLE), CMD_PARCHK(CPK_ENAB));

	commandToAll(PTYPE_INLINE,(cmdSyntaxPtr) buffer, 0);

	for (cp = pf->f_rstring, i = 64;
		i ; i--)
	{
		*cp++ = 0;
	}
	
	/* We do not need to assert any particular protocols
	 * for these ports: we will test them at whatever
	 * rate they happen to be running.
	 */

	for (pass = 1; pass <= pf->f_passes; pass++)
	{
		/* The test data can be anything at all */

		for(fpch = fpchTable, i = nfpchTable ; i; fpch++, i--)
		{
			cp = fpch->csize ?
				&randomPool8[random(POOLSIZE - TEST_STRING_SIZE)] :
				&randomPool7[random(POOLSIZE - TEST_STRING_SIZE)] ;
			for (j = TEST_STRING_SIZE, fcp = fpch->charsOut;  	
				j; j--)
			{
				*fcp++ = *cp++;	
			}
			*fcp = 0;

			fpch->nOut = mapSpeedToSize(fpch->baud);
			fpch->iOut = 0;
			fpch->nIn = 0;
		}

		/* TRANSMIT THE DATA: */
		/* Because of synchronous commands which may have backed up in
		 * the buffers, we may not immediately have room to send the data
		 * However, there should be room within a very short time, say 1/10
		 * second, except for when the loadware is just starting up.  (The
		 * worst case would be when a debugging port is installed).
		 * To be safe, let's try for up to 1 second.  Then something
		 * catastrophic must have occurred.
		 */

		for (retries = 100; retries; retries--)
		{
			for(fpch = fpchTable, i = nfpchTable, anytested = 0;
				i; fpch++, i--)
			{
				if (fpch->iOut < fpch->nOut)
				{
					anytested = 1;
	
					/* Transmit the data */

					fpch->iOut += i2Output(fpch->c_pCh,
						(char far*) &(fpch->charsOut[fpch->iOut]),
						fpch->nOut - fpch->iOut);
				}
			}

			if (!anytested) break;	/* If we didn't need to try to send
									 * anything, we must be all sent,
									 * so stop trying.
									 */
			iiDelay(pB,10);
		}

		iiDelay(pB, 3000);	/* Wait three whole seconds for data to move */


		/* Now we will attempt, just once, to read some data.
		 * if anything appears, that channel is bad.
		 */

		for(fpch = fpchTable, i = nfpchTable;
			i; fpch++, i--)
		{
			j = i2Input(fpch->c_pCh,
				(char far *)&(fpch->charsIn[fpch->nIn]),
				fpch->nOut - fpch->nIn);

			if (j)
			{
				perfect = 0;
				pf->f_rstring[fpch->thisChan] = L_ERR_GOTDATA;
			}
		}
		if (!perfect) break;
	}
	if (perfect)
	{
		pf->f_error = OKAY;
	}
	else
	{
		pf->f_error = LOOPBACK_ERR;	
		pf->f_passfail = pass ;	/* if > f_passes, never totally failed */
	}

	return (pf->f_error == OKAY) ? 0 : 1;
}


/*-------------------------
 * Pause for operator hit
 *-------------------------
 */

char operatorBuffer[128];

void
operatorHit(essential)
int essential;
{
char *pc;

	if (!essential & !interaction) return;

	printf("---> Hit Enter to continue: ");
	pc = gets(operatorBuffer);

	/* No way this could be false, is there? */
	/* Just an arbitrary delta, this... */
	if (pc == operatorBuffer)
	{
		printf("\n");
	}
}	

/*-------------------------------
 * Optionally pause or go fatal
 * after any error.
 *-------------------------------
 */

void
holdIfNeeded()
{
	if (isFatal) rip(20,"\nStopping test at first error, by request");
	if (isHolding) operatorHit(1);
}


/*--------------------------------------
 * initializeBoard(pB, protest)
 *
 * Attempts to reset and initialize
 * the board and then determine whether it is
 * an expandable or non-expandable.
 * 
 * If 'protest' is true, then printf's
 * if errors are returned from any low-level
 * routines.
 *
 * Returns:
 * -1	Some serious trouble.
 *  0	Non-expandable board
 *  1	Expandable board
 *---------------------------------------
 */

int
initializeBoard(i2eBordStrPtr pB, int protest)
{
int stat1, stat2, stat3;

	/* Wipe any debris from previous invocations */
	pB->i2eGoodMap[0] = pB->i2eGoodMap[1] = 
		pB->i2eGoodMap[2] = pB->i2eGoodMap[3] =
		pB->i2eChannelMap[0] = pB->i2eChannelMap[1] = 
		pB->i2eChannelMap[2] = pB->i2eChannelMap[3] = 0;

	if (	!(stat1 = iiReset(pB)) ||
			!(stat2 = iiResetDelay(pB)) ||
			!(stat3 = iiInitialize(pB)) )
	{
		if (protest)
		{
			eprintf("Error initializing board at address 0x%x\n",
				pB->i2eBase);
			eprintf("Data at 0x%x, Status at 0x%x, Valid=0x%x, State=0x%x\n",
				pB->i2eData, pB->i2eStatus, pB->i2eValid, pB->i2eState);
			eprintf("Reset = %d, Delay = %d, Initialize = %d\n",
				stat1, stat2, stat3);
			boardErrorReport(pB);
			eprintf("\n");
		}
		/* In case we must retry */
		pB->i2eValid = I2E_MAGIC;
		pB->i2eState = II_STATE_COLD;
		return -1;
	}

	switch(pB->i2eFifoStyle)
	{
	case FIFO_II:
		return 0;
	case FIFO_IIEX:
		return 1;
	default:
		printf("Impossible!\n");
		return -1;	/* This case should be impossible: would have
					 * gotten a bad family report earlier.
					 */
	}
}


/*--------------------------------------------------
 * boardMessReport(pB)
 * 
 * Displays the power-on reset message as raw data
 *--------------------------------------------------
 */

void
boardMessReport(i2eBordStrPtr pB)
{
register unsigned char *pc = pB->i2ePom.c;
register unsigned char *pcLimit = &(pB->i2ePom.c[sizeof(porStr)]);

	printf("Power-up reset message (in hex):\n");
	while(pc < pcLimit)
	{
		printf(" %02x", *pc++);
	}
	printf("\n");
}

/*---------------------------------------------
 * interpret_po(f)
 *
 * Prints the interpretation of the power-up reset
 * message, prompts for operator hit before continue
 *-------------------------------------------------
 */

interpret_po(pfunk f)
{
register i2eBordStrPtr pB = f->f_pB;

	printf("\n");
	printf("Power-on Reset Message interpreted: %d bytes read\n",
		pB->i2ePomSize);
	boardMessReport(pB);
	
	printf("     Magic Numbers:  0x%x, 0x%x\n",
		pB->i2ePom.e.porMagic1,
		pB->i2ePom.e.porMagic2);
	printf("     Bootstrap firmware version %d.%d.%d \n", 
		pB->i2ePom.e.porVersion,
		pB->i2ePom.e.porRevision,
		pB->i2ePom.e.porSubRev);

	switch(pB->i2eFifoStyle)
	{
		case FIFO_II:
			printf("     IntelliPort-II");
			break;
		case FIFO_IIEX:
			printf("     IntelliPort-IIEX");
			printf(" (%d-Box)", pB->i2ePom.e.porNumBoxes);
			break;
		default:
			printf("     Unknown type (0x%x)", pB->i2eFifoStyle);
			break;
	}
	printf(" FIFO interface \n");
	printf("          %d-Bits wide, %d bytes deep.\n",
		pB->i2eDataWidth16 ? 16 : 8,
		pB->i2eFifoSize);

	printf("     Debug port %sfound\n",
		(pB->i2ePom.e.porDiag2 & POR_DEBUG_PORT) ? "": "not ");
	
	printf("     Uses %d.%d MHz processor with %dK DRAM\n",
		pB->i2ePom.e.porSpeed / 10,
		pB->i2ePom.e.porSpeed % 10,
		pB->i2ePom.e.porMemory * 32);

	printf("     Channels product is configured to have:\n");
	printChanMap(pB->i2eChannelMap);
	printf("     Channels found during power-on self-test:\n");
	printChanMap(pB->i2eGoodMap);


#ifdef NOT_DEFINED
	printf("i2ePom.e.porID       = %4x \n", pB->i2ePom.e.porID);
	printf("i2ePom.e.porBus      = %4x ", pB->i2ePom.e.porBus);
	printf("i2ePom.e.porDiag1    = %4x \n", pB->i2ePom.e.porDiag1);
	printf("i2ePom.e.porDiag2    = %4x ", pB->i2ePom.e.porDiag2);
	printf("i2ePom.e.porFlags    = %4x ", pB->i2ePom.e.porFlags);
	printf("i2eStartMail      = %4x \n", pB->i2eStartMail);
	printf("i2eMaxIrq         = %4x ", pB->i2eMaxIrq);
	printf("i2eChangeIrq      = %4x \n", pB->i2eChangeIrq);
#endif

	if (pB->i2eStartMail != NO_MAIL_HERE)
	{
		printf("     Initial incoming mail: 0x%x\n", pB->i2eStartMail);
	}
	operatorHit(0);
	return 0;
}


/*---------------------------------------------------
 * sensitiveDelay(ms)
 *
 * This calls the standard delay function, and also
 * calls the kbhit()function to that we will
 * have some ctrl-C sensitivity.
 * Additional changes to work around a Borland delay()
 * bug.
 *---------------------------------------------------
 */

void
sensitiveDelay(int ms)
{
clock_t future;
int times;

#ifdef CLUDGEXX
	ii2DelayIO(ms <<delayMult);
#else
	ms <<= delayMult;
	/* BORLAND ASSUMPTION:
	 * that the clock function returns ticks of about 1000/18.2 = 55ms
	 * per tick.  Therefore for delays of more than about 60 mS we can use
	 * clock to validate.  For shorter ones, we have to punt.
	 */

	if (ms < 64)
	{
		while(ms--)
		{
			delay(1);
		}
		return;
	}

	/* If the bug strikes, then the delay should return immediately,
	 * on essentially the same tick of the clock.
	 * If the delay works, the clock is guaranteed to have changed.
	 */

	times = (ms + 63) / 64;
	while(times--)
	{
		do {
			future = clock();
			delay(64);
		} while (future == clock());
	}
#endif
	kbhit();
}


/*---------------------------------------------------
 * fullDownloadFromFile(pB, fileName,  isStandard)
 * 
 * Loads board at pB with loadware fileName, reports
 * error if any trouble.
 *---------------------------------------------------
 */
void
fullDownloadFromFile(i2eBordStrPtr pB, char *fileName, int isStandard)
{
int status;

	if (mainVerbosity & V_VERSION)
	{
		printf("Writing %sLoadware %s...", 
			(isStandard ? "Standard " : ""), fileName);
	}
	status = iiDownloadFromFile(pB, fileName, isStandard);
	if (status != II_DOWN_GOOD)
	{
		if (!(mainVerbosity & V_VERSION))
		{
			eprintf("Error loading %s: ", fileName);
		}
		switch(status)
		{
		case II_DOWN_BAD:
			rip(11,"CRC error on download");
			break;
		case II_DOWN_BADFILE:
			rip(11,"Bad magic number in loadware file");
			break;
		case II_DOWN_TIMEOUT:
			rip(11,"Timed-out waiting for firmware");
			break;
		case II_DOWN_OVER:
			rip(11,"Loadware file too large");
			break;
		case II_DOWN_UNDER:
			rip(11,"Loadware file too small");
			break;
		case II_DOWN_NOFILE:
			rip(11,"Cannot open file");
			break;

		case II_DOWN_BADSTATE:
		case II_DOWN_BADVALID:
		case II_DOWN_CONTINUING:
		default:
			rip(11,"Software error: impossible");
			break;
		}
	}
	if (mainVerbosity & V_VERSION)
	{
		printf("Version %d.%d.%d\n", 
			pB->i2eLVersion,
			pB->i2eLRevision,
			pB->i2eLSub );
	}
	/* This is in case we want to debug loadware: exit test so we
	 * can go into debug or something like that.
	 * If the 'L' and ',' options are both present, we merely
	 * pause before starting test.
	 */

	if (loadOnly)
	{
		if (isHolding)
		{
			printf("-- Loadware written: ");
			operatorHit(1);
		}
		else
		{
			rip(12,"-- Loadware written to board: aborting test\n");
		}
	}

}

/*---------------------------------------------------
 * boardErrorReport(pB)
 *
 * Uses the pB->i2eError field to determine the
 * cause of the error. 
 * Printf's an appropriate error message.
 *---------------------------------------------------
 */

void
boardErrorReport(i2eBordStrPtr pB)
{
	switch(pB->i2eError)
	{
	case I2EE_GOOD:
		return;

	case I2EE_BADADDR:
		eprintf("Proposed board address is invalid.");
		return;

	case I2EE_BADSTATE:
		eprintf("Inappropriate board state: programming error?");
		return;

	case I2EE_INVALID:
		eprintf("Board pointer (%x) references invalid structure.", pB);
		return;

	case I2EE_BADMAGIC:
		eprintf("Power-on Reset Message did not begin properly: %d characters read\n ",
			pB->i2ePomSize);
		boardMessReport(pB);
		return;

	case I2EE_PORM_SHORT:
		eprintf("Power-on Reset Message too short: %d characters received:\n",
			pB->i2ePomSize);
		boardMessReport(pB);
		return;

	case I2EE_PORM_LONG:
		eprintf("Power-on Reset Message too long: %d characters received:\n",
			pB->i2ePomSize);
		boardMessReport(pB);
		return;

	case I2EE_BADBUS:
	case I2EE_BAD_FAMILY:
		eprintf("Power-on Reset Message reports an unsupported board:\n");
		boardMessReport(pB);
		return;

	case I2EE_INCONSIST:
		eprintf("Power-on Reset Message has conflicting information:\n");
		boardMessReport(pB);
		return;

	case I2EE_POSTERR:
		eprintf("Power-on Reset Message reports P.O.S.T. error\n");
		boardMessReport(pB);
		return;
	
	case I2EE_TXE_TIME:
		eprintf("Timed-out waiting for TX fifo to empty while writing loadware");
		return;

	case I2EE_BAD_PORT:
		eprintf("Power-on Reset Message has a discrepancy between actual\n");
		eprintf("     ports found and what the product should support.\n");
		return;

	case I2EE_BAD_IRQ:
		eprintf("Unsupported IRQ selected");
		return;

	case I2EE_NOCHANNELS:
		eprintf("No channel structure have been initialized yet.");
		return;

	default:
		eprintf("Unknown error reported: %xH ", pB->i2eError);
		return;
	}
}


void
xlatePlugCode(int anyflags)
{
	switch(anyflags & PLUG_BITS)
	{
	case PLUG1:
		printf("TxD(3)->RxD(2) | DTR(8)->DCD(20) | RTS(5)->CTS(4)\n");
		break;
	case PLUG2:
		printf("DTR(8)->DSR(11) | RTS(5)->RI(22)\n");
		break;
	case PLUG1 | PLUG2:
		printf("TxD(3)->RxD(2) | DTR(8)->DCD(20) DSR(11) | RTS(5)->CTS(4) RI(22)\n");
		break;
	case PLUG3:
		printf("TxD(3)->Rxd(2) | DTR(8)->CTS(4) | RTS(5)->DCD(20)\n");
		break;
	default:
		break;
	}
}



void
fatalTrap(i2eBordStrPtr pB)
{
	i2SetFatalTrap(pB, 0);	/* Once this has been called, don't trap.
							 * Reason is that rip() does cleanup that
							 * may involve calls to library routines,
							 * and so possible infinite recursion.
							 */
	if(pB->i2eFatal == 2)
	{
		rip(15,"\nBad block count from board: Unrecoverable!\n");
	}
	else
	{
		rip(15,"\nLoadware has detected a fatal error.  Consult LED\n");
	}
}


/*-------------------------
 * MAIN
 *-------------------------
 */

main(argc, argv)
int argc; 
char **argv;
{
pfunk pf;
ptestdef ptd;
int currentFlags = 0;
int currentPlugs = 0;
int i, j, result;
int errCount = 0;
int passCount = 0;
uchar *cp;
i2eBordStrPtr pBarray[4] = {boardInfo, NULL, NULL, NULL};
i2eBordStrPtr pB = boardInfo;
int firstInit = 1;
int status;
int usesComposite = 0;
int cantBeExpandable = 0;
int cantBeNonExpandable = 0;
time_t currentTime;
int wasPowerless;


	printf("\n ------------- i2diag version %s ------------- \n\n", VERSION);
	startTime = time(NULL);

	/* First, all run-time initializations */
	
	runTimeInitialize();

	/* Then parse arguments */

	parseArguments(argc, argv);


#ifdef NOT_DEFINED
	if (testAddress > 03xff && busType == ISA_OPT)
	{
		usage_bad("I/O Addresses over 0x3ff invalid on the ISA bus");
	}
#endif

	if (hasUsedB && (busType == MCA_OPT) && !validMCA())
	{
		usage_bad("BM argument given, but this is not a MicroChannel host");
	}
		
	if (hasUsedB && (busType == EISA_OPT) && !validEISA())
	{
		usage_bad("BE argument given, but this is not an EISA host");
	}

	if (hasUsedB && (busType == PCI_OPT) && !biosPCIpresent())
	{
		usage_bad("BP argument given, but not PCI bios found");
	}
		
	/* Check equipment requirements */

	allFlags = 0;

	/* OR together all the flags of all the selected tests,
	 * smoke out any tests that only apply to the expandable
	 * or the non-expandable boxes.
	 */

	for (j = 0, ptd = testList, pf = argumentList;
			j < realTests; j++, ptd++, pf++)
	{
		if (ptd->sequence && pf->f_passes)
		{
			allFlags |= ptd->flags;
			if (ptd->pTest[0] == NULL) cantBeNonExpandable = 1;
			if (ptd->pTest[1] == NULL) cantBeExpandable = 1;
		}
	}

	if (allFlags == 0)
	{	
		usage_bad("No tests have been specified.");
	}
	if (cantBeExpandable && cantBeNonExpandable)
	{
		usage_bad("Selected tests for expandable only, and for non-expandable only.");
	}

	/* If you select a test that applies only to expandable (or non-exp),
	 * then it must be consistent with any EN or EX argument we gave.
	 * If no EX or EN argument, behave as though we had one: expect the
	 * type of board that matches the test.
	 */

	switch(isExpandable)
	{
	case NONEXP_TYPE:
		if (cantBeNonExpandable)
		{
			usage_bad("Some specified tests only work on expandable boards.");
		}
		break;

	case EXPAND_TYPE:
		if (cantBeExpandable)
		{
			usage_bad("Some specified tests only work on non-expandable boards.");
		}
		break;

	default:
		if (cantBeNonExpandable) isExpandable = EXPAND_TYPE;	
		if (cantBeExpandable) isExpandable = NONEXP_TYPE;
		break;
	}

	if ((allFlags & NEED_ARG_E) && !hasUsedE)
	{
		usage_bad("Some selected tests require the E argument");
	}
		
	if ((allFlags & (PERINTERRUPT)) &&
		chiefInterrupt == 0 && busType != PCI_OPT)
	{
		usage_bad("Some selected tests require the I argument");
	}

	/* Do not require A option (and implicitly, the I option) if
	 * we are running on MCA or EISA buses.
	 */

	if (busType == MCA_OPT || busType == EISA_OPT || busType == PCI_OPT)
	{
		if (!hasUsedA) testAddress = 0;
	}

	/* Make sure the selected board is in some slot, if
	 * microchannel.
	 */

	if (hasUsedB && (busType == MCA_OPT) && !findMCAslot(&testAddress, &chiefInterrupt))
	{
		rip(13,"No board found at designated address or interrupt");
	}
	
	if (hasUsedB && (busType == EISA_OPT) && !findEISAslot(&testAddress, &chiefInterrupt))
	{
		rip(13,"No board found at designated address or interrupt");
	}
	if (hasUsedB && (busType == PCI_OPT) && !findPCIslot(&testAddress, &chiefInterrupt))
	{
		rip(13,"No board found at designated address or interrupt");
	}
	
	if ((allFlags & (PERINTERRUPT | REAL_LW)) &&
		chiefInterrupt == 0)
	{
		usage_bad("Some selected tests require I argument unless on EISA, MCA, or PCI");
	}


	/* Describe what interrupts, channels, are being tested */

	if(mainVerbosity & V_STARTUP)
	{
		/*----- Interrupts ------- */

		if (allFlags & PERINTERRUPT)
		{
			eprintf("Will Test interrupts: ");
			for (i = mainInterrupts, j = 0; i; i >>= 1, j++)
			{
				if(i & 1) eprintf("%2d, ", validIrqList[j]);
			}	
		}
		else
		{
			if (chiefInterrupt)
			{
				eprintf("Using interrupt %d.", chiefInterrupt);
			}
			else
			{
				eprintf("Using no interrupt.");
			}
		}	
		eprintf("\n");

		/*-------- Channels ----------*/

		if (allFlags & PERCHANNEL)
		{
			eprintf("Channels to be tested:\n");
			if (argumentList[0].f_channels[0] == 0 &&
				argumentList[0].f_channels[1] == 0 &&
				argumentList[0].f_channels[2] == 0 &&
				argumentList[0].f_channels[3] == 0)
			{
				eprintf("None specified -- will test all channels detected as present\n");
			}
			else
			{
				printChanMap(argumentList[0].f_channels);
			}
		}

		/*---------- Tests -----------*/

		eprintf("Tests chosen: ");
		for (i = 1; i <= testSequence; i++)
		{
			for (j = 0, ptd = testList;
				j < realTests; j++, ptd++)
			{
				if (ptd->sequence == i) 
				{
					if (argumentList[j].f_passes)
					{
						eprintf("%d, ", j + 1);
					}
					break;
				}
			}
		}
		eprintf("\n");
		printf("\n");

	}	/* Endif on being verbose */

	/* If any per-channel work is to be done, set up the
	 * tables on a one-time basis!
	 */

	if (allFlags & PERCHANNEL)
	{
		allocChannelTable();
	}

	if (allFlags & DEBUG_NEED)
	{
		printf("\nThe debug port register test was selected.\n");
		printf("Ensure the debug-port card is connected.\n");
		printf("Always TURN OFF POWER when connecting the debug-port card.\n");
		operatorHit(0);
	}

	if ((allFlags & NEEDS_INPUT) && interaction)
	{
		printf("\nThe data transmission test requires you to enter the data to be sent.\n");
		printf("Please Enter this data now:\n");
		cp = (uchar *)gets((char *)bbuffer);
		if (cp != bbuffer)
		{
			rip(13,"Unexpected input");
		}

		/* Ensure against overly long stuff! */
	
		bbuffer[126] = '\n';
		bbuffer[127] = 0;

	}

	switch (allFlags & PLUG_BITS)
	{
	case PLUG1 | PLUG2 | PLUG3:
		usesComposite = (PLUG2 | PLUG1);
	case PLUG1 | PLUG3:
	case PLUG2 | PLUG3:
		plugPrompt = 1;
		currentPlugs = 0;
		printf("\nThe selected tests require different loopback plugs be\n");
		printf("used at different times.  This test will stop and prompt you\n");
		printf("to change plugs as needed.\n");
		break;
		/*---------------------------------*/
	case 0:
		break;

		/*---------------------------------*/
	case PLUG1 | PLUG2:
		usesComposite = (PLUG2 | PLUG1);	/* No break: drop through */
	default:
		printf("\nSome selected tests require the following loopback connector:\n");
		xlatePlugCode(allFlags);
		printf("Ensure they are installed on all selected channels\n");
		operatorHit(0);
		break;
	}


	/* Set the address: shouldn't expect an error */
	if (!iiSetAddress(pB, testAddress, sensitiveDelay) )
	{
		boardErrorReport(pB);
		rip(10,0);
	}
				
	/* Now we may spin through the tests */
	for(passCount = 0; passCount < mainPasses; passCount++)
	{
		/* If we are sufficiently verbose and are doing
		 * multiple passes, tell which one!
		 */

		if ((logFd >= 0 || (mainVerbosity & V_NORMAL) && mainPasses > 1))
		{
			currentTime = time(NULL);
			eprintf("------- Testing: main pass %d of %d... ---------  %s",
				passCount + 1, mainPasses, ctime(&currentTime));
		}

		/* Now, perform each test in the order given on command line */
		/* (Note that testSequence starts at 1) */

		for (i = 1; i <= testSequence; i++)
		{
			for (pf = (pfunk) 0, j = 0, ptd = testList;
				j < realTests; j++, ptd++)
			{
				if (ptd->sequence == i)
				{
					pf  = &(argumentList[j]);
					break;
				}
			}

			/* A little paranoia... */
			if (pf == (pfunk)0) 
			{
				printf("Software error: cannot find sequence %d\n",
					i);
				rip(1,0);
			}

			if (!pf->f_passes) continue;	/* Zero passes means zero! */

			/* ----------------------
			 * ATTENTION:
			 * Should print the channels or interrupts tested here
			 * provided ptd->flags & PERCHANNEL or PERINTERRUPT
			 * is set for this test.  Perhaps so constrain that
			 * we only set once.
			 *------------------------
			 */

			/* It is possible we need to change plugs between two tests */
			/* This happens when a test needs a plug, and we don't have it.
			 */

			if (plugPrompt && 
				((ptd->flags & PLUG_BITS) & ~currentPlugs) )
				
			{
				printf("\nInstall Loopbacks on all selected channels:\n");
				currentPlugs = ptd->flags & PLUG_BITS;

				/* Plugs 1 and 2 may be combined.  If out test
				 * suite needed both, prompt once for the combined
				 * version only.
				 */

				if (currentPlugs & (PLUG1 | PLUG2) )
				{
					currentPlugs |= usesComposite;
				}
				xlatePlugCode(currentPlugs);
				operatorHit(1);
			}
			
			/* Write some loadware if needed */

			if ((ptd->flags & DIAG_LW) || (ptd->flags & DIAG_RESET) ||
				(!(currentFlags & REAL_LW) && ptd->flags & REAL_LW) )

			{
				if(!waitForChannels)
				{
					status = initializeBoard(pB, 1);	/* for now, protest any problem */
				}
				else
				{
					printf("Waiting for %d box(es) to be connected...",
						waitForChannels);
					for (wasPowerless = 0; 
						wasPowerless < 3;)
					{
						printf(".");
						status = initializeBoard(pB, 0);	/* No protest */
						iiDelay (pB,150);
						if((status < 0) || 
							(pB->i2eFifoStyle == FIFO_IIEX &&
					 		pB->i2ePom.e.porNumBoxes != waitForChannels))
						{
							wasPowerless = 0;
						}
						else
						{
							wasPowerless++;
						}
					}
					printf("\n");
				}
				
				/* If this is the first time we have initialized this
				 * board, check to see it is the expected type.
				 */

				if (firstInit)
				{
					if (isExpandable < 0)
					{
						isExpandable = status;
					}
					else if (isExpandable && !status)
					{
						rip(10, "Expandable board expected, non-expandable found");
						boardMessReport(pB);
					}
					else if (!isExpandable && status > 0)
					{
						rip(10, "Non-expandable board expected, expandable found");
						boardMessReport(pB);
					}	
					firstInit = 0;
					if (mainVerbosity & V_VERSION) 
					{
						boardMessReport(pB);
					}
					if (isExpandable && (is16bit >= 0))
					{
						if(is16bit && !(pB->i2eDataWidth16))
						{
							rip(10, "Expandable board using 8-bit interface: command line requested 16");
						}
						else if(!is16bit && pB->i2eDataWidth16)
						{
							rip(10, "Expandable board using 16-bit interface: comment line requested 8");
						}
					}
				}	
				if (status < 0) rip(10,"Error Initializing Board");

				if (ptd->flags & (REAL_LW | DIAG_LW))
				{
					fullDownloadFromFile(pB, ptd->loadName[isExpandable],
							(ptd->flags & REAL_LW));
				}

				/* If this is the standard loadware, initialize library
				 * as well.
				 */

				if (ptd->flags & REAL_LW)
				{
					if (!i2InitChannels(pB, N_CHANNELS, chanInfo))
					{
						boardErrorReport(pB);
						rip(10,0);
					}

					if (!i2LinkEnableIrq(1, pBarray, chiefInterrupt, 
							(void (far *)())i2InterruptHandler))
					{
						boardErrorReport(pB);
						rip(10,0);
					}
					i2SetFatalTrap(pB, fatalTrap);
					endSpeculation();	/* If we haven't already, resolve channel selection */
					validateChannels();
					sendEnables();		/* Enable dss & break reporting */
									/* (Have to do this after validate channels!) */
					iiDelay(pB, 500);	/* Provide some extra time for loadware
										 * start-up code to work.  This allows
										 * the initial flow-control packets to
										 * be sent, as well as the commands 
										 * implied by sendEnables().
										 * When actual tests start, their 
										 * expected timings don't need to 
										 * account for startup.
										 */
				}

			}
			/* If by now we don't know whether we are an expandable or
			 * non-expandable board, we must not care at all, so we assume
			 * we are non-expandable.  This happens, for example, when 
			 * running the POS register test.
			 */
			if (isExpandable < 0) isExpandable = NONEXP_TYPE;

			/* Ensure we can in fact run the test for this kind of board */
			
			if (ptd->pTest[isExpandable] == NULL)
			{
				rip(10,"Wrong type of board for test\n");
			}

			if ( (currentFlags & REAL_LW) && !(ptd->flags & REAL_LW) )
			{
				/* SHOULDN'T NEED TO DO THIS! */
				/* Used to have to unlink the std. library interrupts */
			}

			/* If sufficiently verbose, print name of test & number of
			 * passes to be done
			 */

			if (mainVerbosity & V_NORMAL)
			{
				printf("t%-2d: %s Test, %d %s...", j+1, ptd->testName, 
					pf->f_passes,
					(ptd->flags & TIMED_TEST) ? 
					(pf->f_passes == 1 ? "second" : "seconds" ) :
					(pf->f_passes == 1 ? "pass" : "passes" ) );
			}

			/* Now these flags are current */

			currentFlags = ptd->flags;

			/* Perform the test */
		
			result = (*(ptd->pTest[isExpandable]))(pf);
			if (result) errCount++;

			if (mainVerbosity & V_DEBUG) printf("\nreturn value =%d\n", result);

			/* Display the results */
			(*(ptd->pResults[isExpandable]))(pf, ptd);

			/* If using real loadware and we want to output
			 * interrupt information, give a summary.
			 */

			if ((ptd->flags & REAL_LW) && (mainVerbosity & V_IRQ))
			{
				eprintf("Interrupts: DataOut-%d, DataIn-%d\n",
					pB->i2eFifoInInts, pB->i2eFifoOutInts);
			}

			/* Possibly terminate it if the situation is bad */

			if (result && (ptd->flags & E_FATAL) )
			{
				eprintf("--This error is too serious for further testing to be valid\n");
				eprintf("--Terminating test!\n");

				/* The following is quite unlikely,
				 * that the real firmware is resident when a
				 * test yeilds a fatal condition, but try to handle
				 */
				if ( currentFlags & REAL_LW )
				{
					/* Used to have to unlink library routines here */
				}
				rip(2,0);
			}
		} /* End loop on tests */

		if (waitForChannels)
		{
			/* So we will re-test all channels on the board */
			initChannelTable(CONTINUE_ANY);

			currentFlags = 0;
			i2RestoreIrqs();
			printf("\n-------- %s -----------\n", 
				errCount ? "FAILED! FAILED! FAILED!\007" : "PASSED");
			operatorHit(1);
			errCount = 0;
			continue;
		}
		if (errCount && !Persist) break;	
								/* Unless the X option was given
								 * Do not continue major passes after
								 * an error has occurred */
								/* This will ensure passCount < mainPasses */

	} /* End loop on major passes */

	if ( currentFlags & REAL_LW )
	{
		/* Used to have to unlink library routines here */
	}


	if (errCount)
	{
		eprintf("\n %d out of %d pass(es) completely passed:\n", passCount, mainPasses);
		eprintf(" %d test(s) report errors\n", errCount);
		rip(3,0);
	}			
	else
	{
		rip(0, "\n Test passed successfully\n");
	}
	return 0;	/* But, this won't happen! */
}

