/*---------------------------------
 * i2file.c
 * 
 * IntelliPort-II and IntelliPort-IIEX
 *
 * Service routines for downloading 
 * a loadware module to the board, 
 * from a file.
 *
 * Works through the Extremely Low Level
 * Interface Services (i2ellis.c).
 * These are not included there because
 * they pre-suppose access to the file
 * system:  this can only be used from 
 * application-level code.
 *--------------------------------------
 */

/*--------------------------------------
 * Cosmetic Note:
 * Written using tabstops at 4 spaces.
 * vi: set ts=4
 * pr: -e4
 *--------------------------------------
 */

/*-------------------------------------
 * Revision History:
 *
 * 11 October 1991	MAG		First Draft
 *
 *--------------------------------------
 */

/*-------------------------------------
 * Mandatory Includes:
 *-------------------------------------
 */

#include "i2ellis.h"	/* Defines for low-level board interface */
#include <fcntl.h>		/* open */
#include <io.h>			/* read */


/*-------------------------------------
 * iiDownloadFromFile(pB, filename, isStandard)
 *
 * Routine attempts to open the file,
 * download the entire contents to the
 * board as loadware, then close the file.
 *
 * Returns II_DOWN_GOOD if everything ok,
 * Otherwise returns one of the other II_DOWN_???
 * statuses (see i2ellis.h).  Similar in operation
 * to iiDownloadAll() which tries to download 
 * an entire file from a buffer.
 *
 * This routine expects that the board
 * has been reset and initialized already,
 * and is now actually ready for loadware to
 * be written. (Complains otherwise).
 *---------------------------------------
 */

int
iiDownloadFromFile(i2eBordStrPtr pB, char *filename, int isStandard)
{
unsigned char buffer[LOADWARE_BLOCK_SIZE];
int status;
int fd;

	/* We know (from context) board should be ready for
	 * the first block of download.  Complain if not.
	 */

	if (pB->i2eState != II_STATE_READY) return II_DOWN_BADSTATE;

	/* Attempt to open file, and return bad if not possible */
	if ((fd = open(filename, O_RDONLY | O_BINARY)) < 0) 
	{
		return II_DOWN_NOFILE;
	}
	
	/* File is present and open: remember to close it before you leave */

	/* We'll read as much file as there but expect we should
	 * return from the middle of the loop as soon as the II_DOWN_GOOD
	 * state is acheived.  If read runs out of data first, there
	 * wasn't enough.
	 */

	while (read(fd, buffer, LOADWARE_BLOCK_SIZE) > 0)
	{
		status = iiDownloadBlock(pB, (loadHdrStrPtr)buffer, isStandard);

		switch(status)
		{
		case II_DOWN_CONTINUING:
			break;

		/* Note that unlike iiDownloadAll, we do not check here for
		 * the possibility that the file is actually larger than
		 * declared, since the file system may be more subject to
		 * stray stuff tacked on to the end.  Furthermore, if we
		 * reached this point and the firmware reported a good
		 * CRC, that should be good enough, no?
		 */

		case II_DOWN_GOOD:
			close(fd);
			return II_DOWN_GOOD;

		default:
			close (fd);
			return status;
		}
	}

	/* We shouldn't drop out: it means "while" caught us with
	 * nothing left to download, yet the previous DownloadBlock
	 * did not return complete.   Ergo, not enough data to match
	 * the size byte in the header.
	 */
	close (fd);
	return II_DOWN_UNDER;
}



