/*-------------------------------------------------
 * i2lib.c
 *
 * IntelliPort-II and IntelliPort-IIEX 
 *
 * Standard Loadware support library
 *
 * Uses the Extremely Low Level Interface Support (i2ellis.c)
 * Provides an interface to the standard loadware,
 * to support drivers or application code.
 *-------------------------------------------------
 */

/*-------------------------------------
 * Cosmetic note:
 * Written using tabstops at 4 spaces
 * vi: set ts=4
 * pr: -e4
 *-------------------------------------
 */

/*------------------------------------------------------
 * Compilation Advice:
 * Some code (i.e., in handling of pointers) assumes the
 * SMALL or TINY compilation models, that is, only a
 * single data segment unless otherwise defined.
 * Using Borland,
 * bcc -w -1 -mt -a- -f- -c -B i2lib.c
 *-------------------------------------------------------
 */

/*-----------------------------------------------------------------------------
 * Note on Strategy:
 * Once the board has been initialized, it will interrupt us when
 * 1) It has something in the fifo for us to read (incoming data,
 *	flow control packets, or whatever).
 * 2) It has stripped whatever we have sent last time in the FIFO (and
 *	consequently is ready for more).
 * 3) Every 50 milliseconds even when nothing else is going on.
 *
 * When we want to send packets to the board, we must place them in the
 * fifo.  After we have placed a whole number of packets there, we must
 * then hit the board's mailbox.  Once we have done so, we cannot send
 * more to the fifo until the board has interrupted us.  
 * Each filling of the fifo is like getting a bus-load of passengers.  
 * If we were to leave as soon as the first passenger shows up, we would
 * not be very efficient.  Yet we are not sure that there will ever be
 * more passengers.  So we use the 50-millisecond timer to trigger the
 * ultimate sending of the mailbox interrupt, ultimately triggering output.
 *
 * Note also that the buffer sizes declared in i2lib.h are VERY SMALL.
 * This worsens performance considerably, but is done so that a great
 * many channels might use only a little memory.
 *-----------------------------------------------------------------------------
 */


/*-----------------------------------------------------------------------------
 * Revision History:
 * 
 * 0.00 -- 4/16/91	First Draft
 * 0.01 -- 4/29/91	1st beta release 
 * 0.02 -- 6/14/91	Changes to allow small model compilation
 * 0.03 -- 6/17/91	MAG	--	Break reporting protected from interrupts
 * 							routines with in-line asm added for moving
 * 						data to/from ring buffers, replacing a
 * 							variety of methods used previously.
 * 0.04 -- 6/21/91	MAG	--	Initial flow-control packets not queued until
 * 							i2_enable_interrupts time.  Former versions would 
 * 							enqueue them at i2_init_channel time, before we knew
 * 							how many channels were supposed to exist!
 * 
 * 0.05 -- 10/12/91	MAG	--	Major changes: works through the ellis.c routines
 *							now; supports new 16-bit protocol and expandable
 *							boards.
 *		-- 10/24/91	MAG	--  Most changes in place and stable.
 * 0.06 -- 2/20/92  MAG --  Format of CMD_HOTACK corrected:  the command
 *							takes no argument.
 * 0.07 -- 3/11/92	MAG --	Support added to store special packet types at
 *							interrupt level (mostly responses to specific
 *							commands.)
 * 0.08 -- 3/30/92  MAG --	Support added for STAT_MODEM packet
 * 0.09 -- 6/24/93	MAG	--	i2Link... needed to update number of boards
 *							BEFORE turning on the interrupt.
 * 0.10 -- 6/25/93	MAG	--	To avoid gruesome death from a bad board, we
 *							sanity check some incoming.
 *-----------------------------------------------------------------------------
 */

#define VERSION "0.09"





/*-------------------------------------
 * Mandatory Includes 
 *-------------------------------------
 */

#include "i2lib.h"
#include <dos.h>


/*-------------------------------------
 * Globals
 *-------------------------------------
 */

char i2_version[] = VERSION;



/* The interrupt-level code in this module depends on setting up
 * our own stack so that the stack will be large enough.
 * Here is the stack itself; the helper routines are in i2os.h
 * TO_NEW_STACK() and TO_OLD_STACK().
 */

#define STAKSIZE 500	/* number of words on the interrupt-level stack */

unsigned short i2_superstack[STAKSIZE + 2];


/* These are used by the interrupt handler */
static unsigned short	issueSlaveEoi;	/* EOI slave controller? */
static unsigned short	myMaskBits;		/* Mask bits for this interrupt */
static unsigned short	irqSaveMask;	/* Store old mask value here */
static unsigned short	irqNewMask;		/* Put modified mask here */


/* There is code here for saving, modifying, and restoring the
 * interrupt masks (host 8259) and the interrupt vectors.  This
 * is quite host-architecture dependent.
 */

static unsigned short	savedIrqMasks;	/* Master in LSB, slave in MSB */
static unsigned short	maskSaved = 0;	/* Indicates whether the mask is
										 * saved.  Once it is saved, it
										 * cannot get saved again, since
										 * the presumption is, once we've
										 * saved the masks the first time,
										 * we probably modified them and so
										 * the ones we have now are no longer
										 * virgin.
										 */
static unsigned short	currentIrqMasks;	/* Stores the current IRQ Mask */

/* In addition to the eight vectors supported for original
 * boards, there is also the possibility of PCI, so now
 * we modify to support any interrupt, 0-15.
 */
static unsigned long	savedIrqVectors[16];	/* 16 supported vectors */
static unsigned short 	irqSaved[16] = 
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};	/* Saved yet? */

	/* What is the vector number for each interrupt */
static unsigned short	vectorValues[16] = {
		0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
		0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77 };

						/* For each supported interrupt, which bits in the
						 * mask registers (LSB=master, 0x21) must be
						 * cleared to enable the interrupt.
						 * Note whenever high-order interrupts are used,
						 * we need also to unmask the cascade bit.
						 */
static unsigned short	irqBits[16] = {
	0x1,   0x2,   0x4,   0x8,   0x10,   0x20,   0x40,   0x80,
	0x104, 0x204, 0x404, 0x804, 0x1004, 0x2004, 0x4004, 0x8004};

 
static unsigned long	vectTemp;	/* Temporary holding for IRQ vector */
static unsigned short	valTemp;	/* Temporary holding for IRQ value */

/* -----------------------------------------------------
 * By convention, all the boards will
 * use the same interrupt.  The architecture
 * does not specifically require this, but
 * 1) Interrupts are in short supply anyway,
 * 2) When an interrupt occurs, it doesn't take long
 *	to determine who caused it, so the non-interrupting
 *	boards don't take long to be eliminated.
 * 3) When lots of boards interrupt, it takes less time to
 *	poll through than to service separate interrupts.
 *------------------------------------------------------
 */

/* Here, then is a table of board pointers which the interrupt
 * routine should scan through to determine who it must
 * service
 */

#define ABSOLUTE_MAX_BOARDS 8		/* Extreme overestimate!!!! */
									/* Most real drivers stop at 4 */

i2eBordStrPtr i2BoardPtrTable[ABSOLUTE_MAX_BOARDS];
unsigned short i2nBoards = 0;	/* Number of boards here */


/*----------------------------------------------
 * Allocate sufficient board and channel 
 * structures to support the required number
 * of boards and channels!
 *----------------------------------------------
 */

/* They won't be actually allocated here, however! */


/*------------------------------------------
 * Prototypes  for static functions
 *------------------------------------------
 */



static void i2QueueNeeds(i2eBordStrPtr, i2ChanStrPtr, int);
static void i2StripFifo(i2eBordStrPtr);
static void i2StuffFifoBypass(i2eBordStrPtr);
static void i2StuffFifoFlow(i2eBordStrPtr);
static void i2StuffFifoInline(i2eBordStrPtr);

/* Not a documented part of the library routines (careful...)
 * but the Diagnostic i2diag.c finds them useful to help the
 * throughput in certain limited single-threaded operations.
 */

void iiSendPendingMail(i2eBordStrPtr);
void serviceOutgoingFifo(i2eBordStrPtr);	


/*-----------------------------------------------------------
 * i2RestoreIrqs()
 * 
 * Restores any masks and vectors which might have been
 * saved by i2LinkEnableIrq()
 *
 * Is so contrived that any combination of i2LinkEnableIrq's
 * and i2RestoreIrqs should not be able to permanantly corrupt
 * the vector tables.
 *------------------------------------------------------------
 */

void
i2RestoreIrqs()
{
register int vIndex;
int bTemp;
i2eBordStrPtr pB;

	/* Instruct each board to cease sending ping interrupts */

	for (bTemp = 0; bTemp < i2nBoards; bTemp++)
	{
		pB = i2BoardPtrTable[bTemp];
		i2QueueCommands(PTYPE_INLINE, 
			(i2ChanStrPtr)(pB->i2eChannelPtr), 0, 
			1, CMD_PING_REQ(0));
	}

	/* Wait an arbitrary half second.  This could be
	 * construed to allow any pending interrupts to 
	 * occur, although since we are restoring the
	 * interrupt vectors, we clearly have no continued 
	 * interest in the board, and would be expected to
	 * reset and start over if using it further.
	 * Perhaps this allows some things to occur more gracefully.
	 */

	if (i2nBoards)
	{
		pB = i2BoardPtrTable[0];
		iiDelay(pB, 500);
	}

	/* To ensure the board does not try to issue an interrupt
	 * when no one is ready, clear all the bits in host mask
	 * (which serves for all except Reset)
	 */

	for (bTemp = 0; bTemp < i2nBoards; bTemp++)
	{
		pB = i2BoardPtrTable[bTemp];
		iiDisableMailIrq(pB);
	}

	i2nBoards = 0;	/* Clear table of boards to service */

	/* Restore the original masks */

	if (maskSaved)
	{
		WRITE_IRQ_MASKS(savedIrqMasks);
	}

	/* Restore the original vectors */

	for (vIndex = 0; vIndex < 16; vIndex++)
	{
		if (irqSaved[vIndex])
		{
			valTemp = vectorValues[vIndex];
			vectTemp = savedIrqVectors[vIndex];
			WRITE_IRQ_VECTOR(valTemp, vectTemp);
		}
	}
}

			


/*-------------------------------------------------------
 * i2LinkEnableIrq(boardCnt,ppB, irq, handler)
 *
 * boardCnt 
 *		The number of board pointers passed.
 *		SPECIAL CASE: if boardCnt == 0, then
 *		we add the handler to the vector table only.
 *		This feature is added to support diagnostic code
 *		which might patch several interrupts, although the
 *		board is only using one.
 *		This allows i2RestoreIrqs to be used to restore all the
 *		interrupts which were patched.
 * ppB
 *		The address of an array of board structure pointers. 
 *		(NOT of an array of board structures) which
 *		this interrupt should service.
 * irq
 *		The interrupt to use: (see i2cmd.h)
 * handler
 *		The address of the interrupt handler to use.
 *
 * Handles a maximum of 8 boards (takes the first eight
 * without complaining!  Other limits at higher levels
 * limit us first, actually.)
 *
 * For each board pointer in the list,
 * Validates that the board structure is valid and
 * has the standard loadware loaded
 * 
 * Then determines that theinterrupt
 * being requested is supported.
 *
 * Saves the interrupt mask registers if they have not
 * already been saved. Saves the interrupt vector if
 * it has not already been saved.
 * 
 * Should this routine be called multiple times, the
 * effects of previous calls are negated.
 *
 * Sends command to board to select the interrupt
 * (or polling mode) which it is to use.
 * This command must be issued before any other commands
 * can be sent to the board, after the standard
 * loadware has been installed.
 *
 * If this routine is called when the loadware other than
 * the "standard" has been loaded, it does not send the
 * commands, etc. (It assumes some low-level routine is
 * taking care of everything).
 *--------------------------------------------------------
 */

	
int
i2LinkEnableIrq(unsigned short boardCnt,
		i2eBordStrPtr *ppB, int irq, void(far *handler)())
{ 
int bTemp;
int boardIrq;
i2eBordStrPtr pB = ppB[0];
i2ChanStrPtr pCh;
unsigned char tempCommand[10];
int isStandard = 0;		/* Initialization holds if boardCnt == 0 */


	/* Limit the number of boards */

	if(boardCnt > ABSOLUTE_MAX_BOARDS) 
	{
		boardCnt = ABSOLUTE_MAX_BOARDS;
	}

	/* Check the requested interrupt for validity! */

	/* Now at this level we will allow _any interrupt 1-15
	 * to be chosen.  Most boards only support 3,4,5,7,10,11,12,15,
	 * but the PCI boards might get assigned to any ISA interrupt.
	 * (well, I draw the line at 0: this has traditionally been
	 *  the indicator for polling mode, and this library is
	 * not designed to run in polling mode.)
	 * Furthermore, when we _tell the PCI board what interrupt
	 * we are using, we really should _lie to him, unless we
	 * are using an interrupt that is accepted by all the boards,
	 * otherwise the loadware mightn't like it.
	 */
	if (irq <= 0 || irq > 15)
	{
		COMPLETE(pB, I2EE_BAD_IRQ);
	}

	switch(irq)
	{
	case CIR_3:
	case CIR_4:
	case CIR_5:
	case CIR_7:
	case CIR_10:
	case CIR_11:
	case CIR_12:
	case CIR_15:
		boardIrq = irq;
		break;
		
	default:		/* Yet this one we will allow: it means
					 * that this is probably a PCI board
					 * being used.
					 */
		boardIrq = CIR_15;
		break;
	}

	issueSlaveEoi = (irq > 7);		/* Was it on the slave controller? */

	/* Check each board structure for validity */
	for (bTemp = 0; bTemp < boardCnt; bTemp++)
	{
		pB = ppB[bTemp];
		if (pB->i2eValid != I2E_MAGIC)
		{
			COMPLETE(pB, I2EE_BADMAGIC);
		}

		switch(pB->i2eState)
		{
		case II_STATE_STDLOADED:
			isStandard = 1;
			break;

		case II_STATE_LOADED:
			isStandard = 0;
			break;

		default:
			COMPLETE(pB, I2EE_BADSTATE);
		}

		if (pB->i2eChannelCnt == 0)
		{
			COMPLETE(pB, I2EE_NOCHANNELS);
		}
	}
	
	if (boardCnt > 0)
	{
		/* Early checks ok, so we know we will patch something.
	 	* In case this routine was called earlier, negate its effects! 
	 	* It is always safe to do extra restores. 
	 	*/

		i2nBoards = 0;		/* Prevent any commands being sent to the boards 
						 	* by i2RestoreIrqs().  No need, since we are
						 	* just changing from one set of IRQ's to another
						 	* (NO GUARANTEE actually changing interrupts on
						 	* the fly should be robust).
						 	*/
		i2RestoreIrqs();
	}

	

	/* Save the interrupt request mask, but only if it has not been saved */
	if (!maskSaved)
	{
		maskSaved = 1;
		SAVE_IRQ_MASKS(savedIrqMasks);
	}

	valTemp = vectorValues[irq];

	/* Save the interrupt vector, but not if already saved */
	
	if (!irqSaved[irq])	
	{
		irqSaved[irq] = 1;
		SAVE_IRQ_VECTOR(valTemp,vectTemp);
		savedIrqVectors[irq] = vectTemp;	
	}

	/* Patch in the appropriate new vector */
		
	vectTemp = (unsigned long)handler;
	WRITE_IRQ_VECTOR(valTemp, vectTemp);

	/* Update the appropriate interrupt mask */

	SAVE_IRQ_MASKS(currentIrqMasks);
	myMaskBits = irqBits[irq];
	currentIrqMasks &=  ~myMaskBits;
	WRITE_IRQ_MASKS(currentIrqMasks);

	
	if (isStandard)
	{
		/* Notify the boards they may generate interrupts 
	 	* This is done by sending an in-line command to
	 	* channel 0 on each board.  This is why the channels
	 	* have to be defined already.
	 	*/
	
		/* For each board, if the interrupt has never
	 	* been defined, we must do so NOW, directly,
	 	* since board will not send flow control or
	 	* even give an interrupt until this is done.
	 	*/
	
		for (bTemp = 0; bTemp < boardCnt; bTemp++)
		{
			pB = ppB[bTemp];
			pCh = (i2ChanStrPtr) pB->i2eChannelPtr;
	
			i2BoardPtrTable[bTemp] = pB;	/* Store pointer in 
											 * global table so interrupt 
											 * handler can reach it 
											 */
	
			if (pB->i2eUsingIrq == IRQ_UNDEFINED)
			{
				iiDisableMailIrq(pB);
				/* We build up the entire packet header */
	
				CHANNEL_OF(tempCommand) = 0;
				PTYPE_OF(tempCommand) = PTYPE_INLINE;
				CMD_COUNT_OF(tempCommand) = 2;
				(CMD_OF(tempCommand))[0] = CMDVALUE_IRQ;
				(CMD_OF(tempCommand))[1] = boardIrq;
	
				/* Write to FIFO: don't bother to adjust fifo
			 	* capacity for this, since board will respond
			 	* almose immediately.
			 	*/
	
				iiWriteBuf(pB, tempCommand, 4);
				pB->i2eUsingIrq = irq;
				pB->i2eOutMailWaiting |= MB_OUT_STUFFED;
	
				/* This command can be sent in the normal way.
			 	* The board will soon respond with starting
			 	* flow-control packets.  As soon as they come in
			 	* this will trigger the sending of any queued-up
			 	* commands.
			 	* Because the mailbox interrupt has been disabled, 
			 	* there's no way for these packets to come in until
			 	* this command has been queued.
				*/

				i2QueueCommands(PTYPE_INLINE, pCh, 0,
					1, CMD_PING_REQ(50));
			}
			else
			{
				i2QueueCommands(PTYPE_INLINE, pCh, 0,
					2, 
					CMD_SET_IRQ(boardIrq),
					CMD_PING_REQ(50));
			}
			/* Need to update number of boards before you enable mail int */
			i2nBoards = (unsigned short)bTemp + 1;

			iiEnableMailIrq(pB);
			iiSendPendingMail(pB);
		}
	}

	COMPLETE(ppB[0], I2EE_GOOD);
}


/*---------------------------------------
 * iiSendPendingMail(pB)
 *
 * If any outgoing mail bits are set and
 * there is outgoing mailbox is empty,
 * send the mail and clear the bits.
 *---------------------------------------
 */

void
iiSendPendingMail(register i2eBordStrPtr pB)
{
	SAVE_AND_DISABLE_INTS
	if (pB->i2eOutMailWaiting)
	{
		if (iiTrySendMail(pB, pB->i2eOutMailWaiting)) 
		{
			/* If we were already waiting for fifo to empty,
			 * or just sent MB_OUT_STUFFED, then we are
			 * still waiting for it to empty, until we should
			 * receive an MB_IN_STRIPPED from the board.
			 */
			pB->i2eWaitingForEmptyFifo |= 
				(pB->i2eOutMailWaiting & MB_OUT_STUFFED);
			pB->i2eOutMailWaiting = 0;
		}
	}
	RESTORE_INTS
}



/*------------------------------------------------------
 * i2InitChannels(pB, nChannels, pCh)
 *
 * pB 	 		pointer to a board structure
 * nChannels 	number of channels to initialize
 * pCh			pointer to an array of channel structures
 *
 * Routine patches pointer, back-pointers, and 
 * initializes all the elements in the channel structure
 * array.
 *
 * This should be run after the board structure is
 * initialized, through having loaded the standard loadware
 * (Otherwise it complains).
 *
 * In any case, it must be done before any serious work
 * begins initializing the irq's or sending commands...
 *-------------------------------------------------------
 */

int
i2InitChannels(register i2eBordStrPtr pB, 
		int nChannels, register i2ChanStrPtr pCh)
{
int index, stuffIndex;
i2ChanStrPtr *ppCh;

	if (pB->i2eValid != I2E_MAGIC)
	{
		COMPLETE(pB, I2EE_BADMAGIC);
	}

	if (pB->i2eState != II_STATE_STDLOADED)
	{
		COMPLETE(pB, I2EE_BADSTATE);
	}

	pB->i2eChannelPtr = pCh;
	pB->i2eChannelCnt = nChannels;

	pB->i2fbufStrip = pB->i2fbufStuff = 0;
	pB->i2dbufStrip = pB->i2dbufStuff = 0;
	pB->i2bbufStrip = pB->i2bbufStuff = 0;

	/* Sanity on number of channels, for now */
	if (nChannels > ABS_MOST_PORTS) nChannels = ABS_MOST_PORTS;

	for (index = stuffIndex = 0, ppCh = (i2ChanStrPtr *)(pB->i2fBuf); 
			index < nChannels; 
				index++, pCh++)
	{
		/* Set up validity flag according to support level */

		if (pB->i2eGoodMap[index >> 4] & (1 << (index & 0xf)) )
		{
			pCh->validity = CHANNEL_MAGIC | CHANNEL_SUPPORT;
		}
		else
		{
			pCh->validity = CHANNEL_MAGIC;
		}

		pCh->pMyBord = pB;		/* Back-pointer */

		/* Prepare an outgoing flow-control packet to send
		 * as soon as the chance occurred.
		 */
		if(pCh->validity & CHANNEL_SUPPORT)
		{	
			pCh->infl.hd.i2sChannel = index;
			pCh->infl.hd.i2sCount = 5;
			pCh->infl.hd.i2sType = PTYPE_BYPASS;
			pCh->infl.fcmd = 37;
			pCh->infl.asof = 0;
			pCh->infl.room = IBUF_SIZE - 1;

			pCh->whenSendFlow = (IBUF_SIZE * 4)/5;	/* compiler does this! */

				/* The following is similar to calling i2QueueNeeds,
				 * except that this is done in longhand, since we
				 * are setting up initial conditions on many channels
				 * at once.
				 */
			pCh->channelNeeds = NEED_FLOW;	/* Since starting from scratch */
			pCh->sinceLastFlow = 0;			/* No bytes received since last
											 * flow control packet was queued
											 */
		
			stuffIndex++;
			*ppCh++ = pCh;			/* List this channel as needing initial
									 * flow control packet sent
									 */
		}

		/* Don't allow anything to be sent until the status
		 * packets come in from the board.
		 */

		pCh->outfl.asof = 0;
		pCh->outfl.room = 0;

		/* Initialize all the ring buffers */

		pCh->ibufStuff = pCh->ibufStrip = 0;
		pCh->obufStuff = pCh->obufStrip = 0;
		pCh->cbufStuff = pCh->cbufStrip = 0;

		pCh->dataSetIn = 0;
		pCh->hotKeyIn = HOT_CLEAR;
		pCh->channelOptions = 0;
		pCh->bookMarks = 0;
	}
	/* No need to check for wrap here: this is initialization */
	pB->i2fbufStuff = stuffIndex;
	COMPLETE(pB, I2EE_GOOD);
}


/*---------------------------------------------------
 * i2QueueNeeds(pB, pCh, type)
 *
 * pB	-	Pointer to a board structure
 * pCh	-	Pointer to a channel structure
 * type	-	Bit mapped: may include NEED_INLINE, 
 *			NEED_BYPASS, or NEED_FLOW
 *
 * For each type of need selected, if the given
 * channel is not already in the queue, adds it,
 * and sets the flag indicating it is in the queue.
 *---------------------------------------------------
 */


static void
i2QueueNeeds(i2eBordStrPtr pB, i2ChanStrPtr pCh, int type)
{
unsigned short queueIndex;

	/* We turn off all the interrupts during this brief process,
	 * since the interrupt-level code might want to put things
	 * on the queue as well.  
	 */

	SAVE_AND_DISABLE_INTS

	if ((type & NEED_INLINE) && 
		!(pCh->channelNeeds & NEED_INLINE))
	{
		pCh->channelNeeds |= NEED_INLINE;
		queueIndex = pB->i2dbufStuff;
		pB->i2dBuf[queueIndex++] = pCh;
		if (queueIndex >= CH_QUEUE_SIZE)
		{
			queueIndex = 0;
		}
		pB->i2dbufStuff = queueIndex;
	}
			
	if ((type & NEED_BYPASS) && 
		!(pCh->channelNeeds & NEED_BYPASS))
	{
		pCh->channelNeeds |= NEED_BYPASS;
		queueIndex = pB->i2bbufStuff;
		pB->i2bBuf[queueIndex++] = pCh;
		if (queueIndex >= CH_QUEUE_SIZE)
		{
			queueIndex = 0;
		}
		pB->i2bbufStuff = queueIndex;
	}
			
	if ((type & NEED_FLOW) && 
		!(pCh->channelNeeds & NEED_FLOW))
	{
		pCh->channelNeeds |= NEED_FLOW;
		queueIndex = pB->i2fbufStuff;
		pB->i2fBuf[queueIndex++] = pCh;
		if (queueIndex >= CH_QUEUE_SIZE)
		{
			queueIndex = 0;
		}
		pB->i2fbufStuff = queueIndex;
	}

	RESTORE_INTS
}
			


/*------------------------------------------------------------
 * int
 * i2QueueCommands(type, pCh, timeout, nCommands, pCs,...)
 *
 * Queues up some commands to be sent to a channel.
 * To send possibly several bypess or inline commands
 * to the given channel.
 * 
 * type
 *		PTYPE_BYPASS or PTYPE_INLINE
 * pCh
 *		pointer to the channel structure
 * timeout
 *		indicates how many HUNDREDTHS OF SECONDS to wait until there
 *		is room. 
 *		0 = return immediately if no room
 *		negative  = wait forever
 *		positive  = number of 1/100 seconds to wait
 * nCommands
 *		How many commands are being passed?
 * pCs
 *		one or more commands: use the macros in I2CMD.H please
 *
 * Return values:
 * -1	Some kind of nasty error: bad channel structure or 
 *		invalid arguments.
 *  0	No room to send all the commands
 * (+)	Number of commands sent
 *
 * Does not worry here about channel flow control: this is
 * taken care of when we send blocks to the fifo.
 *------------------------------------------------------------
 */

int
i2QueueCommands(int type, i2ChanStrPtr pCh, int timeout,
		int nCommands, cmdSyntaxPtr pCs0,...)
{
int totalsize;
int blocksize;
int lastended;
cmdSyntaxPtr *ppCs;
cmdSyntaxPtr pCs;
int count;
int flag;
i2eBordStrPtr pB;

unsigned short maxBlock;
unsigned short maxBuff;
short bufroom;
unsigned short stuffIndex;
unsigned char *pBuf;
unsigned char *pInsert;
unsigned char *pDest, *pSource;
unsigned short channel;
int cnt;

	/* Make sure the channel exists, otherwise do nothing */

	if ((pCh->validity & (CHANNEL_MAGIC_BITS | CHANNEL_SUPPORT)) !=
		(CHANNEL_MAGIC | CHANNEL_SUPPORT))
	{
		return -1;
	}
	pB = pCh->pMyBord;

	/* Board must also exist, and THE INTERRUPT COMMAND ALREADY SENT */

	if (pB->i2eValid != I2E_MAGIC || 
		pB->i2eUsingIrq == IRQ_UNDEFINED)
	{
		return -1;	
	}

	/* If the board has gone fatal, return bad, and also
	 * hit the trap routine if it exists
	 */

	if (pB->i2eFatal)
	{
		if (pB->i2eFatalTrap != NULLFUNC)
		{
			(*(pB)->i2eFatalTrap)(pB);
		}
		return -1;
	}


	/* Which buffers are we using?  How big are they? */
	switch(type)
	{
	case PTYPE_INLINE:
		flag = INL;
		maxBlock = MAX_OBUF_BLOCK;
		maxBuff = OBUF_SIZE;
		pBuf = pCh->obuf;
		break;
	case PTYPE_BYPASS:
		flag = BYP;
		maxBlock = MAX_CBUF_BLOCK;
		maxBuff = CBUF_SIZE;
		pBuf = pCh->cbuf;
		break;
	default:
		return -1;
	}

	while(1) {
		/* How much room (this time through) ? */
		switch(type)
		{
		case PTYPE_INLINE:
			stuffIndex = pCh->obufStuff;
			bufroom = pCh->obufStrip - stuffIndex;
			break;
		case PTYPE_BYPASS:
			stuffIndex = pCh->cbufStuff;
			bufroom = pCh->cbufStrip - stuffIndex;
			break;
		default:
			return -1;
		}
		if (--bufroom < 0)
		{
			bufroom += maxBuff; 
		}

		/* Determine the total size required for all the
	 	* commands (n.b., these pack together).  Also, check
	 	* that they are all of the same type.
	 	*/

		for (totalsize = blocksize = sizeof(i2CmdHeader), 
				count = nCommands, 
				lastended = 0, ppCs = &pCs0;
			count;
			count--, ppCs++)
		{
			pCs = *ppCs;
	
			/* Make sure command can be sent in the appropriate mode */
			if (!(pCs->flags & flag))
			{
				return -1;
			}
	
			cnt = pCs->length;
	
			/* Will a new block be needed for this one ?
			 * Two possible reasons: too big or previous command
			 * has to be at the end of a packet.
			 */
	
			if ((blocksize + cnt > maxBlock) || lastended)
			{
				blocksize = sizeof(i2CmdHeader);
				totalsize += sizeof(i2CmdHeader);
			}	
	
			totalsize += cnt;
			blocksize += cnt;
	
			/* Check for overflow */
			if (totalsize > bufroom) break;

			/* If this command had to end a block, then
			 * we will make sure to account for it should
			 * there be any more blocks.
			 */
	
			lastended = pCs->flags & END;
		}
		if (totalsize <= bufroom) break;	/* Enough room: proceed */

		serviceOutgoingFifo(pB);	/* Try to throw more things
									 * (maybe not us) in the fifo
									 * if we're not already waiting
									 * for it
									 */

		if (timeout == 0) return 0;		/* Tired of waiting */
		if (timeout > 0) timeout--;		/* So negative values = forever */
		iiDelay(pB, 10);				/* Count 1/100 sec = 10 mS */
										/* Of course this is a little long,
										 * because of this routine's execution
										 * time gets added on!
										 */
	}


	/* We looked ahead to see that all the commands could fit in
	 * the buffer.  Now we will start back and put them in.
	 */

	channel = pCh->infl.hd.i2sChannel;

	pInsert = &pBuf[stuffIndex];	/* Pointer to start of packet */
	pDest = CMD_OF(pInsert);		/* Pointer to start of command */

	/* When we start counting, the block is the size of the header */
	for (blocksize = sizeof(i2CmdHeader), count = nCommands, 
			lastended = 0, ppCs = &pCs0;
		count;
		count--, ppCs++)
	{
		pCs = *ppCs;			/* Points to command protocol structure */


		/* If this is a bookmark request command, post the fact that
		 * a bookmark request is pending.
		 *
		 * NOTE THIS TRICK ONLY WORKS BECAUSE CMD_BMARK_REQ has no
		 * parameters!  The more general solution would be to
		 * reference pCs->cmd[0].
		 */

		if (pCs == CMD_BMARK_REQ)
		{
			pCh->bookMarks++;
		}

		cnt = pCs->length;

		/* If this command would put us over the maximum block size
		 * or if the last command had to be at the end of a block,
		 * we end the existing block here and start a new one.
		 */

		if ((blocksize + cnt > maxBlock) || lastended)
		{
			PTYPE_OF(pInsert) = type;
			CHANNEL_OF(pInsert) = channel;			
			CMD_COUNT_OF(pInsert) = blocksize - sizeof(i2CmdHeader);	
											/* count here does
											 * not include the header 
											 */
			stuffIndex += blocksize;
			if(stuffIndex >= maxBuff)
			{
				stuffIndex = 0;
				pInsert = pBuf;
			}
			pInsert = &pBuf[stuffIndex];	/* Pointer to start of next pkt */
			pDest = CMD_OF(pInsert);
			blocksize = sizeof(i2CmdHeader);
		}

		/* Now we know there is room for this one in the current block */

		blocksize += cnt;		/* Total bytes in this command */
		pSource = pCs->cmd;		/* Copy the command into the buffer */
		while (cnt--)
		{
			*pDest++ = *pSource++;
		}
		
		/* If this command had to end a block, then
		 * we will make sure to account for it should
		 * there be any more blocks.
		 */

		lastended = pCs->flags & END;

	}
	/* Clean up the final block by writing header, etc */

	PTYPE_OF(pInsert) = type;
	CHANNEL_OF(pInsert) = channel;			
	CMD_COUNT_OF(pInsert) = blocksize - sizeof(i2CmdHeader);
											/* count here does
											 * not include the header */
	stuffIndex += blocksize;
	if(stuffIndex >= maxBuff)
	{
		stuffIndex = 0;
		pInsert = pBuf;
	}

	/* Updates the index, and post the need for service */ 
	/* When adding these to the queue of channels, we turn off
	 * the interrupt while doing so, because at interrupt level
	 * we might want to push a channel back to the end of the
	 * queue
	 */

	switch(type)
	{
	case PTYPE_INLINE:
		pCh->obufStuff = stuffIndex;	/* Store buffer pointer */

		/* Add the channel pointer to list of channels needing
		 * service (first come...), if it's not already there
		 */

		i2QueueNeeds(pB, pCh, NEED_INLINE);
		break;

	case PTYPE_BYPASS:
		pCh->cbufStuff = stuffIndex;	/* Store buffer pointer */

		/* Add the channel pointer to list of channels needing
		 * service (first come...), if it's not already there
		 */

		i2QueueNeeds(pB, pCh, NEED_BYPASS);
		break;

	default:
		return -1;	/* impossible, really */
	}
	return nCommands;	/* Good status: number of commands sent */
}



/*------------------------------------------------------------
 * int
 * i2CombineCommands(type, pDest, nCommands, pCs,...)
 *
 * Combines several commands (pCs arguments) into a
 * single command and stores it at pDest.
 *
 * The user is responsible to see that there is enough room 
 * available at pDest to store the combined command.
 *
 * Commands are combined as long as the argument list holds out,
 * the maximum block size is not exceeded, or until a command
 * is encountered which must go at the end of a packet.
 *
 * Commands designated as VIP cannot be combined
 *
 * Using the command at pDest, i2QueueCommands can then be called
 * for a number of channels using a fixed number of arguments.
 * 
 * 
 * type
 *		PTYPE_BYPASS or PTYPE_INLINE
 * pDest
 *		pointer to a cmdSyntax structure which will receive the
 *		resulting combined command.
 * nCommands
 *		How many commands are being passed?
 * pCs
 *		one or more commands: use the macros in I2CMD.H please
 *
 * Return values:
 * -1	Some kind of nasty error: inconsistent arguments?
 *  0	..or..
 * (+)	Number of commands combined.
 *
 *------------------------------------------------------------
 */

int
i2CombineCommands(int type, cmdSyntaxPtr pDest, int nCommands, 
		cmdSyntaxPtr pCs0,...)
{
unsigned short maxBlock;
cmdSyntaxPtr *ppCs;
cmdSyntaxPtr pCs;
int count;
int flag;
int cnt;
int newLength;

unsigned char *pInsert;
unsigned char *pSource;

	/* Make sure the channel exists, otherwise do nothing */

	/* Which buffers are we using?  How big are they? How much room left? */
	switch(type)
	{
	case PTYPE_INLINE:
		flag = INL;
		maxBlock = MAX_OBUF_BLOCK - sizeof(i2CmdHeader);
		break;
	case PTYPE_BYPASS:
		flag = BYP;
		maxBlock = MAX_CBUF_BLOCK - sizeof(i2CmdHeader);
		break;
	default:
		return -1;
	}

	/* Ensure that all commands could be sent in the way
	 * required. (Even if they won't get combined after all).
	 */

	for (count = nCommands, ppCs = &pCs0;
		count;
		count--, ppCs++)
	{
		pCs = *ppCs;

		/* Make sure command can be sent in the appropriate mode */
		if (!(pCs->flags & flag) || (pCs->flags & VIP) )
		{
			return -1;
		}
	}


	pInsert = pDest->cmd;


	/* When we start counting, the block is the size of the header */

	for (newLength = 0, pInsert = pDest->cmd, 
			count = nCommands, ppCs = &pCs0;
		count;)
	{
		pCs = *ppCs++;			/* Points to command protocol structure */

		cnt = pCs->length;

		/* Stop before you combine a command that would
		 * put us over the maximum block length.
		 */

		if (newLength + cnt > maxBlock)
		{
			break;
		}

		newLength += cnt;		/* Total bytes in this command */
		pSource = pCs->cmd;		/* Copy the command into the buffer */
		while (cnt--)
		{
			*pInsert++ = *pSource++;
		}
		count--;
		
		/* If we just added a command which must end a block,
		 * stop combining here.
		 */

		if (pCs->flags & END)
		{
			break;
		}
	}
	pDest->length = newLength;
	pDest->flags = flag;
	return nCommands - count;
	
}



/*------------------------------------------------------------
 * i2GetStatus(pCh,resetBits)
 *
 * Returns the state of data set signals, and whether a break
 * has been received. (see i2lib.h for bit-mapped result).
 * resetBits is a bit-map of any status bits to be cleared:
 * I2_BRK, I2_PAR, I2_FRA, I2_OVR,...
 * These are cleared AFTER the condition is passed.
 * If pCh does not point to a valid channel, returns -1
 * (which would be impossible otherwise.
 *------------------------------------------------------------
 */

int
i2GetStatus(i2ChanStrPtr pCh, int resetBits)
{
unsigned short status;

	/* Make sure the channel exists, otherwise do nothing */

	if ((pCh->validity & (CHANNEL_MAGIC_BITS | CHANNEL_SUPPORT)) !=
		(CHANNEL_MAGIC | CHANNEL_SUPPORT))
	{
		return -1;
	}

	SAVE_AND_DISABLE_INTS
	status = pCh->dataSetIn;

	/* Clear any specified error bits: but note that only actual
	 * error bits can be cleared, regardless of the value passed.
	 */
	if(resetBits)
	{
		pCh->dataSetIn &= ~(resetBits & (I2_BRK | I2_PAR | I2_FRA | I2_OVR));
	}
	RESTORE_INTS
	return status;
}


/*----------------------------------------------
 * i2Input(pCh, fpDest, count)
 *
 * pCh
 *		Pointer to a channel structure
 * fpDest
 *		(Far) Pointer to data buffer
 * count
 *		number of bytes to read
 *
 * Strips data from the input buffer and writes it 
 * to fpDest.
 *
 * If there is a collosal blunder, (invalid structure
 * pointers or the like), returns -1.
 *
 * Otherwise, returns the number of bytes read.
 *
 *-------------------------------------------------
 */

int
i2Input(i2ChanStrPtr pCh, char far *fpDest, int countRequested)
{
int amountToMove;
unsigned short stripIndex;
int count;

	/* Ensure channel structure seems real */

	if ((pCh->validity & (CHANNEL_MAGIC_BITS | CHANNEL_SUPPORT)) !=
		(CHANNEL_MAGIC | CHANNEL_SUPPORT))
	{
		return -1;
	}

	/* initialize some accelerators and private copies */

	stripIndex = pCh->ibufStrip;

	count = pCh->ibufStuff - stripIndex;

	/* If buffer is empty or requested data count was 0,
	 * (trivial case) return without any further thought
	 */ 
	if (count == 0 || countRequested == 0)
	{
		return 0;
	}

	/* Adjust for buffer wrap */
	if (count < 0)
	{
		count += IBUF_SIZE;
	}
	
	/* Don't give more than was asked */
	if (count > countRequested) 
	{
		count = countRequested;
	}

	/* How much could we copy without a wrap? */
	amountToMove = IBUF_SIZE - stripIndex;
	
	if (amountToMove > count)
	{
		amountToMove = count;
	}

	/* Move the first block */
	i2moveData((char far *)(&(pCh->ibuf[stripIndex])),
		fpDest, amountToMove);

	/* If we needed to wrap, do the second data move */
	if (count > amountToMove)
	{
		fpDest += amountToMove;
		i2moveData((char far *)(pCh->ibuf), fpDest, count - amountToMove);
	}
	/* Bump and wrap the stripIndex all at once by the amount
	 * of data read. This method is good regardless of whether
	 * the data was in one or two pieces.
	 */

	stripIndex += count;
	if (stripIndex >= IBUF_SIZE)
	{
		stripIndex -= IBUF_SIZE;
	}	
	pCh->ibufStrip = stripIndex;

	/* Update our flow control information and possibly queue 
	 * ourselves to send it, depending on how much data has
	 * been stripped since the last time a packet was sent.
	 */

	pCh->infl.asof += count;

	if ((pCh->sinceLastFlow += count) >= pCh->whenSendFlow)
	{
		pCh->sinceLastFlow -= pCh->whenSendFlow;
		i2QueueNeeds(pCh->pMyBord, pCh, NEED_FLOW);
	}
	return count;
}

/*----------------------------------------------
 * i2InputLookahead(pCh)
 *
 * pCh
 *		Pointer to a channel structure
 *
 * Peek ahead at data without stripping.
 *
 * If there is a collosal blunder, (invalid structure
 * pointers or the like), returns -1.
 *
 * If there are no characters available, returns 0
 * If there are characters, returns the next character,
 * OR'ed with 0x100 (to create a non-zero return).
 *
 *-------------------------------------------------
 */

int
i2InputLookahead(i2ChanStrPtr pCh)
{

	/* Ensure channel structure seems real */

	if ((pCh->validity & (CHANNEL_MAGIC_BITS | CHANNEL_SUPPORT)) !=
		(CHANNEL_MAGIC | CHANNEL_SUPPORT))
	{
		return -1;
	}


	if (pCh->ibufStuff != pCh->ibufStrip)
	{
		return ((pCh->ibuf[pCh->ibufStrip]) | 0x100);
	}
	return 0;
}



/*----------------------------------------------
 * i2InputFlush(pCh)
 *
 * pCh
 *		Pointer to a channel structure
 *
 * Strips any data from the input buffer 
 *
 * If there is a collosal blunder, (invalid structure
 * pointers or the like), returns -1.
 *
 * Otherwise, returns the number of bytes stripped.
 *
 *-------------------------------------------------
 */

int
i2InputFlush(i2ChanStrPtr pCh)
{
int count;

	/* Ensure channel structure seems real */

	if ((pCh->validity & (CHANNEL_MAGIC_BITS | CHANNEL_SUPPORT)) !=
		(CHANNEL_MAGIC | CHANNEL_SUPPORT))
	{
		return -1;
	}

	/* initialize some accelerators and private copies */

	SAVE_AND_DISABLE_INTS

	count = pCh->ibufStuff - pCh->ibufStrip;

	/* Adjust for buffer wrap */
	if (count < 0)
	{
		count += IBUF_SIZE;
	}
	
	/* Expedient way to zero out the buffer */
	pCh->ibufStrip = pCh->ibufStuff;

	RESTORE_INTS

	/* Update our flow control information and possibly queue 
	 * ourselves to send it, depending on how much data has
	 * been stripped since the last time a packet was sent.
	 */

	pCh->infl.asof += count;

	if ((pCh->sinceLastFlow += count) >= pCh->whenSendFlow)
	{
		pCh->sinceLastFlow -= pCh->whenSendFlow;
		i2QueueNeeds(pCh->pMyBord, pCh, NEED_FLOW);
	}
	return count;
}



/*----------------------------------------------
 * i2InputAvailable(pCh)
 *
 * pCh
 *		Pointer to a channel structure
 *
 *
 * If there is a collosal blunder, (invalid structure
 * pointers or the like), returns -1.
 *
 * Otherwise, returns the number of bytes available
 * in the buffer.
 *
 *-------------------------------------------------
 */

int
i2InputAvailable(i2ChanStrPtr pCh)
{
int count;

	/* Ensure channel structure seems real */

	if ((pCh->validity & (CHANNEL_MAGIC_BITS | CHANNEL_SUPPORT)) !=
		(CHANNEL_MAGIC | CHANNEL_SUPPORT))
	{
		return -1;
	}

	/* initialize some accelerators and private copies */

	count = pCh->ibufStuff - pCh->ibufStrip;

	/* Adjust for buffer wrap */
	if (count < 0)
	{
		count += IBUF_SIZE;
	}
	
	return count;
}



/*------------------------------------------------
 * i2moveData(fpSource, fpDest, count)
 *
 * Moves count bytes from (char far *) fpSource to
 * fpDest
 *
 * No return value.
 *------------------------------------------------
 */

void
i2moveData(char far *fpSource, char far *fpDest, unsigned short count)
{
	I2_MOVE_DATA(fpSource, fpDest, count)
}


/*----------------------------------------------
 * i2Output(pCh, fpSource, count)
 *
 * pCh
 *		Pointer to a channel structure
 * fpSource
 *		(Far) Pointer to data to send
 * count
 *		number of bytes to send
 *
 * Queues the data at fpSource to be sent as data
 * packets to the board.
 *
 * If there is a collosal blunder, (invalid structure
 * pointers or the like), returns -1.
 *
 * Otherwise, returns the number of bytes written.
 *
 * What if there is not enough room for all the data?
 * If pCh->channelOptions & CO_NBLOCK_WRITE is set, then
 * we transfer as many characters as we can now, then return
 *
 * If this bit is clear (default), routine will spin along
 * until all the data is buffered.  Should this occur, the
 * 1-ms delay routine is called while waiting to avoid
 * applications that one cannot break out of.
 *
 *-------------------------------------------------
 */

int
i2Output(i2ChanStrPtr pCh, char far *fpSource, int count)
{
unsigned short channel;
int amountToMove;
unsigned short stuffIndex;
unsigned char *pInsert;
int countOriginal;
i2eBordStrPtr pB;

	/* Ensure channel structure seems real */

	if ((pCh->validity & (CHANNEL_MAGIC_BITS | CHANNEL_SUPPORT)) !=
		(CHANNEL_MAGIC | CHANNEL_SUPPORT))
	{
		return -1;
	}
	/* initialize some accelerators and private copies */

	channel = pCh->infl.hd.i2sChannel;
	stuffIndex = pCh->obufStuff;
	countOriginal = count;
	pB = pCh->pMyBord;

	/* If the board has gone fatal, return bad, and also
	 * hit the trap routine if it exists
	 */

	if (pB->i2eFatal)
	{
		if (pB->i2eFatalTrap != NULLFUNC)
		{
			(*(pB)->i2eFatalTrap)(pB);
		}
		return -1;
	}

	/* Proceed as though we would do everything */
	while(count)
	{
		/* How much room is there? */

		amountToMove = pCh->obufStrip - stuffIndex - 1;
		if (amountToMove < 0)
		{
			amountToMove += OBUF_SIZE;
		}
		/* Subtract off the headers size and see how much
		 * room there is for real data.
		 */

		/* If this is negative, we will discover later */
		amountToMove -= sizeof (i2DataHeader);

		/* Don't move more than the count we were given */
		if (amountToMove > count)
		{
			amountToMove = count;
		}

		/* Don't move more (now) than can go in a single packet */

		if (amountToMove > MAX_OBUF_BLOCK - sizeof(i2DataHeader))
		{
			amountToMove = MAX_OBUF_BLOCK - sizeof(i2DataHeader);
		}

		/* Now we know how much we must move: Note, because the
		 * ring buffers have an overflow area at the end, we
		 * needn't worry about wrapping in the middle of a packet.
		 */

		if (amountToMove > 0)
		{
			/* Had room to move some data: don't know whether the
			 * block size, buffer space, or what was the limiting
			 * factor...
			 */

			pInsert = &(pCh->obuf[stuffIndex]);

			/* Set up the header */
			CHANNEL_OF(pInsert)		= channel;
			PTYPE_OF(pInsert)		= PTYPE_DATA;
			TAG_OF(pInsert)			= 0;
			ID_OF(pInsert)			= ID_ORDINARY_DATA;
			DATA_COUNT_OF(pInsert)	= amountToMove;

			/* Move the data */
			i2moveData(fpSource, (char far *)(DATA_OF(pInsert)), amountToMove);
			
			/* Adjust pointers and indices */
			fpSource += amountToMove;
			
			stuffIndex += amountToMove + sizeof(i2DataHeader);
			if (stuffIndex >= OBUF_SIZE)
			{
				stuffIndex = 0;
			}
			pCh->obufStuff = stuffIndex;	

			count -= amountToMove;

		}
		else
		{
			/* Cannot move data */

			serviceOutgoingFifo(pB);	/* Try to throw more things
										 * (maybe not us) in the fifo
										 * if we're not already waiting
										 * for it
										 */

			if (pCh->channelOptions & CO_NBLOCK_WRITE)
			{
				break;	/* from while(count) */
			}

			/* Otherwise, perform a little delay before trying again */
			iiDelay(pB, 1);
		}
	} /* End of while(count) */

	/* We drop through either when the count expires, or
	 * when there is some count left, but there was a
	 * non-blocking write.
	 */

	if (countOriginal > count)
	{
		i2QueueNeeds(pB, pCh, NEED_INLINE);
	}
	return countOriginal - count;
} 



/*---------------------------------------------
 * i2FlushOutput(pCh)
 * 
 * pCh -- pointer to a channel structure
 *
 * Sends bypass command to start flushing (waiting
 * possibly forever until there is room), then
 * sends inline command to stop flushing output.
 * (again waiting possibly forever).
 *----------------------------------------------
 */

void 
i2FlushOutput(i2ChanStrPtr pCh)
{
	i2QueueCommands(PTYPE_BYPASS,pCh, -1, 1, CMD_STARTFL);
	i2QueueCommands(PTYPE_INLINE,pCh, -1, 1, CMD_STOPFL);
}	


/*------------------------------------------
 * i2DrainOutput(pCh, timeout)
 *
 * Waits up to timeout mS for the output on
 * channel pCh to drain.
 * 
 * Uses the bookmark request command to ask the
 * board to send a bookmark back as soon as
 * all the data is completely sent.
 *
 *-------------------------------------------
 */

int	
i2DrainOutput(i2ChanStrPtr pCh, int timeout)
{
i2eBordStrPtr pB;

	pB = pCh->pMyBord;
	/* If the board has gone fatal, return bad, and also
	 * hit the trap routine if it exists
	 */

	if (pB->i2eFatal)
	{
		if (pB->i2eFatalTrap != NULLFUNC)
		{
			(*(pB)->i2eFatalTrap)(pB);
		}
		return -1;
	}

	while (timeout)
	{
		if (i2QueueCommands(PTYPE_INLINE, pCh, 1, 1, CMD_BMARK_REQ))
		{
			break;	/* from while */
		}
		else
		{
			timeout--;
		}
	}
	while (timeout)
	{
		if (pCh->bookMarks == 0) return 1;
		iiDelay(pB, 1);
		timeout--;
	}
	return 0;
}


/*-------------------------------------
 * i2OutputFree(pCh)
 *
 * pCh -- pointer to a channel structure
 *
 * Returns -1 if very gross error.
 * otherwise returns the amount of bytes
 * still free in the output buffer.
 *-------------------------------------
 */


int
i2OutputFree(i2ChanStrPtr pCh)
{
int amountToMove;
unsigned short stuffIndex;

	/* Ensure channel structure seems real */

	if ((pCh->validity & (CHANNEL_MAGIC_BITS | CHANNEL_SUPPORT)) !=
		(CHANNEL_MAGIC | CHANNEL_SUPPORT))
	{
		return -1;
	}

	stuffIndex = pCh->obufStuff;

	amountToMove = pCh->obufStrip - stuffIndex - 1;

	if (amountToMove < 0)
	{
		amountToMove += OBUF_SIZE;
	}

	/* If this is negative, we will discover later */
	amountToMove -= sizeof (i2DataHeader);

	return (amountToMove < 0) ? 0 : amountToMove;
}



/*---------------------------------------
 * i2StripFifo(pB)
 *
 * pB	-- pointer to the board structure
 *
 * Strips all the available data from the
 * incoming FIFO, identifies the type of
 * packet, and either buffers the data or
 * does what needs to be done.
 *
 * Note there is no overflow checking here:
 * if the board sends more data than it
 * ought to, we will not detect it here,
 * but blindly overflow...
 *---------------------------------------
 */
 
/* A buffer for reading in blocks for unknown channels */

static unsigned char
junkBuffer[IBUF_SIZE];


/* A buffer to read in a status packet.  Because of the
 * size of the count field for these things, the maximum
 * packet size must be less than 
 * MAX_CMD_PACK_SIZE
 */

static unsigned char
cmdBuffer[MAX_CMD_PACK_SIZE + 4];

/* This table changes the bit order from MSR order
 * given by STAT_MODEM packet to status bits used in our
 * library.
 */

static char xlatDss[16] = {
0      | 0     | 0      | 0      ,
0      | 0     | 0      | I2_CTS ,
0      | 0     | I2_DSR | 0      ,
0      | 0     | I2_DSR | I2_CTS ,
0      | I2_RI | 0      | 0      ,
0      | I2_RI | 0      | I2_CTS ,
0      | I2_RI | I2_DSR | 0      ,
0      | I2_RI | I2_DSR | I2_CTS ,
I2_DCD | 0     | 0      | 0      ,
I2_DCD | 0     | 0      | I2_CTS ,
I2_DCD | 0     | I2_DSR | 0      ,
I2_DCD | 0     | I2_DSR | I2_CTS ,
I2_DCD | I2_RI | 0      | 0      ,
I2_DCD | I2_RI | 0      | I2_CTS ,
I2_DCD | I2_RI | I2_DSR | 0      ,
I2_DCD | I2_RI | I2_DSR | I2_CTS };


static void
i2StripFifo(i2eBordStrPtr pB)
{
i2ChanStrPtr pCh;
int channel;
int count;
unsigned short stuffIndex;
int amountToRead;
unsigned char *pc, *pcLimit;
unsigned char uc;


	while(HAS_INPUT(pB))
	{
		/* The first word (or two bytes) will have channel number
		 * and type of packet, possibly other information
		 */
		pB->i2eLeadoffWord[0] = iiReadWord(pB);

		switch(PTYPE_OF(pB->i2eLeadoffWord))
		{
		case PTYPE_DATA:
			channel = CHANNEL_OF(pB->i2eLeadoffWord);	/* Store channel */
			count = iiReadWord(pB);				/* Count is in the next word */

			/* NEW: Check the count for sanity! */
			/* Should the hardware fail, our death is more pleasant */
			/* While an oversize channel is acceptable (just more than
			 * the driver supports), an over-length count clearly means
			 * we are sick!
			 */

			if (((unsigned int)count) > IBUF_SIZE)
			{
				pB->i2eFatal = 2;
				return;		/* Bail out ASAP */
			}

			if (channel >= pB->i2eChannelCnt)	/* Channel is illegally big ?*/
			{
				iiReadBuf(pB, junkBuffer, count);
				break;			/* From switch: ready for next packet */
			}
		
			/* Channel should be valid, then */

			pCh = &(((i2ChanStrPtr)(pB->i2eChannelPtr))[channel]);

			/* If this is a hot-key, merely post its receipt for now 
			 * These are always supposed to be 1-byte packets, so
			 * we won't even check the count.
			 * Also we will post an acknowledgement to the board so
			 * that more data can be forthcoming.
			 * Note that we are not trying to use these sequences in this
			 * driver, merely to robustly ignore them.
			 */

			if(ID_OF(pB->i2eLeadoffWord) == ID_HOT_KEY)
			{
				pCh->hotKeyIn = iiReadWord(pB) & 0xff;
				i2QueueCommands(PTYPE_BYPASS, pCh, 0,
					1, CMD_HOTACK);
				break;	/* From the switch: ready for next packet */
			}

			/* Normal data! 
			 * we crudely assume there is room for the data in our
			 * buffer because the board wouldn't have exceeded his
			 * credit limit.
			 */

			stuffIndex = pCh->ibufStuff;
			amountToRead = IBUF_SIZE - stuffIndex;
			if (amountToRead > count) amountToRead = count;

			/* stuffIndex would have been already adjusted so 
			 * there would always be room for at least one,
			 * and count is always at least one.
			 */

			iiReadBuf(pB, &(pCh->ibuf[stuffIndex]), amountToRead);

			/* Update the stuffIndex by the amount of data moved */
			/* Note we could never ask for more data than would
			 * just fit.  However, we might have read in one more
			 * byte than we wanted because the read rounds up to
			 * even bytes.  If this byte is on the end of the packet,
			 * and is padding, we ignore it.  If the byte is part of
			 * the actual data, we need to move it.
			 */

			stuffIndex += amountToRead;

			if (stuffIndex >= IBUF_SIZE)
			{
				if ((amountToRead & 1) && (count > amountToRead))
				{
					pCh->ibuf[0] = pCh->ibuf[IBUF_SIZE];
					amountToRead++;
					stuffIndex = 1;
				}
				else
				{
					stuffIndex = 0;
				}
			}

			/* If there is anything left over, read it as well */
			if (count > amountToRead)
			{
				amountToRead = count - amountToRead;
				iiReadBuf(pB, &(pCh->ibuf[stuffIndex]), amountToRead);
				stuffIndex += amountToRead;
			}

			/* Update stuff index */

			pCh->ibufStuff = stuffIndex;

			/* Note we do not need to maintain any flow-control credits
			 * at this time:  if we were to increment .asof and decrement
			 * .room, there would be no net effect.
			 * Instead, when we strip data, we will increment .asof and
			 * leave .room unchanged.
			 */

			break;	/* From switch: ready for next packet */
			


		case PTYPE_STATUS:
			count = CMD_COUNT_OF(pB->i2eLeadoffWord);

			iiReadBuf(pB, cmdBuffer, count);

			pc = cmdBuffer;
			pcLimit = &(cmdBuffer[count]);

			while(pc < pcLimit)
			{
				channel = *pc++;	
				if (channel < pB->i2eChannelCnt)
				{
					pCh = &(((i2ChanStrPtr)(pB->i2eChannelPtr))[channel]);
	
					switch (uc = *pc++)
					{
					/* Breaks, Modem signals are easy: just update status */
					case STAT_CTS_UP:
						pCh->dataSetIn |= I2_CTS;
						break;
	
					case STAT_CTS_DN:
						pCh->dataSetIn &= ~I2_CTS;
						break;
	
					case STAT_DCD_UP:
						pCh->dataSetIn |= I2_DCD;
						break;
	
					case STAT_DCD_DN:
						pCh->dataSetIn &= ~I2_DCD;
						break;
	
					case STAT_DSR_UP:
						pCh->dataSetIn |= I2_DSR;
						break;
	
					case STAT_DSR_DN:
						pCh->dataSetIn &= ~I2_DSR;
						break;
	
					case STAT_RI_UP:
						pCh->dataSetIn |= I2_RI ;
						break;
	
					case STAT_RI_DN:
						pCh->dataSetIn &= ~I2_RI ;
						break;
	
					case STAT_BRK_DET:
						pCh->dataSetIn |= I2_BRK;
						break;
						
					/* Bookmarks? one less request we're waiting for */
					case STAT_BMARK:
						pCh->bookMarks--;
						break;

					/* Flow control packets? Update the new credits,
					 * and if someone was waiting for output, queue
					 * him up again.
					 */

					case STAT_FLOW:
						pCh->outfl.room = 
							((flowStatPtr)pc)->room +
							((flowStatPtr)pc)->asof -
							pCh->outfl.asof;

						if (pCh->channelNeeds & NEED_CREDIT)
						{
							pCh->channelNeeds &= ~NEED_CREDIT;
							i2QueueNeeds(pB, pCh, NEED_INLINE);
						}
	
						pc += sizeof(flowStat);
						break;
							
					/* Special packets: */
					/* Just copy the information into the channel structure */

					case STAT_STATUS:

						pCh->channelStatus = *((debugStatPtr)pc);
						pc += sizeof(debugStat);
						break;

					case STAT_TXCNT:

						pCh->channelTcount = *((cntStatPtr)pc);
						pc += sizeof(cntStat);
						break;

					case STAT_RXCNT:

						pCh->channelRcount = *((cntStatPtr)pc);
						pc += sizeof(cntStat);
						break;

					case STAT_BOXIDS:

						pCh->channelBtypes = *((bidStatPtr)pc);
						pc += sizeof(bidStat);
						break;

					case STAT_HWFAIL:

						pCh->channelFail = *((failStatPtr)pc);
						pc += sizeof(failStat);
						break;

						
					/* No explicit match? then 
					 * Might be an error packet... 
					 */
					default:
						switch (uc & STAT_MOD_ERROR)
						{
						case STAT_ERROR:
							if (uc & STAT_E_PARITY) pCh->dataSetIn |= I2_PAR;
							if (uc & STAT_E_FRAMING) pCh->dataSetIn |= I2_FRA;
							if (uc & STAT_E_OVERRUN) pCh->dataSetIn |= I2_OVR;
							break;

						case STAT_MODEM:
							pCh->dataSetIn = (pCh->dataSetIn 
								& ~(I2_RI | I2_CTS | I2_DCD | I2_DSR) )
								| xlatDss[uc & 0xf];
						default:
							break;
						}
						break;
	
					}	/* End of switch on status type */
				}	
				else	/* Or else, channel is invalid */
				{
					/* Even though the channel is invalid, we
					 * must test the status to see how much
					 * additional data it has (to be skipped)
					 */

					switch (*pc++)
					{
					case STAT_FLOW:
						pc += 4;		/* Skip the data */
						break;

					case STAT_CTS_UP:
					case STAT_CTS_DN:
					case STAT_DCD_UP:
					case STAT_DCD_DN:
					case STAT_DSR_UP:
					case STAT_DSR_DN:
					case STAT_RI_UP:
					case STAT_RI_DN:
					case STAT_BRK_DET:
					case STAT_BMARK:
					default:
						break;
					}
				}
	
			}	/* End of while (there is still some status packet left) */

		default:	/* Neither packet? should be impossible */
			break;

		}	/* End of switch on type of packets */
	}

	/* Send acknowledgement to the board even if there was no data! */
	pB->i2eOutMailWaiting |= MB_IN_STRIPPED;
}


/*-------------------------------------
 * i2StuffFifoBypass(pB)
 *
 * pB  -- pointer to the board structure
 *
 * Stuffs as many bypass commands into
 * the fifo as possible.
 *
 * This is simpler than stuffing data or
 * inline commands to fifo, since we
 * do not have flow-control to
 * deal with.
 *-------------------------------------
 */

static void
i2StuffFifoBypass(i2eBordStrPtr pB)
{
i2ChanStrPtr pCh;
unsigned short queueStrip = pB->i2bbufStrip;
unsigned short remains = pB->i2eFifoRemains;
unsigned short stripIndex;
unsigned char *pRemove;
unsigned short packetSize;
unsigned short paddedSize;
unsigned short notClogged = 1;
unsigned short wroteSomething = 0;


	/* Continue processing so long as there are entries,
	 * or there is room in the fifo.
	 * Each entry represents a channel with something to do.
	 */

	while (notClogged && (queueStrip != pB->i2bbufStuff))
	{
		pCh = pB->i2bBuf[queueStrip];	/* pick up the channel pointer */
		stripIndex = pCh->cbufStrip;

		/* as long as there are packets for this channel... */

		while(stripIndex != pCh->cbufStuff)
		{
			pRemove = &(pCh->cbuf[stripIndex]);
			packetSize = CMD_COUNT_OF(pRemove) + sizeof(i2CmdHeader);
			paddedSize = ROUNDUP(packetSize);
			if (remains < paddedSize)

			{
				/* If fifo has no room for the next channel on
				 * the queue, stop right here with that channel
				 * still at the head of the queue.
				 */

				notClogged = 0;
				break;	/* Break from the channel */
			}
			/* Can write to board now */

			iiWriteBuf(pB, pRemove, packetSize);

			wroteSomething = 1;
			remains -= paddedSize;
			pRemove += packetSize;
			stripIndex += packetSize;
			if (stripIndex >= CBUF_SIZE)
			{
				stripIndex = 0;
				pRemove = pCh->cbuf;
			}
		}
		/* Done with this channel. Move to next,
		 * removing this one from the queue of channels
		 * if we cleaned it out (i.e., didn't get clogged.
		 */

		pCh->cbufStrip = stripIndex;
		if (notClogged)
		{
			pCh->channelNeeds &= ~NEED_BYPASS;
			if (++queueStrip >= CH_QUEUE_SIZE)
			{
				queueStrip = 0;
			}
		}
	}	/* Either clogged or finished all the work */

	/* Update queue pointers and how much room in FIFO there is */

	pB->i2bbufStrip = queueStrip;
	pB->i2eFifoRemains = remains;

	if (wroteSomething)
	{
		pB->i2eOutMailWaiting |= MB_OUT_STUFFED;
	}
}
			
			

/*-------------------------------------
 * i2StuffFifoFlow(pB)
 *
 * pB  -- pointer to the board structure
 *
 * Stuffs as many flow control packets
 * into FIFO as possible
 *
 * This is easier even than doing normal
 * bypass commands, because there is
 * always at most one packet, already
 * assembled, for each channel.
 *-------------------------------------
 */

static void
i2StuffFifoFlow(i2eBordStrPtr pB)
{
i2ChanStrPtr pCh;
unsigned short queueStrip = pB->i2fbufStrip;
unsigned short remains = pB->i2eFifoRemains;
unsigned short wroteSomething = 0;
unsigned short paddedSize = ROUNDUP(sizeof(flowIn));


	/* Continue processing so long as there are entries,
	 * or there is room in the fifo.
	 * Each entry represents a channel with something to do.
	 */

	while (queueStrip != pB->i2fbufStuff)
	{
		pCh = pB->i2fBuf[queueStrip];	/* pick up the channel pointer */

		if (remains < paddedSize)
		{
			break;	
		}

		iiWriteBuf(pB,(unsigned char *) &(pCh->infl), sizeof(flowIn));

		wroteSomething = 1;
		remains -= paddedSize;

		pCh->channelNeeds &= ~NEED_FLOW;
		if (++queueStrip >= CH_QUEUE_SIZE)
		{
			queueStrip = 0;
		}

	}	/* Either clogged or finished all the work */

	/* Update queue pointers and how much room in FIFO there is */

	pB->i2fbufStrip = queueStrip;
	pB->i2eFifoRemains = remains;

	if (wroteSomething)
	{
		pB->i2eOutMailWaiting |= MB_OUT_STUFFED;
	}
}
			
			

/*-------------------------------------
 * i2StuffFifoInline(pB)
 *
 * pB  -- pointer to the board structure
 *
 * Stuffs as much data and inline commands 
 * into the fifo as possible.
 *
 * This is the most complex fifo-stuffing
 * operation, since there if now the
 * channel flow-control issue to deal
 * with.
 *-------------------------------------
 */

static void
i2StuffFifoInline(i2eBordStrPtr pB)
{
i2ChanStrPtr pCh;
unsigned short queueStrip = pB->i2dbufStrip;
unsigned short remains = pB->i2eFifoRemains;
unsigned short stripIndex;
unsigned char *pRemove;
unsigned short packetSize;
unsigned short paddedSize;
unsigned short notClogged = 1;
unsigned short wroteSomething = 0;
unsigned short flowsize;


	/* Continue processing so long as there are entries,
	 * or there is room in the fifo.
	 * Each entry represents a channel with something to do.
	 */

	while (notClogged && (queueStrip != pB->i2dbufStuff))
	{
		pCh = pB->i2dBuf[queueStrip];	/* pick up the channel pointer */
		stripIndex = pCh->obufStrip;

		/* as long as there are packets for this channel... */

		while(stripIndex != pCh->obufStuff)
		{
			pRemove = &(pCh->obuf[stripIndex]);

			/* Must determine whether this be a data or command packet 
			 * to correctly calculate the header size and the amount
			 * of flow-control credit this type of packet will use.
			 */

			if (PTYPE_OF(pRemove) == PTYPE_DATA)
			{
				flowsize = DATA_COUNT_OF(pRemove);
				packetSize = flowsize + sizeof(i2DataHeader);
			}
			else
			{
				flowsize = CMD_COUNT_OF(pRemove);
				packetSize = flowsize + sizeof(i2CmdHeader);
			}	
			flowsize = CREDIT_USAGE(flowsize);
			paddedSize = ROUNDUP(packetSize);
			if (remains < paddedSize)

			{
				/* If fifo has no room for the next channel on
				 * the queue, stop right here with that channel
				 * still at the head of the queue.
				 */

				notClogged = 0;
				break;	/* Break from the channel */
			}

			/* Now, we must ensure we have the credits to send
			 * this packet.
			 */

			/* If we don't have enough credits from the board
			 * to send the data, flag the channel that we are
			 * waiting for flow control credit, and break out.
			 * This will clean up this channel and remove us
			 * from the queue of hot things to do.
			 */

			if (pCh->outfl.room < flowsize)
			{
				pCh->channelNeeds |= NEED_CREDIT;
				break;	/* So to do next channel */
			}

			/* Can write to board now */

			iiWriteBuf(pB, pRemove, packetSize);

			/* Update current credits */
			pCh->outfl.room -= flowsize;
			pCh->outfl.asof += flowsize;

			wroteSomething = 1;
			remains -= paddedSize;
			pRemove += packetSize;
			stripIndex += packetSize;
			if (stripIndex >= OBUF_SIZE)
			{
				stripIndex = 0;
				pRemove = pCh->obuf;
			}
		}
		/* Done with this channel. Move to next,
		 * removing this one from the queue of channels
		 * if we cleaned it out (i.e., didn't get clogged.
		 */

		pCh->obufStrip = stripIndex;
		if (notClogged)
		{
			pCh->channelNeeds &= ~NEED_INLINE;
			if (++queueStrip >= CH_QUEUE_SIZE)
			{
				queueStrip = 0;
			}
		}
	}	/* Either clogged or finished all the work */

	/* Update queue pointers and how much room in FIFO there is */

	pB->i2dbufStrip = queueStrip;
	pB->i2eFifoRemains = remains;

	if (wroteSomething)
	{
		pB->i2eOutMailWaiting |= MB_OUT_STUFFED;
	}
}
		

/*---------------------------------------------------
 * i2InterruptHandler
 *
 * The interrupt handler.
 * This must not have any local variables: however,
 * it can call functions with local variables.
 *---------------------------------------------------
 */

static i2eBordStrPtr *ppB;
static i2eBordStrPtr *ppBlimit;

void interrupt
i2InterruptHandler()
{
	/* Set up new stack and data segments (see I2OS.H) */

	TO_NEW_STACK(i2_superstack, STAKSIZE)

	/* Mask out this interrupt type from further interrupts */
	/* This prevents MC and EISA boards, which generate level triggered
	 * interrupts, from posting another interrupt as soon as the
	 * EOI is issued.
	 */

	SAVE_IRQ_MASKS(irqSaveMask);
	irqNewMask = irqSaveMask | myMaskBits;
	WRITE_IRQ_MASKS(irqNewMask);

	/* Issue EOI to master and (optionally) slave controller */
	/* We must do this before servicing the board, to prevent
	 * losing interrupts from ISA boards, which issue edge-triggered
	 * interrupts.
	 */

	if (issueSlaveEoi)
	{
		SLAVE_EOI
	}
	MASTER_EOI

	/* Service every board on the list */
	ppB = i2BoardPtrTable;
	ppBlimit = &(i2BoardPtrTable[i2nBoards]);

	while(ppB < ppBlimit)
	{
		i2ServiceBoard(*ppB++);	
	}

	/* Restore the interrupt masks */

	WRITE_IRQ_MASKS(irqSaveMask);

	/* Restore old stack before returning */
	TO_OLD_STACK(i2_superstack)
}

	

/*---------------------------------------------
 * i2ServiceBoard(pB)
 *
 * pB -- pointer to board structure
 *
 * Normally this is called from interrupt level,
 * but there is deliberately nothing in here
 * specific to being called from interrupt level.
 * All the hardware-specific, interupt-specific
 * things happen at the outer levels.
 *
 * For example, a timer interrupt could drive this
 * routine for some sort of polled operation.
 * The only requirement is that the programmer
 * deal with any atomiticity/concurrency issues
 * that result.
 *
 * This routine responds to the board's having
 * sent mailbox information to the host (which would
 * normally cause an interrupt).  This routine
 * reads the incoming mailbox. If there is no data
 * in it, this board did not create the interrupt
 * and/or has nothing to be done to it. (Except, if
 * we have been waiting to write mailbox data to it,
 * we may do so.
 *
 * Based on the value in the mailbox, we may take
 * various actions.
 *
 * No checking here of pB validity: after all, it
 * shouldn't have been called by the handler unless
 * pB were on the list.  
 *-----------------------------------------------
 */

void
i2ServiceBoard(register i2eBordStrPtr pB)
{
unsigned inmail;

	inmail = iiGetMail(pB);	
	if (inmail != NO_MAIL_HERE)
	{
		/* If the board has gone fatal, nothing to do but
		 * hit a bit that will alert foreground tasks
		 * to protest!
		 */

		if (inmail & MB_FATAL_ERROR)
		{
			pB->i2eFatal = 1;
			return;
		}

		/* Assuming no fatal condition, we proceed to do work */
		if (inmail & MB_IN_STUFFED)
		{
			pB->i2eFifoInInts++;
			i2StripFifo(pB);		/* There might be incoming packets */
		}
	
		if (inmail & MB_OUT_STRIPPED)
		{
			pB->i2eFifoOutInts++;
			pB->i2eWaitingForEmptyFifo = 0;
			pB->i2eFifoRemains = pB->i2eFifoSize;
		}
	
		serviceOutgoingFifo(pB);
	}

	/* Whether there was mail or not, try to send
	 * a mailbox interrupt if one has been waiting
	 */

	iiSendPendingMail(pB);
}

/*--------------------------------------
 * serviceOutgoingFifo(pB)
 *
 * Helper routine to put data in the
 * outgoing fifo, if we aren't already
 * waiting for something to be there
 * 
 * If the fifo has only room for a very
 * little data, go head and hit the board
 * with a mailbox hit immediately.
 * Otherwise, it will have to happen later
 * in the interrupt processing.
 *
 * Since this routine may be called both
 * at interrupt and foreground time, we
 * must turn off interrupts during the 
 * entire process.
 * -------------------------------------
 */


void
serviceOutgoingFifo(i2eBordStrPtr pB)
{	
	SAVE_AND_DISABLE_INTS

	/* If we aren't currently waiting for the board to
	 * empty our fifo, service everything that is 
	 * pending, in priority order (especially, Bypass
	 * before Inline).
	 */
	if (pB->i2eWaitingForEmptyFifo == 0)
	{
		i2StuffFifoFlow(pB);
		i2StuffFifoBypass(pB);
		i2StuffFifoInline(pB);
	}

	/* If less than 32 bytes remain in the fifo, that's little
	 * enough that we don't need to wait for any more: hit the
	 * board now.
	 */

	if (pB->i2eFifoRemains < 32)
	{
		iiSendPendingMail(pB);
	}
	RESTORE_INTS
}

