/*-------------------------------------
 * I2LIB.H
 *
 * IntelliPort-II and IntelliPort-IIEX
 *
 * Defines, structure definitions, and
 * external declarations for i2lib.c
 *-------------------------------------
 */

/*-------------------------------------
 * Cosmetic note:
 * Written using tabstops at 4 spaces
 * vi: set ts=4
 * pr: -e4
 *-------------------------------------
 */

#ifndef I2LIB_H		/* To prevent multiple inclusions */
#define I2LIB_H	1

/*--------------------------------------
 * Mandatory Includes:
 *--------------------------------------
 */
#include "i2ellis.h"
#include "i2pack.h"
#include "i2cmd.h"


/*------------------------------------
 * i2ChanStr -- Channel Structure:
 * Used to track per-channel information
 * for the library routines using
 * standard loadware.
 * Note also, a pointer to an array of 
 * these structures is patched into
 * the i2eBordStr (see i2ellis.h)
 *------------------------------------
 */

/* If we make some limits on the maximum block sizes, we can
 * avoid dealing with buffer wrap:  The wrapping of the buffer is
 * based on where the start of the packet is. Then there is always
 * room for the packet contiguously
 */
#define MAX_OBUF_BLOCK	36	/* Maximum total length of an outgoing data
							 * or in-line command block.
							 * The limit of 36 on data is quite arbitrary
							 * and based more on DOS memory limitations
							 * than the board interface.  However, for
							 * commands, the maximum packet length is
							 * MAX_CMD_PACK_SIZE, because the field
							 * size for the count is only a few bits
							 * (see I2PACK.H) in such packets.  For
							 * data packets, the count field size is
							 * not the limiting factor.
							 * As of this writing, MAX_OBUF_BLOCK <
							 * MAX_CMD_PACK_SIZE, but be careful if
							 * wanting to modify either.
							 */
							
					/* Another note on maximum block sizes: 
					 * We are buffering packets here.  Data is put
					 * into the buffer (if there is room) regardless
					 * of the credits from the board.  The board sends
					 * new credits whenever it has removed from his
					 * buffers a number of characters equal to 80% of
					 * total buffer size.  (Of course, the total buffer
					 * size is what is reported when the very first
					 * set of flow control status packets are received
					 * from the board.  Therefore, to be robust, you must
					 * always fill the board to at least 80% of the current
					 * credit limit, else you might not give it enough
					 * to trigger a new report.  
					 * These conditions are obtained here so long as the
					 * maximum output block size is less than 20% the size
					 * of the board's output buffers.  This is true at 
					 * present by "coincidence" or "infernal knowledge":
					 * the board's output buffers are at least 700 bytes
					 * long (20% = 140 bytes, at least).  The 80% figure
					 * is "official", so the safest strategy might be to
					 * trap the first flow control report and guarantee 
					 * that the effective maxObufBlock is the minimum of
					 * MAX_OBUF_BLOCK and 20% of first reported buffer credit.
					 */

#define MAX_CBUF_BLOCK	6	/* Maximum total length of a bypass command block*/


#define IBUF_SIZE 50	/* character capacity of input buffer per channel */
#define OBUF_SIZE 96	/* character capacity of output buffer per channel */
#define CBUF_SIZE 10	/* character capacity of output bypass buffer */


typedef struct _i2ChanStr {
	
	/* First, back-pointers so that given a pointer to this
	 * structure, you can determine the correct board and
	 * channel number to reference, (say, when issuing commands,
	 * etc. (Note, channel number is in infl.hd.i2sChannel)
	 */

	unsigned short	validity;			/* Indicates whether the
										 * given channel has been initialized,
										 * really exists (or is a missing
										 * channel-- channel 9 on an 8-port
										 * box, for example.
										 */
										
	i2eBordStrPtr	pMyBord;			/* Back-pointer to the board
										 * structure this belongs to
										 */

	flowIn			infl;				/* This structure is initialized
										 * as a completely formed flow-control
										 * command packet, and as such has
										 * the channel number, also the
										 * capacity and "as-of" data 
										 * needed continuously.
										 */
											
	unsigned short	sinceLastFlow;		/* Counts the number of characters 
										 * read from input buffers, since
										 * the last time flow control info
										 * was sent.
										 */

	unsigned short 	whenSendFlow;		/* Determines when new flow 
										 * control is to be sent to 
										 * the board.  Note unlike earlier
										 * manifestations of the driver, 
										 * these packets can be sent
										 * from in-place.
										 */

	unsigned short	channelNeeds;		/* Bit map of important things
										 * which must be done for this
										 * channel. (See bits below )
										 */

	flowStat		outfl;				/* Same type of structure is used to
										 * hold current flow control
										 * information used to control our
										 * output.  "asof" is kept updated
										 * as data is sent, and "room"
										 * never goes to zero.
										 */
	
	/* The incoming ring-buffer: */
	/* Unlike the outgoing buffers, this holds raw data, not packets */

	unsigned char	ibuf[IBUF_SIZE + 2];/* Input ring buffer */
										/* The two extra bytes are used
										 * to hold the byte-padding when
										 * there is room for an odd number
										 * of bytes before we must wrap.
										 */
	unsigned short	ibufStuff;			/* Stuffing index */
	unsigned short	ibufStrip;			/* Stripping index */


	/* The outgoing ring-buffer: Holds Data and command packets.
	 * N.B., even though these are in the channel structure, the
	 * channel is also written here, the easier to send it to
	 * the fifo when ready.  HOWEVER, individual packets here are
	 * NOT padded to even length: the routines for writing blocks
	 * to the the fifo will pad to even byte counts.
	 */

	unsigned char	obuf[OBUF_SIZE+MAX_OBUF_BLOCK+4];	
										/* Output ring buffer */
	unsigned short	obufStuff;			/* Stuffing index */
	unsigned short	obufStrip;			/* Stripping index */

	/* The outgoing bypass-command buffer 
	 * (Unlike earlier manifestations, the flow control packets are
	 * sent directly from the structures.  
	 * As above, the channel number is included in the packet,
	 * but they are NOT padded to even size.
	 */

	unsigned char	cbuf[CBUF_SIZE+MAX_CBUF_BLOCK+2];	
										/* Bypass command ring buffer */
	unsigned short	cbufStuff;			/* Stuffing index */
	unsigned short	cbufStrip;			/* Stripping index */	

	/* The state of incoming data-set signals */

	unsigned short	dataSetIn;		/* Bit-mapped according to below */
									/* Also indicates whether a break has
									 * been detected since last inquiry
									 */

	/* Most recent hot-key identifier detected */

	unsigned short	hotKeyIn;		/* Hot key as sent by the board,
									 * HOT_CLEAR indicates no hot
									 * key detected since last examined.
									 */

	/* Counter of outstanding requests for bookmarks */

	unsigned short	bookMarks;		/* Number of outstanding bookmark
									 * requests. (+ whenever a bookmark
									 * request if queued up, - whenever
									 * a bookmark is received
									 */
	/* Misc options */

	unsigned short	channelOptions;	/* See below */

	/* To store various incoming special packets */

	debugStat	channelStatus;
	cntStat		channelRcount;
	cntStat		channelTcount;
	bidStat		channelBtypes;
	failStat	channelFail;

} i2ChanStr, *i2ChanStrPtr;


/*---------------------------------------------------
 * Manifests and bit-mapps for elements in i2ChanStr
 *---------------------------------------------------
 */

/* validity */
#define CHANNEL_MAGIC_BITS	0xff00
#define CHANNEL_MAGIC		0x5300	/* (validity & CHANNEL_MAGIC_BITS)
									 * == CHANNEL_MAGIC --> structure good 
									 */

#define CHANNEL_SUPPORT		0x0001	/* Indicates channel is supported */
									/* Exists, and passed P.O.S.T. */

/* channelNeeds */
#define	NEED_FLOW		1		/* Indicates flow control has been queued */
#define NEED_INLINE		2		/* Indicates inline commands or data queued */
#define NEED_BYPASS		4		/* Indicates bypass commands queued */
#define	NEED_CREDIT		8		/* Indicates would be sending except has
								 * not sufficient credit. 
								 * The data is still in the channel structure,
								 * but the channel is not enqueued in the
								 * board structure again until there is
								 * a credit received from the board.
								 */


/* dataSetIn :
 * Also the bits for i2GetStatus return value
 */

#define I2_DCD	1
#define I2_CTS	2
#define I2_DSR	4
#define I2_RI	8
#define I2_BRK	0x10	/* Indicates a break was detected: 
						 * i2GetStatus() optionally clears the bit
						 */
						/* i2GetStatus() can optionally clear these bits */
#define	I2_PAR	0x20	/* Indicates a parity error was reveived */
#define	I2_FRA	0x40	/* Indicates a framing error was received */
#define I2_OVR	0x80	/* Indicates an overrun error was received */

/* hotKeyIn */
#define HOT_CLEAR	0x1322	/* Indicates that no hot-key has been detected */


/* channelOptions */
#define	CO_NBLOCK_WRITE	1	/* Writes don't block waiting for buffer. 
							/* (default is, they do wait) 
							 */

/*-------------------------------------------
 * Macros used from user level like functions
 *-------------------------------------------
 */

/* Macros to set and clear channel options */

#define	i2SetOption(pCh, option) pCh->channelOptions |= option
#define	i2ClrOption(pCh, option) pCh->channelOptions &= ~option


/* Macro to set fatal-error trap */
#define i2SetFatalTrap(pB, routine) pB->i2eFatalTrap = routine


/*--------------------------------------------
 * Declarations and prototypes for i2lib.c
 *--------------------------------------------
 */

extern void i2RestoreIrqs(void);
extern int  i2LinkEnableIrq(unsigned short, i2eBordStrPtr *, 
						int, void(far *)(void));
extern int  i2InitChannels(i2eBordStrPtr, int, i2ChanStrPtr);
extern int	i2CombineCommands(int, cmdSyntaxPtr, int, cmdSyntaxPtr,...);
extern int  i2QueueCommands(int, i2ChanStrPtr, int, int, cmdSyntaxPtr,...);
extern int  i2GetStatus(i2ChanStrPtr, int);
extern int  i2Input(i2ChanStrPtr, char far *, int);
extern int  i2InputLookahead(i2ChanStrPtr);
extern int  i2InputFlush(i2ChanStrPtr);
extern int  i2InputAvailable(i2ChanStrPtr);
extern void i2moveData(char far *, char far *, unsigned short);
extern int  i2Output(i2ChanStrPtr, char far *, int);
extern void i2FlushOutput(i2ChanStrPtr);
extern int  i2OutputFree(i2ChanStrPtr);
extern void interrupt i2InterruptHandler(void);
extern void i2ServiceBoard(i2eBordStrPtr);
extern int	i2DrainOutput(i2ChanStrPtr, int);



extern	i2eBordStrPtr i2BoardPtrTable[];	/* Table of board pointers
											 * for interrupt routine.
											 */
extern	unsigned short i2nBoards;			/* Number of boards */


/* Argument to i2QueueCommands */

#define	C_IN_LINE	1
#define C_BYPASS	0


#endif	/* I2LIB_H */
