/*-------------------------------------------------
 * i2os.h
 *
 * IntelliPort-II and IntelliPort-IIEX
 *
 * Defines, definitions, includes which
 * are heavily dependant on o.s., host,
 * compiler, etc.
 *
 * This file is tailored for
 *
 * DOS Environment (or similar)
 * Borland C++ 2.0 (Only C features used, however)
 * (and naturally 80x86 architecture)
 *--------------------------------------------------
 */

/*-------------------------------------
 * Cosmetic note:
 * Written using tabstops at 4 spaces
 * vi: set ts=4
 * pr: -e4
 *-------------------------------------
 */

#ifndef I2OS_H		/* To prevent multiple includes */
#define I2OS_H	1

/*--------------------------------------------------
 * Revision History:
 *
 * 9 October 1991	MAG	First draft
 *--------------------------------------------------
 */
 

/*-------------------------------------------------
 * Required Includes
 *-------------------------------------------------
 */

#include <dos.h>	/* For inportb, etc */



/*------------------------------------
 * Defines for i/o instructions:
 *
 * These assume the Borland C compiler
 * N.B. Borland implements these varieties
 * as in-line functions (although inportw
 * and outportw are not given in the library
 * reference, as are inportb and outportb.
 *-------------------------------------
 */

#define INB inportb
#define OUTB outportb
#define INW inportw
#define OUTW outportw

/*--------------------------------------------
 * Other Defines supported by Borland in-line
 * assembly...
 *
 * Note on register use:
 * The Borland C++ documentation guarantees 
 * the registers SI and DI may be freely used
 * in in-line assembler.
 *
 * In many circumstances, we would expect DX, CX
 * and AX should be safe as well, but we will
 * be extra cautious in these macros.
 * 
 * Except: I'm assuming it's ok to trash ES.
 *--------------------------------------------
 */

#define SAVE_AND_DISABLE_INTS	\
	asm { 						\
		pushf;					\
		cli; 					\
	}


/* the following weird code simulates a "popf", which doesn't work
 * on early 286 chips
 */

#define RESTORE_INTS 	\
	asm { 				\
		jmp short $+3; 	\
		iret; 			\
		push cs; 		\
		call $-2; 		\
	 }




/*-------------------------------------
 * In-line macros to perform string-in
 * and string-out operations.
 * 
 * While in-line versions of IN and OUT
 * are provided by Borland 2.0, I believe
 * there is no such thing for the string
 * i/o instructions.
 *-------------------------------------
 */

#define OUTSW(port,addr,count)	\
	asm {						\
		push cx;				\
		push dx;				\
		mov	si, addr;			\
		mov cx, count;			\
		inc cx;					\
		shr cx, 1;				\
		mov dx, port;			\
		rep outsw;				\
		pop dx;					\
		pop cx;					\
		}	



#define OUTSB(port,addr,count)	\
	asm {						\
		push cx;				\
		push dx;				\
		mov	si, addr;			\
		mov cx, count;			\
		inc cx;					\
		and cl, 0feh;			\
		mov dx, port;			\
		rep outsb;				\
		pop dx;					\
		pop cx;					\
		}



#define INSW(port,addr,count)	\
	asm {						\
		push cx;				\
		push dx;				\
		push ds;				\
		pop	es;					\
		mov	di, addr;			\
		mov cx, count;			\
		inc cx;					\
		shr cx, 1;				\
		mov dx, port;			\
		rep insw;				\
		pop dx;					\
		pop cx;					\
		}	


#define INSB(port,addr,count)	\
	asm {						\
		push cx;				\
		push dx;				\
		push ds;				\
		pop	es;					\
		mov	di, addr;			\
		mov cx, count;			\
		inc cx;					\
		and cl, 0feh;			\
		mov dx, port;			\
		rep insb;				\
		pop dx;					\
		pop cx;					\
		}	


/*-------------------------------
 * Hardware-delay loop
 * 
 * Probably used in only one place
 * (see i2ellis.c)
 * but this helps keep things 
 * together.
 * Note we have unwound the IN
 * instructions. On machines with
 * a reasonable cache, the eight
 * instructions (1 byte each) should
 * fit in cache nicely, and on un-cached
 * machines, the code-fetch would tend
 * not to dominate.  Note that cx is
 * shifted so that "count" still
 * reflects the total number of iterations
 * assuming no unwinding.
 *--------------------------------
 */


#define DELAY1MS(port,count,label)	\
		asm {						\
			push ax;				\
			push cx;				\
			push dx;				\
			mov	dx, port;			\
			mov cx, count;			\
			add cx, 7;				\
			shr cx, 1;				\
			shr cx, 1;				\
			shr cx, 1;				\
		}							\
label:								\
		asm {						\
			in al, dx;				\
			in al, dx;				\
			in al, dx;				\
			in al, dx;				\
			in al, dx;				\
			in al, dx;				\
			in al, dx;				\
			in al, dx;				\
			loop label;				\
			pop dx;					\
			pop	cx;					\
			pop ax;					\
		}


/*---------------------------------------
 * Macros to switch to a new stack,
 * saving stack pointers, and to
 * restore the old stack
 * (Used, for example, in i2lib.c)
 * "heap" is the address of some 
 * buffer which will become the new
 * stack (working down from highest
 * address).  The two words at the
 * two lowest addresses in this
 * stack are for storing the SS and SP
 *--------------------------------------
 */

#define TO_NEW_STACK(heap,size)		\
		asm {						\
			mov	heap, sp;			\
			mov	heap+2, ss;			\
			push ds;					\
			pop	ss;					\
			mov sp, offset heap+size+size;	\
		}

#define	TO_OLD_STACK(heap)			\
		asm {						\
			mov ss, heap+2;			\
			mov sp, heap;			\
		}



/*-----------------------------------------
 * Macros to save the original IRQ vectors
 * and masks, and to patch in new ones.
 *-----------------------------------------
 */

#define	SAVE_IRQ_MASKS(dest)		\
		asm {						\
			in	al, 0a1h;			\
			jmp	short $+2;			\
			jmp	short $+2;			\
			mov	ah, al;				\
			in	al, 21h;			\
			jmp short $+2;			\
			jmp short $+2;			\
			mov dest, ax;			\
		}

#define WRITE_IRQ_MASKS(src)		\
		asm	{						\
			mov	ax, src;			\
			out 21h, al;			\
			jmp short $+2;			\
			jmp short $+2;			\
			mov	al, ah;				\
			jmp short $+2;			\
			jmp short $+2;			\
			out	0a1h, al;			\
		}

			
#define	SAVE_IRQ_VECTOR(value,dest)	\
		asm {						\
			mov	al, byte ptr value;	\
			mov ah, 35h;			\
			int 21h;				\
			mov dest, bx;			\
			mov dest+2, es;			\
		}

#define WRITE_IRQ_VECTOR(value,src)	\
		asm {						\
			push ds;				\
			push dx;				\
			mov al, byte ptr value;	\
			lds dx, dword ptr src;	\
			mov ah, 25h;			\
			int 21h;				\
			pop dx;					\
			pop ds;					\
		}



/*-----------------------------------------
 * Macro to copy data from one far pointer
 * to another.
 *-----------------------------------------
 */

#define	I2_MOVE_DATA(fpSource,fpDest,count)	\
	asm {									\
		push	cx;							\
		push	ds;							\
		push	es;							\
		lds		si, dword ptr fpSource;		\
		les		di, dword ptr fpDest;		\
		cld;								\
		mov		cx,count;					\
		rep		movsb;						\
		pop		es;							\
		pop		ds;							\
		pop		cx;							\
	}


/*----------------------------------------
 * Macros to issue eoi's to host interrupt
 * control (IBM AT 8259-style
 *----------------------------------------
 */

#define MASTER_EOI		\
	asm {				\
		mov	al, 0x20;	\
		out	0x20, al;	\
		jmp short $+2;	\
		jmp short $+2;	\
	}


#define SLAVE_EOI		\
	asm {				\
		mov	al, 0x20;	\
		out	0xa0, al;	\
		jmp short $+2;	\
		jmp short $+2;	\
	}




#endif	/* I2OS_H */
